/*
# X-BASED HEXTRIS
#
#  HextrisN.c
#
###
#
#  Copyright (c) 1993 - 2006	David A. Bagley, bagleyd@tux.org
#
#			All Rights Reserved
#
#  Permission to use, copy, modify, and distribute this software and
#  its documentation for any purpose and without fee is hereby granted,
#  provided that the above copyright notice appear in all copies and
#  that both that copyright notice and this permission notice appear in
#  supporting documentation, and that the name of the author not be
#  used in advertising or publicity pertaining to distribution of the
#  software without specific, written prior permission.
#
#  This program is distributed in the hope that it will be "playable",
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
*/

/* Methods file for Hextris */
#include "file.h"
#include "rngs.h"
#include "HextrisNP.h"

static Point hexagon_unit[MAX_ORIENT][6] = {
	{ {0, 0}, {2, 0}, {1, 1}, {-1, 1}, {-2, 0}, {-1, -1} },
	{ {0, 0}, {1, 1}, {0, 2}, {-1, 1}, {-1, -1}, {0, -2} }
};
static Point hexagon_list[MAX_ORIENT][7];
static Point hexagon_border[7];

#ifdef WINVER
#ifndef LOGPATH
#define LOGPATH "/usr/tmp"
#endif
#ifndef INIFILE
#define INIFILE "whextris.ini"
#endif

#define SECTION "setup"

#else

#ifndef LOGPATH
#ifdef VMS
#define LOGPATH "SYS$SCRATCH:"
#else
#define LOGPATH "/usr/tmp"
#endif
#endif

static void ResizeNextGame(NextHextrisWidget w);
static void DestroyNextGame(Widget old);
#if 0
static void SizeNextGame(NextHextrisWidget w);
#endif
static Boolean SetValuesNextGame(Widget current, Widget request, Widget renew);
static void QuitNextGame(NextHextrisWidget w,
	XEvent *event, char **args, int nArgs);
static void InitializeNextGame(Widget request, Widget renew);
static void ExposeNextGame(Widget renew, XEvent *event, Region region);
static void HideNextGame(NextHextrisWidget w,
	XEvent *event, char **args, int nArgs);
static void EnterNextGame(NextHextrisWidget w,
	XEvent *event, char **args, int nArgs);
static void LeaveNextGame(NextHextrisWidget w,
	XEvent *event, char **args, int nArgs);
static void ShowNextNextGame(NextHextrisWidget w,
	XEvent *event, char **args, int nArgs);

static char defaultTranslationsNextGame[] =
"<KeyPress>q: Quit()\n\
 Ctrl<KeyPress>C: Quit()\n\
 <KeyPress>osfCancel: Hide()\n\
 <KeyPress>Escape: Hide()\n\
 <KeyPress>osfEscape: Hide()\n\
 Ctrl<KeyPress>[: Hide()\n\
 <KeyPress>0x1B: Hide()\n\
 <KeyPress>n: ShowNext()\n\
 <EnterWindow>: Enter()\n\
 <LeaveWindow>: Leave()";

static XtActionsRec actionsListNextGame[] =
{
	{(char *) "Quit", (XtActionProc) QuitNextGame},
	{(char *) "Hide", (XtActionProc) HideNextGame},
	{(char *) "ShowNext", (XtActionProc) ShowNextNextGame},
	{(char *) "Enter", (XtActionProc) EnterNextGame},
	{(char *) "Leave", (XtActionProc) LeaveNextGame}
};

static XtResource resourcesNextGame[] =
{
	/* 18*6+2 */
	{XtNwidth, XtCWidth, XtRDimension, sizeof (Dimension),
	 XtOffset(NextHextrisWidget, core.width),
	 XtRString, (caddr_t) "145"},
	{XtNheight, XtCHeight, XtRDimension, sizeof (Dimension),
	 XtOffset(NextHextrisWidget, core.height),
	 XtRString, (caddr_t) "109"},
	{XtNmono, XtCMono, XtRBoolean, sizeof (Boolean),
	 XtOffset(NextHextrisWidget, hextris.mono),
	 XtRString, (caddr_t) "FALSE"},
	{XtNreverse, XtCReverse, XtRBoolean, sizeof (Boolean),
	 XtOffset(NextHextrisWidget, hextris.reverse),
	 XtRString, (caddr_t) "FALSE"},
	{XtNforeground, XtCForeground, XtRPixel, sizeof (Pixel),
	 XtOffset(NextHextrisWidget, hextris.foreground),
	 XtRString, (caddr_t) XtDefaultForeground},
	{XtNbackground, XtCBackground, XtRPixel, sizeof (Pixel),
	 XtOffset(NextHextrisWidget, hextris.background),
	 XtRString, (caddr_t) XtDefaultBackground},
	{XtNframeColor, XtCColor, XtRPixel, sizeof (Pixel),
	 XtOffset(NextHextrisWidget, hextris.frameColor),
	 XtRString, (caddr_t) "cyan" /*XtDefaultForeground*/},
	{XtNborderColor, XtCColor, XtRPixel, sizeof (Pixel),
	 XtOffset(NextHextrisWidget, hextris.borderColor),
	 XtRString, (caddr_t) "gray25" /*XtDefaultForeground*/},
	{XtNcolumns, XtCColumns, XtRInt, sizeof (int),
	 XtOffset(NextHextrisWidget, hextris.columns),
	 XtRString, (caddr_t) "10"},
	{XtNunits, XtCUnits, XtRInt, sizeof (int),
	 XtOffset(NextHextrisWidget, hextris.units),
	 XtRString, (caddr_t) "4"},
	{XtNcorners, XtCCorners, XtRBoolean, sizeof (Boolean),
	 XtOffset(NextHextrisWidget, hextris.corners),
	 XtRString, (caddr_t) "FALSE"},
	{XtNmixed, XtCMixed, XtRBoolean, sizeof (Boolean),
	 XtOffset(NextHextrisWidget, hextris.mixed),
	 XtRString, (caddr_t) "FALSE"},
	{XtNbonus, XtCBonus, XtRBoolean, sizeof (Boolean),
	 XtOffset(NextHextrisWidget, nextHextris.newBonus),
	 XtRString, (caddr_t) "FALSE"},
	{XtNplainPiece, XtCPlainPiece, XtRBoolean, sizeof (Boolean),
	 XtOffset(NextHextrisWidget, hextris.plainPiece),
	 XtRString, (caddr_t) "FALSE"},
	{XtNshowNext, XtCShowNext, XtRBoolean, sizeof (Boolean),
	 XtOffset(NextHextrisWidget, hextris.showNext),
	 XtRString, (caddr_t) "TRUE"},
	{XtNmenu, XtCMenu, XtRInt, sizeof (int),
	 XtOffset(NextHextrisWidget, hextris.menu),
	 XtRString, (caddr_t) "-1"},
	{XtNstart, XtCBoolean, XtRBoolean, sizeof (Boolean),
	 XtOffset(NextHextrisWidget, hextris.started),
	 XtRString, (caddr_t) "FALSE"},
	{XtNrandomNumber, XtCRandomNumber, XtRInt, sizeof (int),
	 XtOffset(NextHextrisWidget, nextHextris.randomNumber),
	 XtRString, (caddr_t) "-1"},
	{XtNrandomRotation, XtCRandomRotation, XtRInt, sizeof (int),
	 XtOffset(NextHextrisWidget, nextHextris.randomRotation),
	 XtRString, (caddr_t) "-1"},
	{XtNrandomReflection, XtCRandomReflection, XtRInt, sizeof (int),
	 XtOffset(NextHextrisWidget, nextHextris.randomReflection),
	 XtRString, (caddr_t) "-1"},
	{XtNselectCallback, XtCCallback, XtRCallback, sizeof (caddr_t),
	 XtOffset(NextHextrisWidget, hextris.select),
	 XtRCallback, (caddr_t) NULL}
};

NextHextrisClassRec nextHextrisClassRec =
{
	{
		(WidgetClass) & widgetClassRec,		/* superclass */
		(char *) "NextHextris",	/* class name */
		sizeof (NextHextrisRec),	/* widget size */
		NULL,		/* class initialize */
		NULL,		/* class part initialize */
		FALSE,		/* class inited */
		(XtInitProc) InitializeNextGame,	/* initialize */
		NULL,		/* initialize hook */
		XtInheritRealize,	/* realize */
		actionsListNextGame,	/* actions */
		XtNumber(actionsListNextGame),	/* num actions */
		resourcesNextGame,	/* resources */
		XtNumber(resourcesNextGame),	/* num resources */
		NULLQUARK,	/* xrm class */
		TRUE,		/* compress motion */
		TRUE,		/* compress exposure */
		TRUE,		/* compress enterleave */
		TRUE,		/* visible interest */
		(XtWidgetProc) DestroyNextGame,	/* destroy */
		(XtWidgetProc) ResizeNextGame,	/* resize */
		(XtExposeProc) ExposeNextGame,	/* expose */
		(XtSetValuesFunc) SetValuesNextGame,	/* set values */
		NULL,		/* set values hook */
		XtInheritSetValuesAlmost,	/* set values almost */
		NULL,		/* get values hook */
		NULL,		/* accept focus */
		XtVersion,	/* version */
		NULL,		/* callback private */
		defaultTranslationsNextGame,	/* tm table */
		NULL,		/* query geometry */
		NULL,		/* display accelerator */
		NULL		/* extension */
	},
	{
		0		/* ignore */
	}
};

WidgetClass nextHextrisWidgetClass = (WidgetClass) & nextHextrisClassRec;

static void
setNextGame(NextHextrisWidget w, int reason)
{
	hextrisCallbackStruct cb;

	cb.reason = reason;
	XtCallCallbacks((Widget) w, (char *) XtNselectCallback, &cb);
}
#endif

#include "image.h"

static Pixmap dr = 0; /* dummy for future double buffering */

static void
CheckThings(NextHextrisWidget w)
{
	char *buf1 = NULL, *buf2 = NULL;

	if (w->hextris.units < MIN_UNITS || w->hextris.units > MAX_UNITS) {
		intCat(&buf1, "Number of units out of bounds, use ",
			MIN_UNITS);
		stringCat(&buf2, buf1, "...");
		free(buf1);
		intCat(&buf1, buf2, MAX_UNITS);
		free(buf2);
		stringCat(&buf2, buf1, ", defaulting to ");
		free(buf1);
		intCat(&buf1, buf2, DEFAULT_UNITS);
		free(buf2);
		DISPLAY_WARNING(buf1);
		free(buf1);
		w->hextris.units = DEFAULT_UNITS;
		w->hextris.next.units = w->hextris.units;
	}
}

static void
drawUnit(NextHextrisWidget w, int pmid, int cid, int i, int j)
{
	if (i >= 0 && i < MAX_UNITS &&
			j >= 0 && j < MAX_UNITS + 1) {
		hexagon_border[0].x = i * w->hextris.hexWidth +
			w->hextris.offset.x - 1 +
			(((j + 1 + w->hextris.nextThing.size) % 2) ?
			0 : (w->hextris.hexWidth + 1) / 2);
		hexagon_border[0].y = (j - 1) * w->hextris.hexHeight +
			w->hextris.offset.y;
		if (w->hextris.next.corners) {
			fillPolygon((HextrisWidget) w, pmid, cid,
				hexagon_border, True, False);
		} else {
			Point hexagon_borderTrunc[MAX_SIDES + 1];
			Point aPoint, bPoint, aPointTrunc, bPointTrunc;
			int k;

			aPoint.x = aPoint.y = 0;
			bPoint.x = bPoint.y = 0;
			aPointTrunc.x = aPointTrunc.y = 0;
			bPointTrunc.x = -hexagon_border[1].x / 2;
			bPointTrunc.y = hexagon_border[1].y / 2;
			for (k = 0; k < MAX_SIDES; k++) {
				aPoint.x += hexagon_border[k + 1].x;
				aPoint.y += hexagon_border[k + 1].y;
				aPointTrunc.x = (aPoint.x + bPoint.x) / 2;
				aPointTrunc.y = (aPoint.y + bPoint.y) / 2;
				hexagon_borderTrunc[k + 1].x =
					aPointTrunc.x - bPointTrunc.x;
				hexagon_borderTrunc[k + 1].y =
					aPointTrunc.y - bPointTrunc.y;
				bPoint = aPoint;
				bPointTrunc = aPointTrunc;
			}
			hexagon_borderTrunc[0].x = hexagon_border[0].x -
				hexagon_border[1].x / 2;
			hexagon_borderTrunc[0].y = hexagon_border[0].y +
				hexagon_border[1].y / 2 + 1;
			fillPolygon((HextrisWidget) w, pmid, cid,
				hexagon_borderTrunc, False, False);
		}
	}
}

#if 0
static void
clearThing(HextrisWidget w)
{
	XFillRectangle(display, mainWin, w->hextris.inverseGC,
		topMidX,
		topMidY - (w->hextris.hexHeight * HEX_ROW_COL(HEX_SIZE) +
		w->hextris.hexHeight / 3) / 2 + 3,
		topRWidth, w->hextris.hexHeight * HEX_ROW_COL(HEX_SIZE) +
		w->hextris.hexHeight / 3 + 4);
}
#endif

static void
clearThing(NextHextrisWidget w)
{
	FILLRECTANGLE(w, dr, w->hextris.inverseGC,
		w->hextris.offset.x, w->hextris.offset.y,
		w->hextris.viewWidth, w->hextris.viewHeight);
}

/*static void
drawThing(NextHextrisWidget w)
{
	int i, j;

	for (i = 0; i < w->hextris.nextThing.size; i++)
		for (j = 0; j < w->hextris.nextThing.size; j++)
			if (tris[w->hextris.nextThing.units].polyhex
					[w->hextris.nextThing.polyhexNumber].shape[j][i]) {
				drawUnit(w, polyPixmap(i, j,
					(tris[w->hextris.nextThing.units].polyhex
					[w->hextris.nextThing.polyhexNumber].shape),
					w->hextris.nextThing.size),
					w->hextris.nextThing.colorNumber,
					w->hextris.nextThing.xpos + i,
					w->hextris.nextThing.ypos + j);
			}
}*/

static void
drawThing(NextHextrisWidget w)
{
	int i, j;
	int len = HEX_ROW_COL(w->hextris.nextThing.size);

	if (w->hextris.nextThing.polyhexNumber == GAME_IGNORE)
		return;
	clearThing(w);
	for (i = 0; i < len; i++)
		for (j = 0; j < len; j++)
			if (inHexagon(w->hextris.nextThing.size, i, j) != 0 &&
					tris[w->hextris.nextThing.units].polyhex
					[w->hextris.nextThing.polyhexNumber].shape[j][i] != 0) {
				drawUnit(w, polyPixmap(i, j,
					(tris[w->hextris.nextThing.units].polyhex
					[w->hextris.nextThing.polyhexNumber].shape),
					w->hextris.nextThing.size),
					w->hextris.nextThing.colorNumber,
					w->hextris.nextThing.xpos + i,
					w->hextris.nextThing.ypos + j);
			}
}

static void
redoNext(NextHextrisWidget w)
{
	int corners = (w->hextris.next.corners) ? 1 : 0;
	int turnStyle = tris[w->hextris.next.units -
		MIN_UNITS].mode[corners].turnStyle;
	int next_start, i;

	w->hextris.nextThing.units = w->hextris.next.units - MIN_UNITS +
		w->hextris.nextThing.bonus;
	next_start = w->hextris.nextThing.randomNumber %
		tris[w->hextris.nextThing.units].number
		[(w->hextris.nextThing.bonus) ?
		NOMIX : w->hextris.next.mixed][turnStyle];
	w->hextris.nextThing.colorNumber = next_start;

	if (w->hextris.next.mixed && !w->hextris.nextThing.bonus) {
		w->hextris.nextThing.units = 0;
		while (next_start >= tris[w->hextris.nextThing.units].number
				[NOMIX][turnStyle]) {
			next_start -= tris[w->hextris.nextThing.units].number
				[NOMIX][turnStyle];
			w->hextris.nextThing.units++;
		}
	}
	w->hextris.nextThing.polyhexNumber =
		tris[w->hextris.nextThing.units].mode
		[corners].start[next_start][turnStyle];
	if (tris[w->hextris.nextThing.units].mode[corners].turnStyle > NONE)
		for (i = 0; i < w->hextris.nextThing.randomRotation; i++)
			w->hextris.nextThing.polyhexNumber =
				tris[w->hextris.nextThing.units].polyhex
				[w->hextris.nextThing.polyhexNumber].rotation;
	if (tris[w->hextris.nextThing.units].mode[corners].turnStyle == ALL)
		for (i = 0; i < w->hextris.nextThing.randomReflection; i++)
			w->hextris.nextThing.polyhexNumber =
				tris[w->hextris.nextThing.units].polyhex
				[w->hextris.nextThing.polyhexNumber].reflection;
	w->hextris.nextThing.size = tris[w->hextris.nextThing.units].polyhex
		[w->hextris.nextThing.polyhexNumber].size;
	w->hextris.nextThing.xpos =
		(MAX_UNITS - w->hextris.nextThing.size) / 2;
	w->hextris.nextThing.ypos =
		(MAX_UNITS - w->hextris.nextThing.size) / 2;
}

static void
EraseFrame(NextHextrisWidget w)
{
	FILLRECTANGLE(w, dr, w->hextris.inverseGC,
		w->hextris.offset.x, w->hextris.offset.y,
		w->hextris.viewWidth, w->hextris.viewHeight);
}

static void
DrawFrame(NextHextrisWidget w, Boolean focus)
{
	DRAWRECTANGLE(w, dr,
		(focus) ? w->hextris.frameGC : w->hextris.borderGC,
		w->hextris.offset.x - 1, w->hextris.offset.y - 1,
		w->hextris.viewWidth + 1, w->hextris.viewHeight + 1);
}

#if 0
static void
ResetThings(NextHextrisWidget w)
{
}

static void
ResizeThings(NextHextrisWidget w)
{
}
#endif

#ifdef WINVER
static void
SetValuesNextGame(NextHextrisWidget w)
{
	struct tagColor {
		int red, green, blue;
	} color;
	char szBuf[80], buf[20];
	int thing;

	w->hextris.units = GetPrivateProfileInt(SECTION, "units",
		DEFAULTUNITS, INIFILE);
	w->hextris.corners = (BOOL) GetPrivateProfileInt(SECTION, "corners",
		DEFAULTCORNERS, INIFILE);
	w->hextris.mixed = (BOOL) GetPrivateProfileInt(SECTION, "mixed",
		DEFAULTMIXED, INIFILE);
	w->hextris.plainPiece = (BOOL) GetPrivateProfileInt(SECTION, "plainPiece",
		FALSE, INIFILE);
	w->hextris.showNext = (BOOL) GetPrivateProfileInt(SECTION, "showNext",
		FALSE, INIFILE);
	w->hextris.mono = (BOOL) GetPrivateProfileInt(SECTION, "mono",
		DEFAULTMONO, INIFILE);
	w->hextris.reverse = (BOOL) GetPrivateProfileInt(SECTION, "reverse",
		DEFAULTREVERSE, INIFILE);
	/* cyan */
	(void) GetPrivateProfileString(SECTION, "frameColor", "0 255 255",
		szBuf, sizeof (szBuf), INIFILE);
	(void) sscanf(szBuf, "%d %d %d",
		&(color.red), &(color.green), &(color.blue));
	w->hextris.frameGC = RGB(color.red, color.green, color.blue);
	/* gray25 */
	(void) GetPrivateProfileString(SECTION, "borderColor", "64 64 64",
		szBuf, sizeof (szBuf), INIFILE);
	(void) sscanf(szBuf, "%d %d %d",
		&(color.red), &(color.green), &(color.blue));
	w->hextris.borderGC = RGB(color.red, color.green, color.blue);
	/* #AEB2C3 */
	(void) GetPrivateProfileString(SECTION, "background", "255 255 255",
		szBuf, sizeof (szBuf), INIFILE);
	(void) sscanf(szBuf, "%d %d %d",
		&(color.red), &(color.green), &(color.blue));
	w->hextris.inverseGC = RGB(color.red, color.green, color.blue);
	for (thing = 0; thing < MAX_START_POLYHEXES; thing++) {
		(void) sprintf(buf, "thingColor%d", thing);
		(void) GetPrivateProfileString(SECTION, buf,
			polyhexColorString[thing],
			szBuf, sizeof (szBuf), INIFILE);
		(void) sscanf(szBuf, "%d %d %d",
			&(color.red), &(color.green), &(color.blue));
		w->hextris.thingGC[thing] =
			RGB(color.red, color.green, color.blue);
		w->hextris.thingBrightGC[thing] =
			RGB(brighter(color.red), brighter(color.green),
			brighter(color.blue));
		w->hextris.thingDarkGC[thing] =
			RGB(darker(color.red), darker(color.green),
			darker(color.blue));
		w->hextris.thingBrighterGC[thing] =
			RGB(brighter(brighter(color.red)),
			brighter(brighter(color.green)),
			brighter(brighter(color.blue)));
		w->hextris.thingDarkerGC[thing] =
			RGB(darker(darker(color.red)),
			darker(darker(color.green)),
			darker(darker(color.blue)));
	}
}

void
DestroyNextGame(HBRUSH brush)
{
	(void) DeleteObject(brush);
	/*PostQuitMessage(0);*/
}

#else
static void
GetColor(NextHextrisWidget w, int thing)
{
	XGCValues values;
	XtGCMask valueMask;
	XColor colorCell, rgb;
	Boolean gotColor = False;

	valueMask = GCForeground | GCBackground;
	if (w->hextris.reverse) {
		values.background = w->hextris.foreground;
	} else {
		values.background = w->hextris.background;
	}
	if (!w->hextris.mono) {
		if (XAllocNamedColor(XtDisplay(w),
				DefaultColormapOfScreen(XtScreen(w)),
				polyhexColorName[thing],
				&colorCell, &rgb)) {
			values.foreground = w->hextris.thingColor[thing] =
				colorCell.pixel;
			gotColor = True;
		} else {
			char *buf1, *buf2;

			stringCat(&buf1, "Color name \"",
				polyhexColorName[thing]);
			stringCat(&buf2, buf1, "\" is not defined for polyhex ");
			free(buf1);
			intCat(&buf1, buf2, thing);
			free(buf2);
			DISPLAY_WARNING(buf1);
			free(buf1);
		}
	}
	if (!gotColor) {
		if (w->hextris.reverse) {
			values.background = w->hextris.foreground;
			values.foreground = w->hextris.background;
		} else {
			values.background = w->hextris.background;
			values.foreground = w->hextris.foreground;
		}
	}
	if (w->hextris.thingGC[thing])
		XtReleaseGC((Widget) w, w->hextris.thingGC[thing]);
	w->hextris.thingGC[thing] = XtGetGC((Widget) w, valueMask, &values);
	gotColor = False;
	if (!w->hextris.mono) {
		if (XAllocNamedColor(XtDisplay(w),
				DefaultColormapOfScreen(XtScreen(w)),
				polyhexColorName[thing],
				&colorCell, &rgb)) {
			values.foreground = w->hextris.thingColor[thing] =
				colorCell.pixel;
			gotColor = True;
		} else {
			char *buf1, *buf2;

			stringCat(&buf1, "Color name \"",
				polyhexColorName[thing]);
			stringCat(&buf2, buf1, "\" is not defined for polyhex ");
			free(buf1);
			intCat(&buf1, buf2, thing);
			free(buf2);
			DISPLAY_WARNING(buf1);
			free(buf1);
		}
	}
	if (!gotColor) {
		if (w->hextris.reverse) {
			values.background = w->hextris.foreground;
			values.foreground = w->hextris.background;
		} else {
			values.background = w->hextris.background;
			values.foreground = w->hextris.foreground;
		}
	}
	if (w->hextris.thingGC[thing])
		XtReleaseGC((Widget) w, w->hextris.thingGC[thing]);
	w->hextris.thingGC[thing] = XtGetGC((Widget) w, valueMask, &values);
	gotColor = False;
	if (!w->hextris.mono) {
		if (XAllocNamedColor(XtDisplay(w),
				DefaultColormapOfScreen(XtScreen(w)),
				polyhexColorName[thing],
				&colorCell, &rgb)) {
			values.foreground = w->hextris.thingColor[thing] =
				brighter((HextrisWidget) w, colorCell.pixel);
			gotColor = True;
		} else {
			char *buf1, *buf2;

			stringCat(&buf1, "Color name \"",
				polyhexColorName[thing]);
			stringCat(&buf2, buf1, "\" is not defined for polyhex ");
			free(buf1);
			intCat(&buf1, buf2, thing);
			free(buf2);
			DISPLAY_WARNING(buf1);
			free(buf1);
		}
	}
	if (!gotColor) {
		if (w->hextris.reverse) {
			values.background = w->hextris.foreground;
			values.foreground = w->hextris.background;
		} else {
			values.background = w->hextris.background;
			values.foreground = w->hextris.foreground;
		}
	}
	if (w->hextris.thingBrightGC[thing])
		XtReleaseGC((Widget) w, w->hextris.thingBrightGC[thing]);
	w->hextris.thingBrightGC[thing] = XtGetGC((Widget) w, valueMask,
		&values);
	gotColor = False;
	if (!w->hextris.mono) {
		if (XAllocNamedColor(XtDisplay(w),
				DefaultColormapOfScreen(XtScreen(w)),
				polyhexColorName[thing],
				&colorCell, &rgb)) {
			values.foreground = w->hextris.thingColor[thing] =
				darker((HextrisWidget) w, colorCell.pixel);
			gotColor = True;
		} else {
			char *buf1, *buf2;

			stringCat(&buf1, "Color name \"",
				polyhexColorName[thing]);
			stringCat(&buf2, buf1, "\" is not defined for polyhex ");
			free(buf1);
			intCat(&buf1, buf2, thing);
			free(buf2);
			DISPLAY_WARNING(buf1);
			free(buf1);
		}
	}
	if (!gotColor) {
		if (w->hextris.reverse) {
			values.background = w->hextris.foreground;
			values.foreground = w->hextris.background;
		} else {
			values.background = w->hextris.background;
			values.foreground = w->hextris.foreground;
		}
	}
	if (w->hextris.thingDarkGC[thing])
		XtReleaseGC((Widget) w, w->hextris.thingDarkGC[thing]);
	w->hextris.thingDarkGC[thing] = XtGetGC((Widget) w, valueMask,
		&values);
	gotColor = False;
	if (!w->hextris.mono) {
		if (XAllocNamedColor(XtDisplay(w),
				DefaultColormapOfScreen(XtScreen(w)),
				polyhexColorName[thing],
				&colorCell, &rgb)) {
			values.foreground = w->hextris.thingColor[thing] =
				brighter((HextrisWidget) w,
				brighter((HextrisWidget) w, colorCell.pixel));
			gotColor = True;
		} else {
			char *buf1, *buf2;

			stringCat(&buf1, "Color name \"",
				polyhexColorName[thing]);
			stringCat(&buf2, buf1, "\" is not defined for polyhex ");
			free(buf1);
			intCat(&buf1, buf2, thing);
			free(buf2);
			DISPLAY_WARNING(buf1);
			free(buf1);
		}
	}
	if (!gotColor) {
		if (w->hextris.reverse) {
			values.background = w->hextris.foreground;
			values.foreground = w->hextris.background;
		} else {
			values.background = w->hextris.background;
			values.foreground = w->hextris.foreground;
		}
	}
	if (w->hextris.thingBrighterGC[thing])
		XtReleaseGC((Widget) w, w->hextris.thingBrighterGC[thing]);
	w->hextris.thingBrighterGC[thing] = XtGetGC((Widget) w, valueMask,
		&values);
	gotColor = False;
	if (!w->hextris.mono) {
		if (XAllocNamedColor(XtDisplay(w),
				DefaultColormapOfScreen(XtScreen(w)),
				polyhexColorName[thing],
				&colorCell, &rgb)) {
			values.foreground = w->hextris.thingColor[thing] =
				darker((HextrisWidget) w,
				darker((HextrisWidget) w, colorCell.pixel));
			gotColor = True;
		} else {
			char *buf1, *buf2;

			stringCat(&buf1, "Color name \"",
				polyhexColorName[thing]);
			stringCat(&buf2, buf1, "\" is not defined for polyhex ");
			free(buf1);
			intCat(&buf1, buf2, thing);
			free(buf2);
			DISPLAY_WARNING(buf1);
			free(buf1);
		}
	}
	if (!gotColor) {
		if (w->hextris.reverse) {
			values.background = w->hextris.foreground;
			values.foreground = w->hextris.background;
		} else {
			values.background = w->hextris.background;
			values.foreground = w->hextris.foreground;
		}
	}
	if (w->hextris.thingDarkerGC[thing])
		XtReleaseGC((Widget) w, w->hextris.thingDarkerGC[thing]);
	w->hextris.thingDarkerGC[thing] = XtGetGC((Widget) w, valueMask,
		&values);
}

static void
SetAllColors(NextHextrisWidget w)
{
	XGCValues values;
	XtGCMask valueMask;
	int polyhex;

	valueMask = GCForeground | GCBackground;

	if (w->hextris.reverse) {
		values.background = w->hextris.background;
		values.foreground = w->hextris.foreground;
	} else {
		values.foreground = w->hextris.background;
		values.background = w->hextris.foreground;
	}
	if (w->hextris.inverseGC)
		XtReleaseGC((Widget) w, w->hextris.inverseGC);
	w->hextris.inverseGC = XtGetGC((Widget) w, valueMask, &values);
	if (w->hextris.mono) {
		if (w->hextris.reverse) {
			values.background = w->hextris.foreground;
			values.foreground = w->hextris.background;
		} else {
			values.foreground = w->hextris.foreground;
			values.background = w->hextris.background;
		}
	} else {
		values.foreground = w->hextris.frameColor;
		values.background = w->hextris.background;
	}
	if (w->hextris.frameGC)
		XtReleaseGC((Widget) w, w->hextris.frameGC);
	w->hextris.frameGC = XtGetGC((Widget) w, valueMask, &values);
	if (w->hextris.mono) {
		if (w->hextris.reverse) {
			values.background = w->hextris.foreground;
			values.foreground = w->hextris.background;
		} else {
			values.foreground = w->hextris.foreground;
			values.background = w->hextris.background;
		}
	} else {
		values.foreground = w->hextris.borderColor;
		values.background = w->hextris.background;
	}
	if (w->hextris.borderGC)
		XtReleaseGC((Widget) w, w->hextris.borderGC);
	w->hextris.borderGC = XtGetGC((Widget) w, valueMask, &values);
	for (polyhex = 0; polyhex < MAX_START_POLYHEXES; polyhex++)
		GetColor(w, polyhex);
}

static Boolean
SetValuesNextGame(Widget current, Widget request, Widget renew)
{
	NextHextrisWidget c = (NextHextrisWidget) current;
	NextHextrisWidget w = (NextHextrisWidget) renew;
	Boolean redraw = False, setColors = False;

	CheckThings(w);
	if (w->hextris.showNext != c->hextris.showNext) {
		redraw = True;
	}
	if (w->hextris.borderColor != c->hextris.borderColor ||
			w->hextris.frameColor != c->hextris.frameColor ||
			w->hextris.reverse != c->hextris.reverse ||
			w->hextris.mono != c->hextris.mono) {
		SetAllColors(w);
		redraw = True;
	} else if (w->hextris.background != c->hextris.background ||
			w->hextris.foreground != c->hextris.foreground ||
			setColors) {
		SetAllColors(w);
		redraw = True;
	} else if (w->hextris.plainPiece != c->hextris.plainPiece) {
		redraw = True;
	}
#if 0
	if (w->hextris.units != c->hextris.units) {
		w->hextris.next.units = w->hextris.units;
		ResetThings(w);
	} else if (w->hextris.corners != c->hextris.corners) {
		w->hextris.next.corners = w->hextris.corners;
		ResetThings(w);
	} else if (w->hextris.mixed != c->hextris.mixed) {
		w->hextris.next.mixed = w->hextris.mixed;
		ResetThings(w);
	}
#endif
	if (w->hextris.menu != -1) {
		switch (w->hextris.menu) {
		case 3:
			w->hextris.menu = -1;
			clearThing(w);
			redraw = True;
			break;
		default:
			w->hextris.menu = -1;
			break;
		}
	}
	if (w->nextHextris.randomNumber == GAME_CLEARNEXT) {
		if (w->hextris.showNext) {
			drawThing(w);
		} else {
			clearThing(w);
		}
		w->nextHextris.randomNumber = GAME_IGNORE;
		redraw = True;
	} else if (w->nextHextris.randomNumber != GAME_IGNORE) {
		w->hextris.nextThing.randomNumber = w->nextHextris.randomNumber;
		w->hextris.nextThing.randomRotation = w->nextHextris.randomRotation;
		w->hextris.nextThing.randomReflection = w->nextHextris.randomReflection;
		w->hextris.nextThing.bonus = w->nextHextris.newBonus;
		w->hextris.next.units = w->hextris.units;
		w->hextris.next.corners = w->hextris.corners;
		w->hextris.next.mixed = w->hextris.mixed;
		redoNext(w);
		if (w->hextris.showNext) {
			drawThing(w);
		}
		w->nextHextris.randomNumber = GAME_IGNORE;
		redraw = True;
	}
	return (redraw);
}

static void
QuitNextGame(NextHextrisWidget w, XEvent *event, char **args, int nArgs)
{
	Display *display = XtDisplay(w);

	XtCloseDisplay(display);
	exit(0);
}

static void
DestroyNextGame(Widget old)
{
	NextHextrisWidget w = (NextHextrisWidget) old;
	int thing;

	for (thing = 0; thing < MAX_START_POLYHEXES; thing++) {
		XtReleaseGC(old, w->hextris.thingGC[thing]);
		XtReleaseGC(old, w->hextris.thingBrightGC[thing]);
		XtReleaseGC(old, w->hextris.thingDarkGC[thing]);
		XtReleaseGC(old, w->hextris.thingBrighterGC[thing]);
		XtReleaseGC(old, w->hextris.thingDarkerGC[thing]);
	}
	XtReleaseGC(old, w->hextris.borderGC);
	XtReleaseGC(old, w->hextris.frameGC);
	XtReleaseGC(old, w->hextris.inverseGC);
	XtRemoveCallbacks(old, XtNselectCallback, w->hextris.select);
}
#endif

#define SQRT2_2 0.70710678 /* sqrt(2) / 2 */

#ifndef WINVER
static
#endif
void
ResizeNextGame(NextHextrisWidget w)
{
	int i;
#ifdef WINVER
	RECT rect;

	/* Determine size of client area */
	(void) GetClientRect(w->core.hWnd, &rect);
	w->core.width = rect.right;
	w->core.height = rect.bottom;
#endif
	w->hextris.hexHeight = MIN(3 * (w->core.height - 2) /
		(3 * MAX_UNITS + 1),
		(w->core.width - 2) * SQRT2_2 / MAX_UNITS);
	w->hextris.hexWidth = (int) (w->hextris.hexHeight / SQRT2_2);
	w->hextris.viewWidth = w->hextris.hexWidth *
		MAX_UNITS + 2;
	w->hextris.viewHeight = w->hextris.hexHeight *
		 (3 * MAX_UNITS + 1) / 3 + 1;
	w->hextris.offset.x = ((int) w->core.width -
		w->hextris.viewWidth) / 2;
	w->hextris.offset.y = ((int) w->core.height -
		w->hextris.viewHeight) / 2;

	for (i = 0; i < MAX_SIDES; i++) {
		hexagon_list[CORNERS][i].x =
			(w->hextris.hexWidth - 2) / 2 *
			hexagon_unit[CORNERS][i].x;
		hexagon_list[CORNERS][i].y =
			(2 * w->hextris.hexHeight / 3 - 1) / 2 *
			hexagon_unit[CORNERS][i].y;
		hexagon_list[NOCORN][i].x =
			(w->hextris.hexWidth - 1) / 4 *
			hexagon_unit[NOCORN][i].x;
		hexagon_list[NOCORN][i].y =
			3 * (2 * w->hextris.hexHeight / 3 - 1) / 4 *
			hexagon_unit[NOCORN][i].y;
		hexagon_border[i].x =
			w->hextris.hexWidth / 2 * hexagon_unit[CORNERS][i].x;
		hexagon_border[i].y =
			w->hextris.hexHeight / 3 * hexagon_unit[CORNERS][i].y;
	}
	hexagon_list[CORNERS][MAX_SIDES].x =
		-hexagon_list[CORNERS][MAX_SIDES / 2].x;
	hexagon_list[CORNERS][MAX_SIDES].y =
		-hexagon_list[CORNERS][MAX_SIDES / 2].y;
	hexagon_list[NOCORN][MAX_SIDES].x =
		-hexagon_list[NOCORN][MAX_SIDES / 2].x;
	hexagon_list[NOCORN][MAX_SIDES].y =
		-hexagon_list[NOCORN][MAX_SIDES / 2].y;
	hexagon_border[MAX_SIDES].x = -hexagon_border[MAX_SIDES / 2].x;
	hexagon_border[MAX_SIDES].y = -hexagon_border[MAX_SIDES / 2].y;
}

#if 0
#ifndef WINVER
static
#endif
void
SizeNextGame(NextHextrisWidget w)
{
	ResetThings(w);
	ResizeNextGame(w);
}
#endif

#ifndef WINVER
static
#endif
void
InitializeNextGame(
#ifdef WINVER
NextHextrisWidget w, HBRUSH brush
#else
Widget request, Widget renew
#endif
)
{
	int corners, mixed, turnStyle;
#ifndef WINVER
	int poly;
	NextHextrisWidget w = (NextHextrisWidget) renew;
#endif

#ifdef WINVER
	SetValuesNextGame(w);
#else
	w->hextris.mono = (DefaultDepthOfScreen(XtScreen(w)) < 2 ||
		w->hextris.mono);
	for (poly = 0; poly < MAX_START_POLYHEXES; poly++) {
		w->hextris.thingGC[poly] = NULL;
		w->hextris.thingBrightGC[poly] = NULL;
		w->hextris.thingDarkGC[poly] = NULL;
		w->hextris.thingBrighterGC[poly] = NULL;
		w->hextris.thingDarkerGC[poly] = NULL;
	}
	w->hextris.borderGC = NULL;
	w->hextris.frameGC = NULL;
	w->hextris.inverseGC = NULL;
#endif

	w->hextris.quit = False;
	w->hextris.focus = False;
	w->hextris.next.units = w->hextris.units;
	w->hextris.next.corners = w->hextris.corners;
	w->hextris.next.mixed = w->hextris.mixed;
	w->hextris.nextThing.polyhexNumber = GAME_IGNORE;
	CheckThings(w);
#if 0
	SizeNextGame(w);
#else
	ResizeNextGame(w);
#endif
#ifdef WINVER
	brush = CreateSolidBrush(w->hextris.inverseGC);
	SETBACK(w->core.hWnd, brush);
	(void) SRAND(time(NULL));
#else
	(void) SRAND(getpid());
	SetAllColors(w);
#endif
	corners = (w->hextris.next.corners) ? 1 : 0;
	mixed = (w->hextris.next.mixed) ? 1 : 0;
	readThings();
	turnStyle = tris[w->hextris.next.units -
		MIN_UNITS].mode[corners].turnStyle;
	/* use intCat */
	if (turnStyle < 0 || turnStyle >= MAX_TURNSTYLE_TYPES) {
		(void) fprintf(stderr, "NextHextris: corrupted polyhex file\n");
		(void) fprintf(stderr, "\tturnStyle = %d.\n", turnStyle);
		exit(1);
	}
	w->hextris.numberPolyhexes = tris[w->hextris.next.units -
		MIN_UNITS].number[mixed][turnStyle];
	if (w->hextris.numberPolyhexes <= 0 ||
			w->hextris.numberPolyhexes > MAX_START_POLYHEXES) {
		(void) fprintf(stderr, "NextHextris: corrupted polyhex file\n");
		(void) fprintf(stderr, "\tnumberPolyhexes = %d.\n",
			w->hextris.numberPolyhexes);
		exit(1);
	}
}

#ifndef WINVER
static
#endif
void
ExposeNextGame(
#ifdef WINVER
NextHextrisWidget w
#else
Widget renew, XEvent *event, Region region
#endif
)
{
#ifndef WINVER
	NextHextrisWidget w = (NextHextrisWidget) renew;

	if (!w->core.visible)
		return;
#endif
	EraseFrame(w);
#if 0
	ResizeThings(w);
#endif
	DrawFrame(w, w->hextris.focus);
	if (w->hextris.showNext) {
		drawThing(w);
	}
}

#ifndef WINVER
static
#endif
void
HideNextGame(NextHextrisWidget w
#ifndef WINVER
, XEvent *event, char **args, int nArgs
#endif
)
{
	setNextGame(w, GAME_HIDE);
}

#ifdef WINVER
void
ValueNextGame(NextHextrisWidget w,
		int randomNumber, int randomRotation, int randomReflection,
		int units, Boolean corners, Boolean mixed, Boolean newBonus)
{
	w->hextris.nextThing.randomNumber = randomNumber;
	w->hextris.nextThing.randomRotation = randomRotation;
	w->hextris.nextThing.randomReflection = randomReflection;
	w->hextris.next.units = w->hextris.units = units;
	w->hextris.next.corners = w->hextris.corners = corners;
	w->hextris.next.mixed = w->hextris.mixed = mixed;
	w->hextris.nextThing.bonus = newBonus;
	redoNext(w);
	if (w->hextris.showNext) {
		drawThing(w);

	}
}
#endif

#ifndef WINVER
static
#endif
void
EnterNextGame(NextHextrisWidget w
#ifndef WINVER
, XEvent *event, char **args, int nArgs
#endif
)
{
	w->hextris.focus = True;
	DrawFrame(w, w->hextris.focus);
}

#ifndef WINVER
static
#endif
void
LeaveNextGame(NextHextrisWidget w
#ifndef WINVER
, XEvent *event, char **args, int nArgs
#endif
)
{
	w->hextris.focus = False;
	DrawFrame(w, w->hextris.focus);
}

#ifndef WINVER
static
#endif
void
ShowNextNextGame(NextHextrisWidget w
#ifndef WINVER
, XEvent *event, char **args, int nArgs
#endif
)
{
	if (!w->hextris.paused) {
		w->hextris.showNext = !w->hextris.showNext;
		if (w->hextris.showNext) {
			drawThing(w);
		} else {
			clearThing(w);
		}
		FLUSH(w);
	}
}
