/*
# X-BASED HEXTRIS
#
#  HextrisP.h
#
###
#
#  Copyright (c) 1993 - 2006		David A. Bagley, bagleyd@tux.org
#
#  Taken from GENERIC X-BASED TETRIS
#
#  Copyright (c) 1992 - 1995		Q. Alex Zhao, azhao@cc.gatech.edu
#
#			All Rights Reserved
#
#  Permission to use, copy, modify, and distribute this software and
#  its documentation for any purpose and without fee is hereby granted,
#  provided that the above copyright notice appear in all copies and
#  that both that copyright notice and this permission notice appear in
#  supporting documentation, and that the name of the author not be
#  used in advertising or publicity pertaining to distribution of the
#  software without specific, written prior permission.
#
#  This program is distributed in the hope that it will be "playable",
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
*/

/* Private header file for Hextris */

#ifndef _HextrisP_h
#define _HextrisP_h

#include "xwin.h"
#include "timer.h"
#include "polyhex.h"
#include "Hextris.h"

#ifdef WINVER
extern void DestroyGame(HBRUSH brush);
extern void ResizeGame(HextrisWidget w);
extern void SizeGame(HextrisWidget w);
extern void InitializeGame(HextrisWidget w, HBRUSH brush);
extern void ExposeGame(HextrisWidget w);
extern void HideGame(HextrisWidget w);
extern void ClearGame(HextrisWidget w);
extern void EnterGame(HextrisWidget w);
extern void LeaveGame(HextrisWidget w);
extern void MoveGameInput(HextrisWidget w, int direction);
extern void SoundGame(HextrisWidget w);
extern Boolean IncrementGameLevel(HextrisWidget w);
extern Boolean DecrementGameLevel(HextrisWidget w);
extern Boolean IncrementGameUnits(HextrisWidget w);
extern Boolean DecrementGameUnits(HextrisWidget w);
extern Boolean IncrementGameColumns(HextrisWidget w);
extern Boolean DecrementGameColumns(HextrisWidget w);
extern void FixedFillGame(HextrisWidget w);
extern void SuddenAppearGame(HextrisWidget w);
extern Boolean CornersGame(HextrisWidget w);
extern Boolean MixedGame(HextrisWidget w);
extern void BonusGame(HextrisWidget w);
extern void CwGame(HextrisWidget w);
extern void PlainPieceGame(HextrisWidget w);
extern void FalldownGame(HextrisWidget w);
extern void PauseGame(HextrisWidget w);
extern void ShowNextGame(HextrisWidget w);
extern void GridGame(HextrisWidget w);
extern void Fill0Game(HextrisWidget w);
extern void Fill1Game(HextrisWidget w);
extern void Fill2Game(HextrisWidget w);
extern void Fill3Game(HextrisWidget w);
extern void Fill4Game(HextrisWidget w);
extern void Fill5Game(HextrisWidget w);
extern void Fill6Game(HextrisWidget w);
extern void Fill7Game(HextrisWidget w);
extern void Fill8Game(HextrisWidget w);
extern void Fill9Game(HextrisWidget w);
extern void Fill10Game(HextrisWidget w);
extern void setGame(HextrisWidget w, int reason);
extern void setGameNext(HextrisWidget w, int reason,
	int randomNumber, int randomRotation, int randomReflection,
	int units, Boolean corners, Boolean mixed, Boolean newBonus);

#else

/* This gets around C's inability to do inheritance */
typedef struct _HextrisClassPart {
	int         ignore;
} HextrisClassPart;

typedef struct _HextrisClassRec {
	CoreClassPart coreClass;
	HextrisClassPart hextrisClass;
} HextrisClassRec;

extern HextrisClassRec hextrisClassRec;
#endif

#ifndef BUMPSOUND
#define BUMPSOUND "bump"
#endif

#define SYMBOL ':'

#define IGNORE_DIR (-1)

#define NUM_FLASHES 4

#define MAX_SIDES 6
#define MAX_ORIENT 2
#define MIN_LEVELS 1
#define MAX_LEVELS 20

#define XPOS(s,y,j,i)   (((y+i+s)%2)?((s%2)*2-1)*(j%2):0)
#define XINC(y,d,i)     (((((y+!i)%2)&&(d!=1))||(((y+i)%2)&&(d==1)))*d)

#define THRESHOLD(x)	((x)*10)

/*** types ***/

typedef enum {FALL, DROP, MOVE_LEFT, MOVE_RIGHT, ROTATE, REFLECT} move_t;

typedef struct {
	int         units;
	Boolean     corners, mixed, bonus;
} Spec;

typedef struct {
	int         units, polyhexNumber;
	int         xpos, ypos;
	int         size, colorNumber, direction;
	int         bonus, randomRotation, randomReflection;
	long        randomNumber;
} Thing;

typedef struct {
	int         pmid, cid;
} Field;

typedef struct _HextrisPart {
	Thing       curThing, nextThing;
	Field      *field;
	GC          thingGC[MAX_START_POLYHEXES];
	GC          thingBrightGC[MAX_START_POLYHEXES];
	GC          thingDarkGC[MAX_START_POLYHEXES];
	GC          thingBrighterGC[MAX_START_POLYHEXES];
	GC          thingDarkerGC[MAX_START_POLYHEXES];
	Boolean     fixedFill, suddenAppear, cw, plainPiece, showNext, grid;
	Boolean     invert, sound, resetTime, firstFall, paused, gotNew;
	int         level, fill, completedRows;
	int         startLevel, score, newRows;
	Spec        current, next;
	int         units, columns, rows;
	Boolean     corners, mixed, bonus;
	int         delay, numberPolyhexes, pauseStart;
	int         currentPosition;
	Boolean     fall, started, quit, act;
	Boolean     mono, reverse, focus;
	Boolean     scoreOnly, versionOnly;
	int         hexWidth, hexHeight;
	Position    viewWidth, viewHeight;
	Point       offset;
	GC          frameGC, borderGC, inverseGC, xorGC;
#ifdef WINVER
	char        userName[81], scoreFile[81], bumpSound[81];
#else
	int         menu;
	char       *userName, *scoreFile, *bumpSound;
	Pixel       foreground, background;
	Pixel       frameColor, borderColor;
	Pixel       thingColor[MAX_START_POLYHEXES];
	XtCallbackList select;
#endif
} HextrisPart;

typedef struct _HextrisRec {
	CorePart    core;
	HextrisPart hextris;
} HextrisRec;

#ifdef WINVER
extern int brighter(const int light);
extern int darker(const int light);
#else
extern Pixel brighter(HextrisWidget w, Pixel pixel);
extern Pixel darker(HextrisWidget w, Pixel pixel);
#endif
extern void fillPolygon(HextrisWidget w, int pm, int color, Point *points,
	Boolean corners, Boolean graphicalXor);
#endif /* _HextrisP_h */
