/*
# X-BASED TERTRIS
#
#  TertrisN.c
#
###
#
#  Copyright (c) 1993 - 2006	David A. Bagley, bagleyd@tux.org
#
#			All Rights Reserved
#
#  Permission to use, copy, modify, and distribute this software and
#  its documentation for any purpose and without fee is hereby granted,
#  provided that the above copyright notice appear in all copies and
#  that both that copyright notice and this permission notice appear in
#  supporting documentation, and that the name of the author not be
#  used in advertising or publicity pertaining to distribution of the
#  software without specific, written prior permission.
#
#  This program is distributed in the hope that it will be "playable",
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
*/

/* Methods file for Tertris */
#include "file.h"
#include "rngs.h"
#include "TertrisNP.h"

static Point triangle_unit[MAX_ORIENT][3] = {
	{ {0, 0}, {-1, -1}, { 2, 0} },
	{ {0, 0}, { 1,  1}, {-2, 0} }
};
static Point triangle_list[MAX_ORIENT][4];
static Point triangle_border[MAX_ORIENT][4];

#ifdef WINVER
#ifndef LOGPATH
#define LOGPATH "/usr/tmp"
#endif
#ifndef INIFILE
#define INIFILE "wtertris.ini"
#endif

#define SECTION "setup"

#else

#ifndef LOGPATH
#ifdef VMS
#define LOGPATH "SYS$SCRATCH:"
#else
#define LOGPATH "/usr/tmp"
#endif
#endif

static void ResizeNextGame(NextTertrisWidget w);
static void DestroyNextGame(Widget old);
#if 0
static void SizeNextGame(NextTertrisWidget w);
#endif
static Boolean SetValuesNextGame(Widget current, Widget request, Widget renew);
static void QuitNextGame(NextTertrisWidget w,
	XEvent *event, char **args, int nArgs);
static void InitializeNextGame(Widget request, Widget renew);
static void ExposeNextGame(Widget renew, XEvent *event, Region region);
static void HideNextGame(NextTertrisWidget w,
	XEvent *event, char **args, int nArgs);
static void EnterNextGame(NextTertrisWidget w,
	XEvent *event, char **args, int nArgs);
static void LeaveNextGame(NextTertrisWidget w,
	XEvent *event, char **args, int nArgs);
static void ShowNextNextGame(NextTertrisWidget w,
	XEvent *event, char **args, int nArgs);

static char defaultTranslationsNextGame[] =
"<KeyPress>q: Quit()\n\
 Ctrl<KeyPress>C: Quit()\n\
 <KeyPress>osfCancel: Hide()\n\
 <KeyPress>Escape: Hide()\n\
 <KeyPress>osfEscape: Hide()\n\
 Ctrl<KeyPress>[: Hide()\n\
 <KeyPress>0x1B: Hide()\n\
 <KeyPress>n: ShowNext()\n\
 <EnterWindow>: Enter()\n\
 <LeaveWindow>: Leave()";

static XtActionsRec actionsListNextGame[] =
{
	{(char *) "Quit", (XtActionProc) QuitNextGame},
	{(char *) "Hide", (XtActionProc) HideNextGame},
	{(char *) "ShowNext", (XtActionProc) ShowNextNextGame},
	{(char *) "Enter", (XtActionProc) EnterNextGame},
	{(char *) "Leave", (XtActionProc) LeaveNextGame}
};

static XtResource resourcesNextGame[] =
{
	/* 18*6+2 */
	{XtNwidth, XtCWidth, XtRDimension, sizeof (Dimension),
	 XtOffset(NextTertrisWidget, core.width),
	 XtRString, (caddr_t) "132"},
	{XtNheight, XtCHeight, XtRDimension, sizeof (Dimension),
	 XtOffset(NextTertrisWidget, core.height),
	 XtRString, (caddr_t) "137"},
	{XtNmono, XtCMono, XtRBoolean, sizeof (Boolean),
	 XtOffset(NextTertrisWidget, tertris.mono),
	 XtRString, (caddr_t) "FALSE"},
	{XtNreverse, XtCReverse, XtRBoolean, sizeof (Boolean),
	 XtOffset(NextTertrisWidget, tertris.reverse),
	 XtRString, (caddr_t) "FALSE"},
	{XtNforeground, XtCForeground, XtRPixel, sizeof (Pixel),
	 XtOffset(NextTertrisWidget, tertris.foreground),
	 XtRString, (caddr_t) XtDefaultForeground},
	{XtNbackground, XtCBackground, XtRPixel, sizeof (Pixel),
	 XtOffset(NextTertrisWidget, tertris.background),
	 XtRString, (caddr_t) XtDefaultBackground},
	{XtNframeColor, XtCColor, XtRPixel, sizeof (Pixel),
	 XtOffset(NextTertrisWidget, tertris.frameColor),
	 XtRString, (caddr_t) "cyan" /*XtDefaultForeground*/},
	{XtNborderColor, XtCColor, XtRPixel, sizeof (Pixel),
	 XtOffset(NextTertrisWidget, tertris.borderColor),
	 XtRString, (caddr_t) "gray25" /*XtDefaultForeground*/},
	{XtNcolumns, XtCColumns, XtRInt, sizeof (int),
	 XtOffset(NextTertrisWidget, tertris.columns),
	 XtRString, (caddr_t) "10"},
	{XtNunits, XtCUnits, XtRInt, sizeof (int),
	 XtOffset(NextTertrisWidget, tertris.units),
	 XtRString, (caddr_t) "4"},
	{XtNdiagonal, XtCDiagonal, XtRBoolean, sizeof (Boolean),
	 XtOffset(NextTertrisWidget, tertris.diagonal),
	 XtRString, (caddr_t) "FALSE"},
	{XtNmixed, XtCMixed, XtRBoolean, sizeof (Boolean),
	 XtOffset(NextTertrisWidget, tertris.mixed),
	 XtRString, (caddr_t) "FALSE"},
	{XtNbonus, XtCBonus, XtRBoolean, sizeof (Boolean),
	 XtOffset(NextTertrisWidget, nextTertris.newBonus),
	 XtRString, (caddr_t) "FALSE"},
	{XtNplainPiece, XtCPlainPiece, XtRBoolean, sizeof (Boolean),
	 XtOffset(NextTertrisWidget, tertris.plainPiece),
	 XtRString, (caddr_t) "FALSE"},
	{XtNshowNext, XtCShowNext, XtRBoolean, sizeof (Boolean),
	 XtOffset(NextTertrisWidget, tertris.showNext),
	 XtRString, (caddr_t) "TRUE"},
	{XtNmenu, XtCMenu, XtRInt, sizeof (int),
	 XtOffset(NextTertrisWidget, tertris.menu),
	 XtRString, (caddr_t) "-1"},
	{XtNstart, XtCBoolean, XtRBoolean, sizeof (Boolean),
	 XtOffset(NextTertrisWidget, tertris.started),
	 XtRString, (caddr_t) "FALSE"},
	{XtNrandomNumber, XtCRandomNumber, XtRInt, sizeof (int),
	 XtOffset(NextTertrisWidget, nextTertris.randomNumber),
	 XtRString, (caddr_t) "-1"},
	{XtNrandomRotation, XtCRandomRotation, XtRInt, sizeof (int),
	 XtOffset(NextTertrisWidget, nextTertris.randomRotation),
	 XtRString, (caddr_t) "-1"},
	{XtNrandomReflection, XtCRandomReflection, XtRInt, sizeof (int),
	 XtOffset(NextTertrisWidget, nextTertris.randomReflection),
	 XtRString, (caddr_t) "-1"},
	{XtNselectCallback, XtCCallback, XtRCallback, sizeof (caddr_t),
	 XtOffset(NextTertrisWidget, tertris.select),
	 XtRCallback, (caddr_t) NULL}
};

NextTertrisClassRec nextTertrisClassRec =
{
	{
		(WidgetClass) & widgetClassRec,		/* superclass */
		(char *) "NextTertris",	/* class name */
		sizeof (NextTertrisRec),	/* widget size */
		NULL,		/* class initialize */
		NULL,		/* class part initialize */
		FALSE,		/* class inited */
		(XtInitProc) InitializeNextGame,	/* initialize */
		NULL,		/* initialize hook */
		XtInheritRealize,	/* realize */
		actionsListNextGame,	/* actions */
		XtNumber(actionsListNextGame),	/* num actions */
		resourcesNextGame,	/* resources */
		XtNumber(resourcesNextGame),	/* num resources */
		NULLQUARK,	/* xrm class */
		TRUE,		/* compress motion */
		TRUE,		/* compress exposure */
		TRUE,		/* compress enterleave */
		TRUE,		/* visible interest */
		(XtWidgetProc) DestroyNextGame,	/* destroy */
		(XtWidgetProc) ResizeNextGame,	/* resize */
		(XtExposeProc) ExposeNextGame,	/* expose */
		(XtSetValuesFunc) SetValuesNextGame,	/* set values */
		NULL,		/* set values hook */
		XtInheritSetValuesAlmost,	/* set values almost */
		NULL,		/* get values hook */
		NULL,		/* accept focus */
		XtVersion,	/* version */
		NULL,		/* callback private */
		defaultTranslationsNextGame,	/* tm table */
		NULL,		/* query geometry */
		NULL,		/* display accelerator */
		NULL		/* extension */
	},
	{
		0		/* ignore */
	}
};

WidgetClass nextTertrisWidgetClass = (WidgetClass) & nextTertrisClassRec;

static void
setNextGame(NextTertrisWidget w, int reason)
{
	tertrisCallbackStruct cb;

	cb.reason = reason;
	XtCallCallbacks((Widget) w, (char *) XtNselectCallback, &cb);
}
#endif

#include "image.h"

static Pixmap dr = 0; /* dummy for future double buffering */

static void
CheckThings(NextTertrisWidget w)
{
	char *buf1 = NULL, *buf2 = NULL;

	if (w->tertris.units < MIN_UNITS || w->tertris.units > MAX_UNITS) {
		intCat(&buf1, "Number of units out of bounds, use ",
			MIN_UNITS);
		stringCat(&buf2, buf1, "...");
		free(buf1);
		intCat(&buf1, buf2, MAX_UNITS);
		free(buf2);
		stringCat(&buf2, buf1, ", defaulting to ");
		free(buf1);
		intCat(&buf1, buf2, DEFAULT_UNITS);
		free(buf2);
		DISPLAY_WARNING(buf1);
		free(buf1);
		w->tertris.units = DEFAULT_UNITS;
		w->tertris.next.units = w->tertris.units;
	}
}

static void
drawUnit(NextTertrisWidget w, int pmid, int cid, int i, int j)
{
	int orient = (i + j + w->tertris.nextThing.size) % MAX_ORIENT;

	triangle_list[orient][0].x = (i - 1) * (w->tertris.triWidth / 2) +
		w->tertris.offset.x;
	triangle_list[orient][0].y = (j - 2) * w->tertris.triHeight +
		 w->tertris.offset.y +
		((orient != 0) ? 0 : w->tertris.triHeight - 1);
	fillPolygon((TertrisWidget) w, pmid, cid, orient,
		triangle_list[orient], False);
}

#if 0
static void
clearThing(TertrisWidget w)
{
	XFillRectangle(display, mainWin, w->tertris.inverseGC,
		w->tertris.offset.x,
		topMidY - w->tertris.triHeight * HEX_ROW(HEX_SIZE) / 2,
		w->tertris.viewWidth,
		w->tertris.triHeight * HEX_ROW(HEX_SIZE) + 2);
}
#endif

static void
clearThing(NextTertrisWidget w)
{
	FILLRECTANGLE(w, dr, w->tertris.inverseGC,
		w->tertris.offset.x, w->tertris.offset.y,
	        w->tertris.viewWidth, w->tertris.viewHeight);
}

static void
drawThing(NextTertrisWidget w)
{
	int i, j;
	int wid = HEX_COL(w->tertris.nextThing.size);
	int ht = HEX_ROW(w->tertris.nextThing.size);

	if (w->tertris.nextThing.polyiamondNumber == GAME_IGNORE)
		return;
	clearThing(w);
	for (i = 0; i < wid; i++)
		for (j = 0; j < ht; j++)
			if (inHexagon(w->tertris.nextThing.size, i, j) != 0 &&
				tris[w->tertris.nextThing.units].polyiamond
				[w->tertris.nextThing.polyiamondNumber].shape[j][i] != 0) {
				drawUnit(w, polyPixmap(i, j,
					(tris[w->tertris.nextThing.units].polyiamond
					[w->tertris.nextThing.polyiamondNumber].shape),
					w->tertris.nextThing.size),
					w->tertris.nextThing.colorNumber,
					w->tertris.nextThing.xpos + i,
					w->tertris.nextThing.ypos + j);
			}
}

static void
redoNext(NextTertrisWidget w)
{
	int diagonal = (w->tertris.next.diagonal) ? 1 : 0;
	int turnStyle = tris[w->tertris.next.units -
		MIN_UNITS].mode[diagonal].turnStyle;
	int next_start, i;

	w->tertris.nextThing.units = w->tertris.next.units - MIN_UNITS +
		w->tertris.nextThing.bonus;
	next_start = w->tertris.nextThing.randomNumber %
		tris[w->tertris.nextThing.units].mode
		[diagonal].number[(w->tertris.nextThing.bonus) ?
	       	NOMIX : w->tertris.next.mixed][turnStyle];
	w->tertris.nextThing.colorNumber = next_start;

	if (w->tertris.next.mixed && !w->tertris.nextThing.bonus) {
		w->tertris.nextThing.units = 0;
		while (next_start >= tris[w->tertris.nextThing.units].mode
				[diagonal].number[NOMIX][turnStyle]) {
			next_start -= tris[w->tertris.nextThing.units].mode
				[diagonal].number[NOMIX][turnStyle];
			w->tertris.nextThing.units++;
		}
	}
	w->tertris.nextThing.polyiamondNumber =
		tris[w->tertris.nextThing.units].mode
		[diagonal].start[next_start][turnStyle];
	if (tris[w->tertris.nextThing.units].mode[diagonal].turnStyle > NONE)
		for (i = 0; i < w->tertris.nextThing.randomRotation; i++)
			w->tertris.nextThing.polyiamondNumber =
				tris[w->tertris.nextThing.units].polyiamond
				[w->tertris.nextThing.polyiamondNumber].rotation;
	if (tris[w->tertris.nextThing.units].mode[diagonal].turnStyle == ALL)
		for (i = 0; i < w->tertris.nextThing.randomReflection; i++)
			w->tertris.nextThing.polyiamondNumber =
				tris[w->tertris.nextThing.units].polyiamond
				[w->tertris.nextThing.polyiamondNumber].reflection;
	w->tertris.nextThing.size = tris[w->tertris.nextThing.units].polyiamond
		[w->tertris.nextThing.polyiamondNumber].size;
	w->tertris.nextThing.xpos =
		(MAX_UNITS - w->tertris.nextThing.size) / 2;
	w->tertris.nextThing.ypos =
		(MAX_UNITS - w->tertris.nextThing.size) / 2;
}

static void
EraseFrame(NextTertrisWidget w)
{
	FILLRECTANGLE(w, dr, w->tertris.inverseGC,
		w->tertris.offset.x, w->tertris.offset.y,
		w->tertris.viewWidth, w->tertris.viewHeight);
}

static void
DrawFrame(NextTertrisWidget w, Boolean focus)
{
	DRAWRECTANGLE(w, dr, (focus) ? w->tertris.frameGC : w->tertris.borderGC,
		w->tertris.offset.x - 1, w->tertris.offset.y - 1,
		w->tertris.viewWidth + 1, w->tertris.viewHeight + 1);
}

#if 0
static void
ResetThings(NextTertrisWidget w)
{
}

static void
ResizeThings(NextTertrisWidget w)
{
}
#endif

#ifdef WINVER
static void
SetValuesNextGame(NextTertrisWidget w)
{
	struct tagColor {
		int red, green, blue;
	} color;
	char szBuf[80], buf[20];
	int thing;

	w->tertris.units = GetPrivateProfileInt(SECTION, "units",
		DEFAULTUNITS, INIFILE);
	w->tertris.diagonal = (BOOL) GetPrivateProfileInt(SECTION, "diagonal",
		DEFAULTDIAGONAL, INIFILE);
	w->tertris.mixed = (BOOL) GetPrivateProfileInt(SECTION, "mixed",
		DEFAULTMIXED, INIFILE);
	w->tertris.plainPiece = (BOOL) GetPrivateProfileInt(SECTION, "plainPiece",
		FALSE, INIFILE);
	w->tertris.showNext = (BOOL) GetPrivateProfileInt(SECTION, "showNext",
		FALSE, INIFILE);
	w->tertris.mono = (BOOL) GetPrivateProfileInt(SECTION, "mono",
		DEFAULTMONO, INIFILE);
	w->tertris.reverse = (BOOL) GetPrivateProfileInt(SECTION, "reverse",
		DEFAULTREVERSE, INIFILE);
	/* cyan */
	(void) GetPrivateProfileString(SECTION, "frameColor", "0 255 255",
		szBuf, sizeof (szBuf), INIFILE);
	(void) sscanf(szBuf, "%d %d %d",
		&(color.red), &(color.green), &(color.blue));
	w->tertris.frameGC = RGB(color.red, color.green, color.blue);
	/* gray25 */
	(void) GetPrivateProfileString(SECTION, "borderColor", "64 64 64",
		szBuf, sizeof (szBuf), INIFILE);
	(void) sscanf(szBuf, "%d %d %d",
		&(color.red), &(color.green), &(color.blue));
	w->tertris.borderGC = RGB(color.red, color.green, color.blue);
	/* #AEB2C3 */
	(void) GetPrivateProfileString(SECTION, "background", "255 255 255",
		szBuf, sizeof (szBuf), INIFILE);
	(void) sscanf(szBuf, "%d %d %d",
		&(color.red), &(color.green), &(color.blue));
	w->tertris.inverseGC = RGB(color.red, color.green, color.blue);
	for (thing = 0; thing < MAX_START_POLYIAMONDS; thing++) {
		(void) sprintf(buf, "thingColor%d", thing);
		(void) GetPrivateProfileString(SECTION, buf,
			polyiamondColorString[thing],
			szBuf, sizeof (szBuf), INIFILE);
		(void) sscanf(szBuf, "%d %d %d",
			&(color.red), &(color.green), &(color.blue));
		w->tertris.thingGC[thing] =
			RGB(color.red, color.green, color.blue);
		w->tertris.thingBrightGC[thing] =
			RGB(brighter(color.red), brighter(color.green),
			brighter(color.blue));
		w->tertris.thingDarkGC[thing] =
			RGB(darker(color.red), darker(color.green),
			darker(color.blue));
	}
}

void
DestroyNextGame(HBRUSH brush)
{
	(void) DeleteObject(brush);
	/*PostQuitMessage(0);*/
}

#else
static void
GetColor(NextTertrisWidget w, int thing)
{
	XGCValues values;
	XtGCMask valueMask;
	XColor colorCell, rgb;
	Boolean gotColor = False;

	valueMask = GCForeground | GCBackground;
	if (w->tertris.reverse) {
		values.background = w->tertris.foreground;
	} else {
		values.background = w->tertris.background;
	}
	if (!w->tertris.mono) {
		if (XAllocNamedColor(XtDisplay(w),
				DefaultColormapOfScreen(XtScreen(w)),
				polyiamondColorName[thing],
				&colorCell, &rgb)) {
			values.foreground = w->tertris.thingColor[thing] =
				colorCell.pixel;
			gotColor = True;
		} else {
			char *buf1, *buf2;

			stringCat(&buf1, "Color name \"",
				polyiamondColorName[thing]);
			stringCat(&buf2, buf1, "\" is not defined for polyiamond ");
			free(buf1);
			intCat(&buf1, buf2, thing);
			free(buf2);
			DISPLAY_WARNING(buf1);
			free(buf1);
		}
	}
	if (!gotColor) {
		if (w->tertris.reverse) {
			values.background = w->tertris.foreground;
			values.foreground = w->tertris.background;
		} else {
			values.background = w->tertris.background;
			values.foreground = w->tertris.foreground;
		}
	}
	if (w->tertris.thingGC[thing])
		XtReleaseGC((Widget) w, w->tertris.thingGC[thing]);
	w->tertris.thingGC[thing] = XtGetGC((Widget) w, valueMask, &values);
	gotColor = False;
	if (!w->tertris.mono) {
		if (XAllocNamedColor(XtDisplay(w),
				DefaultColormapOfScreen(XtScreen(w)),
				polyiamondColorName[thing],
				&colorCell, &rgb)) {
			values.foreground = w->tertris.thingColor[thing] =
				brighter((TertrisWidget) w, colorCell.pixel);
			gotColor = True;
		} else {
			char *buf1, *buf2;

			stringCat(&buf1, "Color name \"",
				polyiamondColorName[thing]);
			stringCat(&buf2, buf1, "\" is not defined for polyiamond ");
			free(buf1);
			intCat(&buf1, buf2, thing);
			free(buf2);
			DISPLAY_WARNING(buf1);
			free(buf1);
		}
	}
	if (!gotColor) {
		if (w->tertris.reverse) {
			values.background = w->tertris.foreground;
			values.foreground = w->tertris.background;
		} else {
			values.background = w->tertris.background;
			values.foreground = w->tertris.foreground;
		}
	}
	if (w->tertris.thingBrightGC[thing])
		XtReleaseGC((Widget) w, w->tertris.thingBrightGC[thing]);
	w->tertris.thingBrightGC[thing] = XtGetGC((Widget) w, valueMask,
		&values);
	gotColor = False;
	if (!w->tertris.mono) {
		if (XAllocNamedColor(XtDisplay(w),
				DefaultColormapOfScreen(XtScreen(w)),
				polyiamondColorName[thing],
				&colorCell, &rgb)) {
			values.foreground = w->tertris.thingColor[thing] =
				darker((TertrisWidget) w, colorCell.pixel);
			gotColor = True;
		} else {
			char *buf1, *buf2;

			stringCat(&buf1, "Color name \"",
				polyiamondColorName[thing]);
			stringCat(&buf2, buf1, "\" is not defined for polyiamond ");
			free(buf1);
			intCat(&buf1, buf2, thing);
			free(buf2);
			DISPLAY_WARNING(buf1);
			free(buf1);
		}
	}
	if (!gotColor) {
		if (w->tertris.reverse) {
			values.background = w->tertris.foreground;
			values.foreground = w->tertris.background;
		} else {
			values.background = w->tertris.background;
			values.foreground = w->tertris.foreground;
		}
	}
	if (w->tertris.thingDarkGC[thing])
		XtReleaseGC((Widget) w, w->tertris.thingDarkGC[thing]);
	w->tertris.thingDarkGC[thing] = XtGetGC((Widget) w, valueMask,
		&values);
}

static void
SetAllColors(NextTertrisWidget w)
{
	XGCValues values;
	XtGCMask valueMask;
	int polyiamond;

	valueMask = GCForeground | GCBackground;

	if (w->tertris.reverse) {
		values.background = w->tertris.background;
		values.foreground = w->tertris.foreground;
	} else {
		values.foreground = w->tertris.background;
		values.background = w->tertris.foreground;
	}
	if (w->tertris.inverseGC)
		XtReleaseGC((Widget) w, w->tertris.inverseGC);
	w->tertris.inverseGC = XtGetGC((Widget) w, valueMask, &values);
	if (w->tertris.mono) {
		if (w->tertris.reverse) {
			values.background = w->tertris.foreground;
			values.foreground = w->tertris.background;
		} else {
			values.foreground = w->tertris.foreground;
			values.background = w->tertris.background;
		}
	} else {
		values.foreground = w->tertris.frameColor;
		values.background = w->tertris.background;
	}
	if (w->tertris.frameGC)
		XtReleaseGC((Widget) w, w->tertris.frameGC);
	w->tertris.frameGC = XtGetGC((Widget) w, valueMask, &values);
	if (w->tertris.mono) {
		if (w->tertris.reverse) {
			values.background = w->tertris.foreground;
			values.foreground = w->tertris.background;
		} else {
			values.foreground = w->tertris.foreground;
			values.background = w->tertris.background;
		}
	} else {
		values.foreground = w->tertris.borderColor;
		values.background = w->tertris.background;
	}
	if (w->tertris.borderGC)
		XtReleaseGC((Widget) w, w->tertris.borderGC);
	w->tertris.borderGC = XtGetGC((Widget) w, valueMask, &values);
	for (polyiamond = 0; polyiamond < MAX_START_POLYIAMONDS; polyiamond++)
		GetColor(w, polyiamond);
}

static Boolean
SetValuesNextGame(Widget current, Widget request, Widget renew)
{
	NextTertrisWidget c = (NextTertrisWidget) current;
	NextTertrisWidget w = (NextTertrisWidget) renew;
	Boolean redraw = False, setColors = False;

	CheckThings(w);
	if (w->tertris.showNext != c->tertris.showNext) {
		redraw = True;
	}
	if (w->tertris.borderColor != c->tertris.borderColor ||
			w->tertris.frameColor != c->tertris.frameColor ||
			w->tertris.reverse != c->tertris.reverse ||
			w->tertris.mono != c->tertris.mono) {
		SetAllColors(w);
		redraw = True;
	} else if (w->tertris.background != c->tertris.background ||
			w->tertris.foreground != c->tertris.foreground ||
			setColors) {
		SetAllColors(w);
		redraw = True;
	} else if (w->tertris.plainPiece != c->tertris.plainPiece) {
		redraw = True;
	}
#if 0
	if (w->tertris.units != c->tertris.units) {
		w->tertris.next.units = w->tertris.units;
		ResetThings(w);
	} else if (w->tertris.diagonal != c->tertris.diagonal) {
		w->tertris.next.diagonal = w->tertris.diagonal;
		ResetThings(w);
	} else if (w->tertris.mixed != c->tertris.mixed) {
		w->tertris.next.mixed = w->tertris.mixed;
		ResetThings(w);
	}
#endif
	if (w->tertris.menu != -1) {
		switch (w->tertris.menu) {
		case 3:
			w->tertris.menu = -1;
			clearThing(w);
			redraw = True;
			break;
		default:
			w->tertris.menu = -1;
			break;
		}
	}
	if (w->nextTertris.randomNumber == GAME_CLEARNEXT) {
		if (w->tertris.showNext) {
			drawThing(w);
		} else {
			clearThing(w);
		}
		w->nextTertris.randomNumber = GAME_IGNORE;
		redraw = True;
	} else if (w->nextTertris.randomNumber != GAME_IGNORE) {
		w->tertris.nextThing.randomNumber = w->nextTertris.randomNumber;
		w->tertris.nextThing.randomRotation = w->nextTertris.randomRotation;
		w->tertris.nextThing.randomReflection = w->nextTertris.randomReflection;
		w->tertris.nextThing.bonus = w->nextTertris.newBonus;
		w->tertris.next.units = w->tertris.units;
		w->tertris.next.diagonal = w->tertris.diagonal;
		w->tertris.next.mixed = w->tertris.mixed;
		redoNext(w);
		if (w->tertris.showNext) {
			drawThing(w);
		}
		w->nextTertris.randomNumber = GAME_IGNORE;
		redraw = True;
	}
	return (redraw);
}

static void
QuitNextGame(NextTertrisWidget w, XEvent *event, char **args, int nArgs)
{
	Display *display = XtDisplay(w);

	XtCloseDisplay(display);
	exit(0);
}

static void
DestroyNextGame(Widget old)
{
	NextTertrisWidget w = (NextTertrisWidget) old;
	int thing;

	for (thing = 0; thing < MAX_START_POLYIAMONDS; thing++) {
		XtReleaseGC(old, w->tertris.thingGC[thing]);
		XtReleaseGC(old, w->tertris.thingBrightGC[thing]);
		XtReleaseGC(old, w->tertris.thingDarkGC[thing]);
	}
	XtReleaseGC(old, w->tertris.borderGC);
	XtReleaseGC(old, w->tertris.frameGC);
	XtReleaseGC(old, w->tertris.inverseGC);
	XtRemoveCallbacks(old, XtNselectCallback, w->tertris.select);
}
#endif

#define SQRT3_2 0.8660254 /* sqrt(3) / 2 */

#ifndef WINVER
static
#endif
void
ResizeNextGame(NextTertrisWidget w)
{
	int i, j;
#ifdef WINVER
	RECT rect;

	/* Determine size of client area */
	(void) GetClientRect(w->core.hWnd, &rect);
	w->core.width = rect.right;
	w->core.height = rect.bottom;
#endif
	w->tertris.triHeight = MIN((w->core.height - 2) / (MAX_UNITS / 2 + 2),
		(w->core.width - 2) * SQRT3_2 / (MAX_UNITS / 2 + 1));
	w->tertris.triWidth = (int) (w->tertris.triHeight / SQRT3_2);
	w->tertris.viewWidth = w->tertris.triWidth *
		(MAX_UNITS / 2 + 1) + 1;
	w->tertris.viewHeight = w->tertris.triHeight * (MAX_UNITS / 2 + 2);
	w->tertris.offset.x = ((int) w->core.width -
		w->tertris.viewWidth) / 2;
	w->tertris.offset.y = ((int) w->core.height -
		w->tertris.viewHeight) / 2;

	for (j = 0; j < MAX_ORIENT; j++) {
		for (i = 0; i < 3; i++) {
			triangle_list[j][i].x =
				(w->tertris.triWidth - 1) / 2 *
				triangle_unit[j][i].x;
			triangle_list[j][i].y =
				(w->tertris.triHeight - 1) *
				triangle_unit[j][i].y;
			triangle_border[j][i].x =
				w->tertris.triWidth / 2 * triangle_unit[j][i].x;
				triangle_border[j][i].y =
				w->tertris.triHeight * triangle_unit[j][i].y;
		}
		triangle_list[j][3].x = triangle_list[j][1].x;
		triangle_list[j][3].y = -triangle_list[j][1].y;
		triangle_border[j][3].x = triangle_border[j][1].x;
		triangle_border[j][3].y = -triangle_border[j][1].y;
	}
}

#if 0
#ifndef WINVER
static
#endif
void
SizeNextGame(NextTertrisWidget w)
{
	ResetThings(w);
	ResizeNextGame(w);
}
#endif

#ifndef WINVER
static
#endif
void
InitializeNextGame(
#ifdef WINVER
NextTertrisWidget w, HBRUSH brush
#else
Widget request, Widget renew
#endif
)
{
	int diagonal, mixed, turnStyle;
#ifndef WINVER
	int poly;
	NextTertrisWidget w = (NextTertrisWidget) renew;
#endif

#ifdef WINVER
	SetValuesNextGame(w);
#else
	w->tertris.mono = (DefaultDepthOfScreen(XtScreen(w)) < 2 ||
		w->tertris.mono);
	for (poly = 0; poly < MAX_START_POLYIAMONDS; poly++) {
		w->tertris.thingGC[poly] = NULL;
		w->tertris.thingBrightGC[poly] = NULL;
		w->tertris.thingDarkGC[poly] = NULL;
	}
	w->tertris.borderGC = NULL;
	w->tertris.frameGC = NULL;
	w->tertris.inverseGC = NULL;
#endif

	w->tertris.quit = False;
	w->tertris.focus = False;
	w->tertris.next.units = w->tertris.units;
	w->tertris.next.diagonal = w->tertris.diagonal;
	w->tertris.next.mixed = w->tertris.mixed;
	w->tertris.nextThing.polyiamondNumber = GAME_IGNORE;
	CheckThings(w);
#if 0
	SizeNextGame(w);
#else
	ResizeNextGame(w);
#endif
#ifdef WINVER
	brush = CreateSolidBrush(w->tertris.inverseGC);
	SETBACK(w->core.hWnd, brush);
	(void) SRAND(time(NULL));
#else
	(void) SRAND(getpid());
	SetAllColors(w);
#endif
	diagonal = (w->tertris.next.diagonal) ? 1 : 0;
	mixed = (w->tertris.next.mixed) ? 1 : 0;
	readThings();
	turnStyle = tris[w->tertris.next.units -
		MIN_UNITS].mode[diagonal].turnStyle;
	/* use intCat */
	if (turnStyle < 0 || turnStyle >= MAX_TURNSTYLE_TYPES) {
		(void) fprintf(stderr, "NextTertris: corrupted polyiamond file\n");
		(void) fprintf(stderr, "\tturnStyle = %d.\n", turnStyle);
		exit(1);
	}
	w->tertris.numberPolyiamonds = tris[w->tertris.next.units -
		MIN_UNITS].mode[diagonal].number[mixed][turnStyle];
	if (w->tertris.numberPolyiamonds <= 0 ||
			w->tertris.numberPolyiamonds > MAX_START_POLYIAMONDS) {
		(void) fprintf(stderr, "NextTertris: corrupted polyiamond file\n");
		(void) fprintf(stderr, "\tnumberPolyiamonds = %d.\n",
			w->tertris.numberPolyiamonds);
		exit(1);
	}
}

#ifndef WINVER
static
#endif
void
ExposeNextGame(
#ifdef WINVER
NextTertrisWidget w
#else
Widget renew, XEvent *event, Region region
#endif
)
{
#ifndef WINVER
	NextTertrisWidget w = (NextTertrisWidget) renew;

	if (!w->core.visible)
		return;
#endif
	EraseFrame(w);
#if 0
	ResizeThings(w);
#endif
	DrawFrame(w, w->tertris.focus);
	if (w->tertris.showNext) {
		drawThing(w);
	}
}

#ifndef WINVER
static
#endif
void
HideNextGame(NextTertrisWidget w
#ifndef WINVER
, XEvent *event, char **args, int nArgs
#endif
)
{
	setNextGame(w, GAME_HIDE);
}

#ifdef WINVER
void
ValueNextGame(NextTertrisWidget w,
		int randomNumber, int randomRotation, int randomReflection,
		int units, Boolean diagonal, Boolean mixed, Boolean newBonus)
{
	w->tertris.nextThing.randomNumber = randomNumber;
	w->tertris.nextThing.randomRotation = randomRotation;
	w->tertris.nextThing.randomReflection = randomReflection;
	w->tertris.next.units = w->tertris.units = units;
	w->tertris.next.diagonal = w->tertris.diagonal = diagonal;
	w->tertris.next.mixed = w->tertris.mixed = mixed;
	w->tertris.nextThing.bonus = newBonus;
	redoNext(w);
	if (w->tertris.showNext) {
		drawThing(w);

	}
}
#endif

#ifndef WINVER
static
#endif
void
EnterNextGame(NextTertrisWidget w
#ifndef WINVER
, XEvent *event, char **args, int nArgs
#endif
)
{
	w->tertris.focus = True;
	DrawFrame(w, w->tertris.focus);
}

#ifndef WINVER
static
#endif
void
LeaveNextGame(NextTertrisWidget w
#ifndef WINVER
, XEvent *event, char **args, int nArgs
#endif
)
{
	w->tertris.focus = False;
	DrawFrame(w, w->tertris.focus);
}

#ifndef WINVER
static
#endif
void
ShowNextNextGame(NextTertrisWidget w
#ifndef WINVER
, XEvent *event, char **args, int nArgs
#endif
)
{
	if (!w->tertris.paused) {
		w->tertris.showNext = !w->tertris.showNext;
		if (w->tertris.showNext) {
			drawThing(w);
		} else {
			clearThing(w);
		}
		FLUSH(w);
	}
}
