/*
# X-BASED TERTRIS
#
#  TertrisP.h
#
###
#
#  Copyright (c) 1993 - 2006		David A. Bagley, bagleyd@tux.org
#
#  Taken from GENERIC X-BASED TETRIS
#
#  Copyright (c) 1992 - 1995		Q. Alex Zhao, azhao@cc.gatech.edu
#
#			All Rights Reserved
#
#  Permission to use, copy, modify, and distribute this software and
#  its documentation for any purpose and without fee is hereby granted,
#  provided that the above copyright notice appear in all copies and
#  that both that copyright notice and this permission notice appear in
#  supporting documentation, and that the name of the author not be
#  used in advertising or publicity pertaining to distribution of the
#  software without specific, written prior permission.
#
#  This program is distributed in the hope that it will be "playable",
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
*/

/* Private header file for Tertris */

#ifndef _TertrisP_h
#define _TertrisP_h

#include "xwin.h"
#include "timer.h"
#include "polyiamond.h"
#include "Tertris.h"

#ifdef WINVER
extern void DestroyGame(HBRUSH brush);
extern void ResizeGame(TertrisWidget w);
extern void SizeGame(TertrisWidget w);
extern void InitializeGame(TertrisWidget w, HBRUSH brush);
extern void ExposeGame(TertrisWidget w);
extern void HideGame(TertrisWidget w);
extern void ClearGame(TertrisWidget w);
extern void EnterGame(TertrisWidget w);
extern void LeaveGame(TertrisWidget w);
extern void MoveGameInput(TertrisWidget w, int direction);
extern void SoundGame(TertrisWidget w);
extern Boolean IncrementGameLevel(TertrisWidget w);
extern Boolean DecrementGameLevel(TertrisWidget w);
extern Boolean IncrementGameUnits(TertrisWidget w);
extern Boolean DecrementGameUnits(TertrisWidget w);
extern Boolean IncrementGameColumns(TertrisWidget w);
extern Boolean DecrementGameColumns(TertrisWidget w);
extern void FixedFillGame(TertrisWidget w);
extern void SuddenAppearGame(TertrisWidget w);
extern Boolean DiagonalGame(TertrisWidget w);
extern Boolean MixedGame(TertrisWidget w);
extern void BonusGame(TertrisWidget w);
extern void CwGame(TertrisWidget w);
extern void PlainPieceGame(TertrisWidget w);
extern void FalldownGame(TertrisWidget w);
extern void PauseGame(TertrisWidget w);
extern void ShowNextGame(TertrisWidget w);
extern void GridGame(TertrisWidget w);
extern void Fill0Game(TertrisWidget w);
extern void Fill1Game(TertrisWidget w);
extern void Fill2Game(TertrisWidget w);
extern void Fill3Game(TertrisWidget w);
extern void Fill4Game(TertrisWidget w);
extern void Fill5Game(TertrisWidget w);
extern void Fill6Game(TertrisWidget w);
extern void Fill7Game(TertrisWidget w);
extern void Fill8Game(TertrisWidget w);
extern void Fill9Game(TertrisWidget w);
extern void Fill10Game(TertrisWidget w);
extern void setGame(TertrisWidget w, int reason);
extern void setGameNext(TertrisWidget w, int reason,
	int randomNumber, int randomRotation, int randomReflection,
	int units, Boolean diagonal, Boolean mixed, Boolean newBonus);

#else

/* This gets around C's inability to do inheritance */
typedef struct _TertrisClassPart {
	int         ignore;
} TertrisClassPart;

typedef struct _TertrisClassRec {
	CoreClassPart coreClass;
	TertrisClassPart tertrisClass;
} TertrisClassRec;

extern TertrisClassRec tertrisClassRec;
#endif

#ifndef BUMPSOUND
#define BUMPSOUND "bump"
#endif

#define SYMBOL ':'

#define IGNORE_DIR (-1)

#define NUM_FLASHES 4

#define MAX_SIDES 6
#define MAX_ORIENT 2
#define MIN_LEVELS 1
#define MAX_LEVELS 20

#define THRESHOLD(x)	((x)*10)

/*** types ***/

typedef enum {FALL, DROP, MOVE_LEFT, MOVE_RIGHT, ROTATE, REFLECT} move_t;

typedef struct {
	int         units;
	Boolean     diagonal, mixed, bonus;
} Spec;

typedef struct {
	int         units, polyiamondNumber;
	int         xpos, ypos;
	int         size, colorNumber, direction;
	int         bonus, randomRotation, randomReflection;
	long        randomNumber;
} Thing;

typedef struct {
	int         pmid, cid;
} Field;

typedef struct _TertrisPart {
	Thing       curThing, nextThing;
	Field      *field;
	GC          thingGC[MAX_START_POLYIAMONDS];
	GC          thingBrightGC[MAX_START_POLYIAMONDS];
	GC          thingDarkGC[MAX_START_POLYIAMONDS];
	Boolean     fixedFill, suddenAppear, cw, plainPiece, showNext, grid;
	Boolean     invert, sound, resetTime, firstFall, paused, gotNew;
	int         level, fill, completedRows;
	int         startLevel, score, newRows;
	Spec        current, next;
	int         units, columns, rows;
	Boolean     diagonal, mixed, bonus;
	int         delay, numberPolyiamonds, pauseStart;
	int         currentPosition;
	Boolean     fall, started, quit, act;
	Boolean     mono, reverse, focus;
	Boolean     scoreOnly, versionOnly;
	int         triWidth, triHeight;
	Position    viewWidth, viewHeight;
	Point       offset;
	GC          frameGC, borderGC, inverseGC, xorGC;
#ifdef WINVER
	char        userName[81], scoreFile[81], bumpSound[81];
#else
	int         menu;
	char       *userName, *scoreFile, *bumpSound;
	Pixel       foreground, background;
	Pixel       frameColor, borderColor;
	Pixel       thingColor[MAX_START_POLYIAMONDS];
	XtCallbackList select;
#endif
} TertrisPart;

typedef struct _TertrisRec {
	CorePart    core;
	TertrisPart  tertris;
} TertrisRec;

#ifdef WINVER
extern int brighter(const int light);
extern int darker(const int light);
#else
extern Pixel brighter(TertrisWidget w, Pixel pixel);
extern Pixel darker(TertrisWidget w, Pixel pixel);
#endif
extern void fillPolygon(TertrisWidget w, int pm, int color, Boolean invert,
	Point *points, Boolean graphicalXor);
#endif /* _TertrisP_h */
