/*
# MOTIF/X-BASED TERTRIS
#
#  xtertris.c
#
###
#
#  Copyright (c) 1993 - 2006	David Albert Bagley, bagleyd@tux.org
#
#  Taken from GENERIC X-BASED TETRIS
#
#  Copyright (c) 1992 - 1995	Q. Alex Zhao, azhao@cc.gatech.edu
#
#                   All Rights Reserved
#
#  Permission to use, copy, modify, and distribute this software and
#  its documentation for any purpose and without fee is hereby granted,
#  provided that the above copyright notice appear in all copies and
#  that both that copyright notice and this permission notice appear in
#  supporting documentation, and that the name of the author not be
#  used in advertising or publicity pertaining to distribution of the
#  software without specific, written prior permission.
#
#  This program is distributed in the hope that it will be "playable",
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
*/

/*-
  Version 7: 05/05/31 X/Windows
  Version 1: 95/06/21 Generic X
 */

#include "file.h"
#ifdef WINVER
#include "TertrisP.h"
#include "TertrisNP.h"
#include "wtertris.h"
#define TITLE "wtertris"

static TertrisRec widget;
static NextTertrisRec widget2;

#ifndef SCOREPATH
#ifdef UNIXDELIM
#define SCOREPATH "c:/WINDOWS"
#else
#define SCOREPATH "c:\\WINDOWS"
#endif
#endif
#define PRINT_MESSAGE(b) (void) MessageBox(widget.core.hWnd, (LPCSTR) b, "Warning", MB_OK);
#else
#include "xwin.h"
#include <X11/Shell.h>
#include <X11/cursorfont.h>
#ifdef HAVE_MOTIF
#include <Xm/PanedW.h>
#include <Xm/RowColumn.h>
#include <Xm/Label.h>
#include <Xm/LabelG.h>
#include <Xm/MessageB.h>
#include <Xm/PushBG.h>
#include <Xm/CascadeB.h>
#include <Xm/Scale.h>
#include <Xm/ToggleB.h>
#include <Xm/ToggleBG.h>
#include <Xm/Form.h>
#include <Xm/DialogS.h>
#define PRINT_MESSAGE(b) PrintState(message, b)
#else
#define PRINT_MESSAGE(b) XtWarning(b)
#endif
#include "Tertris.h"
#include "TertrisN.h"
#ifdef HAVE_XPM
#include <X11/xpm.h>
#include "icons/tertris.xpm"
#endif
#include "icons/tertris.xbm"
#ifndef SCOREPATH
#ifdef VMS
#define SCOREPATH "SYS$LOGIN:"
#else
#define SCOREPATH "/var/games/altris"
#endif
#endif
#endif

#ifndef WINVER
static const char aboutHelp[] = {
"Tertris Version 7.2.2\n"
"Send bugs (reports or fixes) to the author: "
"David Bagley <bagleyd@tux.org>\n"
"The latest version is at: "
"http://www.tux.org/~bagleyd/altris.html\n"
};

static const char optionsHelp[] = {
"[-geometry [{width}][x{height}][{+-}{xoff}[{+-}{yoff}]]]\n"
"[-display [{host}]:[{vs}]] [-[no]mono] [-[no]{reverse|rv}]\n"
"[-{foreground|fg} {color}] [-{background|bg} {color}]\n"
"[-frame {color}] [-{border|bd} {color}] [-[no]sound]\n"
"[-bumpSound {filename}] [-level {starting_level}]\n"
"[-units {size_of_polyiamond}] [-columns {columns}]\n"
"[-fill {rows}] [-[no]fixedFill] [-[no]suddenAppear]\n"
"[-[no]diagonal] [-[no]mixed] [-[no]bonus] [-[no]cw]\n"
"[-[no]plain] [-[no]showNext] [-[no]grid]\n"
"[-[no]sound] [-bumpSound {filename}]\n"
"[-userName {string}] [-scoreFile {filename}] [-scores]\n"
"[-version]\n"
};
#endif

#if defined(HAVE_MOTIF) || defined(WINVER)
static const char descriptionHelp[] = {
"The ever popular falling tetromino game now with triangles.  "
"Enhanced to allow all polyiamond classes that do not have\n"
"pieces with holes.  Mixing polyiamond classes (so as to "
"permit polyiamonds made up by a variable number triangles)\n"
"can also be selected, as well as, 1 set of diagonal "
"polyiamonds that do not have pieces with holes.\n"
"\n"
"To make these games interesting and \"do-able\" the ability "
"to rotate gets turned off if it is easier than the 5\n"
"triangle polyiamond.  An added ability \"reflect\" "
"(right-left -> left-right) allows one to attempt the 6 polyiamond\n"
"pieces.\n"
"\n"
"To get high scores, do not use \"show next\" mode, try "
"starting at a higher level and try to let \"tertris\" eat more\n"
"than one line at a time.  Polyiamonds with diagonal or "
"with a higher number of triangles earn more points.  Mixed\n"
"mode earns less points.  Play fast and see how well you "
"can do in the bonus mode.  Bonuses are set when the\n"
"level is changed automatically.  (Bonus mode not available "
"for 8 triangles or diagonal mode).\n"
"\n"
"The program lists the top 15 players at the end, each player "
"has at most three scores.\n"
};

static const char featuresHelp[] = {
"Click \"mouse-left\" button , or press \"J\", \"j\", "
"\"S\", or \"s\" keys to move the falling polyiamond left.\n"
"\n"
"Click \"mouse-center\" button, or press \"K\", \"k\", "
"\"D\", or \"d\" keys to rotate the falling polyiamond.\n"
"Note this may be turned off if the depending on the "
"\"turn-style\" but normally this is \"on\".\n"
"\n"
"Click \"mouse-right\" button, or press \"L\", \"l\", "
"\"F\", or \"f\" keys to move the falling polyiamond right.\n"
"\n"
"Click \"mouse-scroll-up\", or press \"I\", \"i\", "
"\"E\", or \"e\" keys to flip (reflect) the falling polyiamond.\n"
"Note this may be turned off if the depending on the "
"\"turn-style\" but normally this is \"off\".\n"
"\n"
"Click \"mouse-scroll-down\", or press \"space-bar\" "
"to drop the falling polyiamond.\n"
"\n"
"The key pad and arrow keys are defined similar.\n"
"Key pad is defined for Tertris as:\n"
"  /    Reflect\n"
"  8    Rotate\n"
"  ^\n"
"4<5>6  Left, Rotate, Right\n"
"  v\n"
"  2    Drop\n"
"\n"
"Press \"C\" or \"c\" keys to clear.\n"
"\n"
"Press \">\" or \".\" keys to speed up "
"the falling polyiamond.\n"
"\n"
"Press \"<\" or \",\" keys to slow down "
"the falling polyiamond.\n"
"This may be turned off to avoid cheating.\n"
"\n"
"Press \"+\" or \"=\" keys to increase the "
"triangles in the falling polyiamond.  Note: To keep the\n"
"level of play similar, the \"turn style\" can change.  "
"The ability to reflect polyiamonds is available\n"
"only when the number of triangles >= 5.  The ability "
"to rotate polyiamonds, is available only when\n"
"the number of triangles is >= 4.  Bonus polyiamonds "
"still have the current \"turn style\".\n"
"\n"
"Press \"-\" or \"_\" keys to decrease the triangles in "
"the falling polyiamond.\n"
"\n"
"Press \"]\" or \"}\" keys to increase the column/row "
"size.\n"
"\n"
"Press \"[\" or \"{\" keys to decrease the column/row "
"size.\n"
"\n"
"Press \"`\", \"1\", \"2\", \"3\", \"4\", \"5\", "
"\"6\", \"7\", \"8\", \"9\", or \"0\" keys\n"
"(not the keypad 0-9) to change the number of fill lines.  "
"\"`\" for 0 lines and \"0\" for 10 lines.\n"
"\n"
"Press \"X\" or \"x\" to toggle between fixed fill or "
"random fill.\n"
"\n"
"Press \"R\" or \"r\" to toggle between sudden initial "
"appearance or gradual initial appearance.\n"
"\n"
"Press \"G\" or \"g\" to toggle diagonal mode.  This "
"may not be available for current number of triangles.\n"
"\n"
"Press \"M\" or \"m\" to toggle mixed mode.  This "
"may not be available for current number of triangles.\n"
"\n"
"Press \"B\" or \"b\" to toggle bonus mode.  This "
"may not be available for current number of triangles.\n"
"\n"
"Press \"W\" or \"w\" to toggle clockwise or counter "
"clockwise turning.\n"
"\n"
"Press \"P\" or \"p\" to toggle plain pieces or "
"three dimensional pieces.\n"
"\n"
"Press \"N\" or \"n\" to toggle show next.\n"
"\n"
"Press \"#\" key to toggle the grid.\n"
"\n"
"Press \"|\" key to pause.\n"
"\n"
"Press \"@\" key to toggle the sound.\n"
"\n"
"Press \"Esc\" key to hide and pause program.  When the "
"tertris window loses its focus, it also goes to\n"
"\"paused\" mode and iconifies itself.\n"
"\n"
"Press \"Q\", \"q\", or \"CTRL-C\" keys to kill "
"program.\n"
};

static const char referencesHelp[] = {
"Based at one time on Q. Alex Zhao's Tetris.\n"
};
#endif

#ifndef SCOREFILE
#define SCOREFILE "tertris.scores"
#endif

#define NEVER (-1)
#define FILENAMELEN 1024
#define USERNAMELEN 120
#define DATELEN 27
#define MESSAGELEN (USERNAMELEN+64)
#define TITLELEN 2048
#define MAXSCORES 15
#define MAXUSERSCORES 3
#define NOACCESS ""
#define NOBODY "nobody"

typedef struct {
	int score, level, rows;
	char name[USERNAMELEN];  /* host[USERNAMELEN] */
	char date[DATELEN];
} GameRecord;

static GameRecord tertrisRecord[MAXSCORES];
static char messageDsp[MESSAGELEN] = "Welcome";

#ifdef HAVE_MOTIF
static Widget scoreText, maxScoreText, message, rowsText, levelSlider;
static Widget unitsSlider, fieldSlider, fillSlider;
static Widget fixedFillSwitch, suddenAppearSwitch;
static Widget diagonalSwitch, mixedSwitch, bonusSwitch;
static Widget cwSwitch, plainSwitch, showNextSwitch;
static Widget gridSwitch, soundSwitch;
static char buff[21];
static Widget setupDialog = NULL, descriptionDialog, featuresDialog;
static Widget optionsDialog, referencesDialog, aboutDialog;
static Widget restartDialog;
static Widget menuBar;
static Arg arg[3];
#else
#ifndef WINVER
static Widget shell;
#endif
static char titleDsp[TITLELEN] = "";
#endif
static char scoreFileName[USERNAMELEN] = SCOREFILE;
#ifdef WINVER
#define MAXPROGNAME 80
static char progDsp[MAXPROGNAME] = TITLE;
static char userNameDsp[USERNAMELEN] = "Guest";
#else
static Pixmap tertrisIcon = None;
static Widget topLevel, tertris, nextTertris;
static char *progDsp;
static char userNameDsp[USERNAMELEN] = "";

#ifdef HAVE_MOTIF
static void
PrintState(Widget w, char *msg)
{
	XmString xmstr;

	if (!XtIsSubclass(w, xmLabelWidgetClass))
		XtError("PrintState() requires a Label Widget");
	xmstr = XmStringCreateLtoR(msg, XmSTRING_DEFAULT_CHARSET);
	XtVaSetValues(w, XmNlabelString, xmstr, NULL);
}
#endif

static void
PrintRecords(void)
{
	int i;

	(void) printf("	    TERTRIS  HALL OF FAME\n\n");
	(void) printf("   # USER	     SCORE LEVEL ROWS\n");
	for (i = 0; i < MAXSCORES; i++) {
		if (tertrisRecord[i].score > 0) {
			(void) printf("%4d %-12s%9d %5d %5d\n", i + 1,
				tertrisRecord[i].name, tertrisRecord[i].score,
				tertrisRecord[i].level, tertrisRecord[i].rows);
		} else  {
			break;
		}
	}
}
#endif

static void
InitRecords(void)
{
	int i;

	for (i = 0; i < MAXSCORES; i++) {
		tertrisRecord[i].score = NEVER;
		tertrisRecord[i].level = 0;
		tertrisRecord[i].rows = 0;
		(void) strncpy(tertrisRecord[i].name,
			NOACCESS, USERNAMELEN);
		tertrisRecord[i].date[0] = '\0';
	}
}

/* #define DEBUG 1 */

static void
ReadRecords(void)
{
	FILE *fp;
	int s = NEVER, l, r, i;
	char userName[USERNAMELEN];
	char *buf1 = NULL, *buf2 = NULL;
	char *fname, *lname, *name;

	stringCat(&buf1, CURRENTDELIM, scoreFileName);
	lname = buf1;
	stringCat(&buf1, SCOREPATH, FINALDELIM);
	stringCat(&buf2, buf1, SCOREFILE);
	free(buf1);
	fname = buf2;
	name = fname;
	if ((fp = fopen(name, "r")) == NULL) {
		/* Try current directory in case its not installed yet. */
		name = lname;
		if ((fp = fopen(name, "r")) == NULL) {
			stringCat(&buf1, "Can not read ", fname);
			stringCat(&buf2, buf1, " or ");
			free(buf1);
			stringCat(&buf1, buf2, lname);
			free(buf2);
			PRINT_MESSAGE(buf1);
			free(buf1);
			free(lname);
			free(fname);
			return;
		}
/* Probably annoying */
#if 0
		else {
			stringCat(&buf1, "Can not read ", fname);
			stringCat(&buf2, buf1, ", falling back to ");
			free(buf1);
			stringCat(&buf1, buf2, lname);
			free(buf2);
			PRINT_MESSAGE(buf1);
			free(buf1);
		}
#endif
	}
	free(lname);
	free(fname);
	for (i = 0; i < MAXSCORES; i++)
		tertrisRecord[i].score = 0;
	for (i = 0; i < MAXSCORES; i++) {
		if (fscanf(fp, "%d %d %d %s\n", &s, &l, &r, userName) > 0) {
			tertrisRecord[i].score = s;
			tertrisRecord[i].level = l;
			tertrisRecord[i].rows = r;
			(void) strncpy(tertrisRecord[i].name,
				userName, USERNAMELEN);
		} else {
			break;
		}
	}
	(void) fclose(fp);
}

static void
WriteRecords(void)
{
	FILE *fp;
	int i;
	char *buf1 = NULL, *buf2 = NULL;
	char *fname, *lname, *name;

	stringCat(&buf1, CURRENTDELIM, scoreFileName);
	lname = buf1;
	stringCat(&buf1, SCOREPATH, FINALDELIM);
	stringCat(&buf2, buf1, SCOREFILE);
	free(buf1);
	fname = buf2;
	name = fname;
	if ((fp = fopen(name, "w")) == NULL) {
		/* Try current directory in case its not installed yet. */
		name = lname;
		if ((fp = fopen(name, "w")) == NULL) {
			stringCat(&buf1, "Can not write to ", fname);
			stringCat(&buf2, buf1, " or ");
			free(buf1);
			stringCat(&buf1, buf2, lname);
			free(buf2);
			PRINT_MESSAGE(buf1);
			free(buf1);
			free(lname);
			free(fname);
			return;
		}
/* Probably annoying */
#if 0
		else {
			stringCat(&buf1, "Can not write to ", fname);
			stringCat(&buf2, buf1, ", falling back to ");
			free(buf1);
			stringCat(&buf1, buf2, lname);
			free(buf2);
			PRINT_MESSAGE(buf1);
			free(buf1);
		}
#endif
	}
	{
#if HAVE_FCNTL_H
		int lfd;
		char lockfile[FILENAMELEN];

		(void) strncpy(lockfile, name, FILENAMELEN - 6);
		(void) strcat(lockfile, ".lock");
		while (((lfd = open(lockfile, O_CREAT | O_EXCL, 0644)) < 0) &&
				errno == EEXIST)
			(void) sleep(1);
		if (lfd < 0) {
#if 1
			(void) fprintf(stderr,
				"Lock file exists... guessing its an old one.\n");
#else
			(void) fprintf(stderr,
				"Lock file exists... score not recorded - sorry.\n");
			return;
#endif
		}
#endif
		for (i = 0; i < MAXSCORES; i++) {
			if (tertrisRecord[i].score <= 0)
				break;
			(void) fprintf(fp, "%d %d %d %s\n",
				tertrisRecord[i].score,
				tertrisRecord[i].level,
				tertrisRecord[i].rows,
				tertrisRecord[i].name);
		}
		(void) fprintf(fp, "\n");
#if HAVE_FCNTL_H
		(void) close(lfd);
		(void) unlink(lockfile);
#endif
		(void) fclose(fp);
	}
	free(lname);
	free(fname);
}

#ifndef WINVER
/* There's probably a better way to assure that they are the same but I don't
 *    know it off hand. */
static void
MakeEquivalent(String *userName, String *scoreFile, int *level,
		int *units, Boolean *diagonal, Boolean *mixed, Boolean *bonus,
		Boolean *showNext)
{
	Boolean mono, reverse, plainPiece, grid;
	Boolean scoreOnly, versionOnly;
	Pixel foreground, background, borderColor, frameColor;

	XtVaGetValues(tertris,
		XtNuserName, userName,
		XtNscoreFile, scoreFile,
		XtNlevel, level,
		XtNunits, units,
		XtNdiagonal, diagonal,
		XtNmixed, mixed,
		XtNbonus, bonus,
		XtNplainPiece, &plainPiece,
		XtNshowNext, showNext,
		XtNgrid, &grid,
		XtNmono, &mono,
		XtNreverse, &reverse,
		XtNforeground, &foreground,
		XtNbackground, &background,
		XtNborderColor, &borderColor,
		XtNframeColor, &frameColor,
		XtNscoreOnly, &scoreOnly,
		XtNversionOnly, &versionOnly, NULL);
	if (versionOnly) {
		(void) printf(aboutHelp);
		exit(0);
	}
	if (scoreOnly) {
		InitRecords();
		ReadRecords();
		PrintRecords();
		exit(0);
	}
	XtVaSetValues(nextTertris,
		XtNlevel, *level,
		XtNunits, *units,
		XtNdiagonal, *diagonal,
		XtNmixed, *mixed,
		XtNbonus, *bonus,
		XtNplainPiece, plainPiece,
		XtNshowNext, *showNext,
		XtNgrid, grid,
		XtNmono, mono,
		XtNreverse, reverse,
		XtNforeground, foreground,
		XtNbackground, background,
		XtNborderColor, borderColor,
		XtNframeColor, frameColor,
		NULL);
}
#endif

static void
PrintStatus(
#ifndef HAVE_MOTIF
	char *msg, int level,
#endif
	int numberRows, int yourScore)
{
	if (tertrisRecord[0].score < yourScore) {
		tertrisRecord[0].score = yourScore;
	}
#ifdef HAVE_MOTIF
	(void) sprintf(buff, "%d", numberRows);
	PrintState(rowsText, buff);
	(void) sprintf(buff, "%d", yourScore);
	PrintState(scoreText, buff);
	(void) sprintf(buff, "%d", tertrisRecord[0].score);
	PrintState(maxScoreText, buff);
#else
	(void) sprintf(titleDsp, "%s: %d level %d rows @ (%d/%d) - %s",
		progDsp, level, numberRows, yourScore,
		tertrisRecord[0].score, msg);
#ifdef WINVER
	SetWindowText(widget.core.hWnd, (LPSTR) titleDsp);
#else
	XtVaSetValues(XtParent(tertris), XtNtitle, titleDsp, NULL);
#endif
#endif
}

#define CPSCORE(a,b) a.score = b.score, a.level = b.level, a.rows = b.rows,\
	(void) strncpy(a.name, b.name, USERNAMELEN)

static Boolean
HandleGameOver(int lastLevel, int numberRows, int yourScore)
{
	int i, j = 0, k = 0, rank = MAXSCORES;
	GameRecord current, temp[MAXSCORES];
	Boolean newRecord = False;

	ReadRecords();	/* Maybe its been updated by another */
	current.score = yourScore;
	current.level = lastLevel;
	current.rows = numberRows;
	(void) strncpy(current.name, userNameDsp, USERNAMELEN);
	for (i = 0; i < MAXSCORES; i++) {
		if (j == 0 && current.score > tertrisRecord[i].score) {
			CPSCORE(temp[i], current);
			j = 1;
			rank = i;
		} else {
			CPSCORE(temp[i], tertrisRecord[i - j]);
		}
	}
	j = 0;
	if (rank < MAXSCORES) {
		for (i = 0; i < MAXSCORES; i++) {
			if (!strcmp(temp[i].name, userNameDsp)) {
				if (j < MAXUSERSCORES) {
					CPSCORE(tertrisRecord[k], temp[i]);
					if (i == rank)
						newRecord = True;
					k++;
				}
				j++;
			} else {
				CPSCORE(tertrisRecord[k], temp[i]);
				k++;
			}
		}
		for (i = k; i < MAXSCORES; i++) {
			tertrisRecord[i].score = 0;
		}
	}
	if (newRecord) {
		WriteRecords();
		PrintStatus(
#ifndef HAVE_MOTIF
			messageDsp, lastLevel,
#endif
			numberRows, yourScore);
		return True;
	}
	return False;
}
#ifdef WINVER
static LRESULT CALLBACK
ShowNext(HWND hDlg, UINT message, WPARAM wParam, LPARAM lParam)
{
	HBRUSH brush = (HBRUSH) NULL;
	PAINTSTRUCT paint;

	if (widget2.core.hWnd == NULL)
		widget2.core.hWnd = hDlg;
	if (GetFocus()) {
		if (!widget2.tertris.focus) {
			widget2.core.hDC = GetDC(widget2.core.hWnd);
			(void) SelectObject(widget2.core.hDC, GetStockObject(NULL_BRUSH));
			EnterNextGame(&widget2);
			(void) EndPaint(widget2.core.hWnd, &paint);
		}
	} else {
		if (widget2.tertris.focus) {
			widget2.core.hDC = GetDC(widget2.core.hWnd);
			(void) SelectObject(widget2.core.hDC, GetStockObject(NULL_BRUSH));
			LeaveNextGame(&widget2);
			(void) EndPaint(widget2.core.hWnd, &paint);
		}
	}
	switch (message) {
		case WM_DESTROY:
			DestroyNextGame(brush);
			(void) DestroyWindow(widget2.core.hWnd);
			widget2.core.hWnd = NULL;
			return TRUE;
		case WM_SIZE:
			ResizeNextGame(&widget2);
			(void) InvalidateRect(widget2.core.hWnd, NULL, TRUE);
			break;
		case WM_PAINT:
			ExposeNextGame(&widget2);
			(void) InvalidateRect(widget2.core.hWnd, NULL, TRUE);
			break;
		case WM_COMMAND:
			switch (LOWORD(wParam)) {
				case IDM_EXIT:
					DestroyNextGame(brush);
					(void) DestroyWindow(widget2.core.hWnd);
					widget2.core.hWnd = NULL;
					return TRUE;
				case IDM_HIDE:
					HideNextGame(&widget2);
					break;
#if 0
				case IDM_CLEAR:
					widget2.core.hDC = GetDC(widget2.core.hWnd);
					(void) SelectObject(widget2.core.hDC, GetStockObject(NULL_PEN));
					ShowNextGame(&widget2);
					(void) ReleaseDC(widget2.core.hWnd, widget2.core.hDC);
					break;
#endif
				case IDM_SHOWNEXT:
					if (widget2.tertris.showNext) {
						ShowWindow(widget2.core.hWnd, SW_SHOWNOACTIVATE);
					} else {
						ShowWindow(widget2.core.hWnd, SW_HIDE);
					}
					break;
#if 0
				case IDM_SHOWNEXT:
					if (widget2.core.hWnd == NULL)
						widget2.core.hWnd = CreateDialog(widget2.core.hInstance,
							"ShowNext", widget2.core.hWnd, (DLGPROC) ShowNext);
					else {
						DestroyNextGame(brush);
						(void) DestroyWindow(widget2.core.hWnd);
						widget2.core.hWnd = NULL;
						return TRUE;
					}
					break;
#endif
			}
			break;
		/*default:
			return (DefWindowProc(widget2.core.hWnd, message, wParam, lParam)); */
	}
	return FALSE;
}
#endif

static void
Initialize(
#ifdef WINVER
TertrisWidget w, HBRUSH brush, HBRUSH brush2
#else
Widget w
#endif
)
{
	int level, units;
	Boolean diagonal, mixed, bonus, showNext;
	char *userName, *scoreFile;

#ifdef WINVER
	InitializeGame(w, brush);
	level = w->tertris.level;
	units = widget2.tertris.units = w->tertris.units;
	diagonal = widget2.tertris.diagonal = w->tertris.diagonal;
	mixed = widget2.tertris.mixed = w->tertris.mixed;
	bonus = widget2.tertris.bonus = w->tertris.bonus;
	userName = w->tertris.userName;
	scoreFile = w->tertris.scoreFile;
	InitializeNextGame(&widget2, brush2);
	(void) CreateDialog(widget.core.hInstance,
		"ShowNext", widget.core.hWnd, (DLGPROC) ShowNext);
	if (!widget2.tertris.showNext)
		ShowWindow(widget2.core.hWnd, SW_HIDE);
	showNext = widget.tertris.showNext = widget2.tertris.showNext;
#else
	MakeEquivalent(&userName, &scoreFile, &level, &units,
		&diagonal, &mixed, &bonus, &showNext);
#ifdef HAVE_MOTIF
	if (level < MIN_LEVELS || level > MAX_LEVELS)
		level = DEFAULTLEVELS;
	XmScaleSetValue(levelSlider, level);
#endif
#endif
	InitRecords();
	ReadRecords();
#ifndef WINVER
	(void) strncpy(userNameDsp, userName, USERNAMELEN);
#endif
	if (!strcmp(userName, "(null)") || !strcmp(userName, NOBODY) ||
			!strcmp(userName, NOACCESS)) {
#ifdef WINVER
		(void) strncpy(userNameDsp, userName, USERNAMELEN);
#else
		char * login = getlogin();

		if (login == NULL) {
			(void) strcpy(userNameDsp, "");
		} else {
			(void) sprintf(userNameDsp, "%s", login);
		}
		if (!strcmp(userNameDsp, "(null)") ||
				!strcmp(userNameDsp, NOBODY) ||
				!strcmp(userNameDsp, NOACCESS))
			/* It really IS nobody */
			(void) sprintf(userNameDsp, "%s", "guest");
#endif
	}
	if (strcmp(scoreFile, ""))
		(void) strncpy(scoreFileName, scoreFile, USERNAMELEN);
	PrintStatus(
#ifndef HAVE_MOTIF
		messageDsp, level,
#endif
		0, 0);
}

#ifdef WINVER
void
setGame(TertrisWidget w, int reason)
#else
static void
CallbackGame(Widget w, caddr_t clientData, tertrisCallbackStruct * callData)
#endif
{
#ifndef WINVER
	int reason = callData->reason;
#endif
	int level, units, columns, fill, score, completedRows;
	Boolean fixedFill, suddenAppear, diagonal, mixed, bonus;
	Boolean cw, plainPiece, grid, showNext, sound;

	(void) strcpy(messageDsp, "");
#ifdef WINVER
	level = w->tertris.level;
	units = w->tertris.units;
	columns = w->tertris.columns;
	fill = w->tertris.fill;
	fixedFill = w->tertris.fixedFill;
	suddenAppear = w->tertris.suddenAppear;
	diagonal = w->tertris.next.diagonal;
	mixed = w->tertris.next.mixed;
	bonus = w->tertris.bonus;
	cw = w->tertris.cw;
	plainPiece = w->tertris.plainPiece;
	grid = w->tertris.grid;
	showNext = w->tertris.showNext;
	sound = w->tertris.sound;
	completedRows = w->tertris.completedRows;
	score = w->tertris.score;
#else
	XtVaGetValues(w,
		XtNlevel, &level,
		XtNunits, &units,
		XtNcolumns, &columns,
		XtNfill, &fill,
		XtNfixedFill, &fixedFill,
		XtNsuddenAppear, &suddenAppear,
		XtNdiagonal, &diagonal,
		XtNmixed, &mixed,
		XtNbonus, &bonus,
		XtNcw, &cw,
		XtNplainPiece, &plainPiece,
		XtNgrid, &grid,
		XtNsound, &sound,
		XtNcompletedRows, &completedRows,
		XtNscore, &score, NULL);
	XtVaGetValues(nextTertris,
		XtNshowNext, &showNext, NULL);
#endif
	switch (reason) {
		case GAME_HIDE:
#ifdef WINVER
			ShowWindow(w->core.hWnd, SW_SHOWMINIMIZED);
#else
			(void) XIconifyWindow(XtDisplay(topLevel),
				XtWindow(topLevel),
				XScreenNumberOfScreen(XtScreen(topLevel)));
#ifndef HAVE_MOTIF
			(void) XIconifyWindow(XtDisplay(shell),
				XtWindow(shell),
				XScreenNumberOfScreen(XtScreen(shell)));
#endif
#endif
			break;
#ifndef WINVER
		case GAME_RESTART_QUERY:
#ifdef HAVE_MOTIF
			XtManageChild(restartDialog);
#else
			XtVaSetValues(tertris,
				XtNmenu, 4, NULL); /* menu choice */
#endif
			break;
#endif
		case GAME_STATUS:
			PrintStatus(
#ifndef HAVE_MOTIF
				messageDsp, level,
#endif
				completedRows, score);
#ifdef HAVE_MOTIF
			XmScaleSetValue(levelSlider, level);
#endif
			break;
#ifndef WINVER
		case GAME_DRAWNEXT:
			XtVaSetValues(nextTertris,
				XtNrandomNumber, callData->randomNumber,
				XtNrandomRotation, callData->randomRotation,
				XtNrandomReflection, callData->randomReflection,
				XtNunits, callData->units,
				XtNdiagonal, callData->diagonal,
				XtNmixed, callData->mixed,
				XtNbonus, callData->newBonus, NULL);
			break;
#endif
		case GAME_SHOWNEXT:
			showNext = !showNext;
#ifdef WINVER
			ShowNextNextGame(&widget2);
#else
			XtVaSetValues(nextTertris,
				XtNshowNext, showNext, NULL);

#ifdef HAVE_MOTIF
			XtVaSetValues(nextTertris,
				XtNrandomNumber, GAME_CLEARNEXT, NULL);
			if (setupDialog != NULL)
				XmToggleButtonSetState(showNextSwitch,
					showNext, True);
#else
			if (showNext) {
				int x, y, width;

				XtVaGetValues(topLevel,
					XtNx, &x,
					XtNy, &y,
					XtNwidth, &width, NULL);
				XtVaSetValues(shell,
					XtNx, x + width,
					XtNy, y, NULL);
				XtRealizeWidget(shell);
				XtVaSetValues(nextTertris,
					XtNrandomNumber, GAME_CLEARNEXT,
					NULL);
			} else {
				XtVaSetValues(nextTertris,
					XtNrandomNumber, GAME_CLEARNEXT,
					NULL);
				XtUnrealizeWidget(shell);
			}
#endif
#endif
			break;
		case GAME_DECUNITS:
			units--;
#ifdef WINVER
			w->tertris.units = units;
#else
			XtVaSetValues(tertris,
				XtNunits, units, NULL);
#ifdef HAVE_MOTIF
			if (setupDialog != NULL)
				XmScaleSetValue(unitsSlider, units);
#endif
#endif
			break;
		case GAME_INCUNITS:
			units++;
#ifdef WINVER
			w->tertris.units = units;
#else
			XtVaSetValues(tertris,
				XtNunits, units, NULL);
#ifdef HAVE_MOTIF
			if (setupDialog != NULL)
				XmScaleSetValue(unitsSlider, units);
#endif
#endif
			break;
		case GAME_DECCOLUMNS:
			columns -= 2;
#ifdef WINVER
			w->tertris.columns = columns;
#else
			XtVaSetValues(tertris,
				XtNcolumns, columns, NULL);
#ifdef HAVE_MOTIF
			if (setupDialog != NULL)
				XmScaleSetValue(fieldSlider, columns);
#endif
#endif
			break;
		case GAME_INCCOLUMNS:
			columns += 2;
#ifdef WINVER
			w->tertris.columns = columns;
#else
			XtVaSetValues(tertris,
				XtNcolumns, columns, NULL);
#ifdef HAVE_MOTIF
			if (setupDialog != NULL)
				XmScaleSetValue(fieldSlider, columns);
#endif
#endif
			break;
		case GAME_FIXEDFILL:
			fixedFill = !fixedFill;
#ifdef WINVER
			w->tertris.fixedFill = fixedFill;
#else
			XtVaSetValues(tertris,
				XtNfixedFill, fixedFill, NULL);
#ifdef HAVE_MOTIF
			if (setupDialog != NULL)
				XmToggleButtonSetState(fixedFillSwitch,
					fixedFill, True);
#endif
#endif
			break;
		case GAME_SUDDENAPPEAR:
			suddenAppear = !suddenAppear;
#ifdef WINVER
			w->tertris.suddenAppear = suddenAppear;
#else
			XtVaSetValues(tertris,
				XtNsuddenAppear, suddenAppear, NULL);
#ifdef HAVE_MOTIF
			if (setupDialog != NULL)
				XmToggleButtonSetState(suddenAppearSwitch,
					suddenAppear, True);
#endif
#endif
			break;
		case GAME_DIAGONAL:
			diagonal = !diagonal;
#ifdef WINVER
			w->tertris.diagonal = diagonal;
#else
			XtVaSetValues(tertris,
				XtNdiagonal, diagonal, NULL);
#ifdef HAVE_MOTIF
			if (setupDialog != NULL)
				XmToggleButtonSetState(diagonalSwitch,
					diagonal, True);
#endif
#endif
			break;
		case GAME_MIXED:
			mixed = !mixed;
#ifdef WINVER
			w->tertris.mixed = mixed;
#else
			XtVaSetValues(tertris,
				XtNmixed, mixed, NULL);
#ifdef HAVE_MOTIF
			if (setupDialog != NULL)
				XmToggleButtonSetState(mixedSwitch,
					mixed, True);
#endif
#endif
			break;
		case GAME_BONUS:
			bonus = !bonus;
#ifdef WINVER
			w->tertris.bonus = bonus;
#else
			XtVaSetValues(tertris,
				XtNbonus, bonus, NULL);
#ifdef HAVE_MOTIF
			if (setupDialog != NULL)
				XmToggleButtonSetState(bonusSwitch,
					bonus, True);
#endif
#endif
			break;
		case GAME_CW:
			cw = !cw;
#ifdef WINVER
			w->tertris.cw = cw;
#else
			XtVaSetValues(tertris,
				XtNcw, cw, NULL);
#ifdef HAVE_MOTIF
			if (setupDialog != NULL)
				XmToggleButtonSetState(cwSwitch, cw, True);
#endif
#endif
			break;
		case GAME_PLAINPIECE:
			plainPiece = !plainPiece;
#ifdef WINVER
			w->tertris.plainPiece = plainPiece;
			widget2.tertris.plainPiece = plainPiece;
#else
			XtVaSetValues(tertris,
				XtNplainPiece, plainPiece, NULL);
			XtVaSetValues(nextTertris,
				XtNplainPiece, plainPiece, NULL);
#ifdef HAVE_MOTIF
			if (setupDialog != NULL)
				XmToggleButtonSetState(plainSwitch, plainPiece,
					True);
#endif
#endif
			break;
		case GAME_GRID:
			grid = !grid;
#ifdef WINVER
			w->tertris.grid = grid;
			widget2.tertris.grid = grid;
#else
			XtVaSetValues(tertris,
				XtNgrid, grid, NULL);
			XtVaSetValues(nextTertris,
				XtNgrid, grid, NULL);
#ifdef HAVE_MOTIF
			if (setupDialog != NULL)
				XmToggleButtonSetState(gridSwitch, grid, True);
#endif
#endif
			break;
		case GAME_SOUND:
			sound = !sound;
#ifdef WINVER
			w->tertris.sound = sound;
#else
			XtVaSetValues(tertris,
				XtNsound, sound, NULL);
#ifdef HAVE_MOTIF
			if (setupDialog != NULL)
				XmToggleButtonSetState(soundSwitch,
					sound, True);
#endif
#endif
			break;
		case GAME_FILL0:
		case GAME_FILL1:
		case GAME_FILL2:
		case GAME_FILL3:
		case GAME_FILL4:
		case GAME_FILL5:
		case GAME_FILL6:
		case GAME_FILL7:
		case GAME_FILL8:
		case GAME_FILL9:
		case GAME_FILL10:
			fill = reason - GAME_FILL0;
#ifdef WINVER
			w->tertris.fill = fill;
#else
			XtVaSetValues(tertris,
				XtNfill, fill, NULL);
#ifdef HAVE_MOTIF
			if (setupDialog != NULL)
				XmScaleSetValue(fillSlider, fill);
#endif
#endif
			break;
		case GAME_GAMEOVER:
			if (HandleGameOver(level, completedRows, score))
				(void) sprintf(messageDsp,
					"Congratulations %s!!", userNameDsp);
			else
				(void) strncpy(messageDsp, "Game Over!", MESSAGELEN);
			break;
	}
#ifdef HAVE_MOTIF
	PrintState(message, messageDsp);
#endif
}

#ifdef WINVER
void
setGameNext(TertrisWidget w, int reason,
		int randomNumber, int randomRotation, int randomReflection,
		int units, Boolean diagonal, Boolean mixed, Boolean newBonus)
{
	if (reason == GAME_DRAWNEXT) {
		ValueNextGame(&widget2,
			randomNumber, randomRotation, randomReflection,
			units, diagonal, mixed, newBonus);
	}
}
#endif

#ifdef WINVER
void
setNextGame(NextTertrisWidget w, int reason)
#else
static void
CallbackNextGame(Widget w, caddr_t clientData, tertrisCallbackStruct * callData)
#endif
{
#ifndef WINVER
	int reason = callData->reason;
#endif

	switch (reason) {
		case GAME_HIDE:
#ifdef WINVER
			ShowWindow(w->core.hWnd, SW_SHOWMINIMIZED);
#else
			(void) XIconifyWindow(XtDisplay(topLevel),
				XtWindow(topLevel),
				XScreenNumberOfScreen(XtScreen(topLevel)));
#ifndef HAVE_MOTIF
			(void) XIconifyWindow(XtDisplay(shell),
				XtWindow(shell),
				XScreenNumberOfScreen(XtScreen(shell)));
#endif
#endif
			break;
	}
}

#ifdef WINVER
static LRESULT CALLBACK
About(HWND hDlg, UINT message, WPARAM wParam, LPARAM lParam)
{
	switch (message) {
		case WM_COMMAND:
			if (LOWORD(wParam) == IDOK) {
				(void) EndDialog(hDlg, TRUE);
				return TRUE;
			}
			break;
	}
	return FALSE;
}

static LRESULT CALLBACK
WindowProc(HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam)
{
	HBRUSH brush = (HBRUSH) NULL;
	HBRUSH brush2 = (HBRUSH) NULL;
	PAINTSTRUCT paint;

	widget.core.hWnd = hWnd;
	if (GetFocus()) {
		if (!widget.tertris.focus) {
			widget.core.hDC = GetDC(hWnd);
			(void) SelectObject(widget.core.hDC, GetStockObject(NULL_BRUSH));
			EnterGame(&widget);
			(void) EndPaint(hWnd, &paint);
		}
	} else {
		if (widget.tertris.focus) {
			widget.core.hDC = GetDC(hWnd);
			(void) SelectObject(widget.core.hDC, GetStockObject(NULL_BRUSH));
			LeaveGame(&widget);
			(void) EndPaint(hWnd, &paint);
		}
	}
	switch (message) {
		case WM_CREATE:
			Initialize(&widget, brush, brush2);
			break;
		case WM_DESTROY:
			DestroyGame(brush);
			break;
		case WM_SIZE:
			ResizeGame(&widget);
			(void) InvalidateRect(hWnd, NULL, TRUE);
			break;
		case WM_PAINT:
			widget.core.hDC = BeginPaint(hWnd, &paint);
			(void) SelectObject(widget.core.hDC, GetStockObject(NULL_PEN));
			ExposeGame(&widget);
			(void) EndPaint(hWnd, &paint);
			break;
		case WM_TIMER:
			KillTimer(hWnd, 1);
			widget.core.hDC = GetDC(hWnd);
			(void) SelectObject(widget.core.hDC, GetStockObject(NULL_PEN));
			FalldownGame(&widget);
			(void) ReleaseDC(hWnd, widget.core.hDC);
			SetTimer(hWnd, 1,
				(unsigned) 7 * widget.tertris.delay, NULL);
			break;
		case WM_RBUTTONDOWN:
			widget.core.hDC = GetDC(hWnd);
			(void) SelectObject(widget.core.hDC, GetStockObject(NULL_PEN));
			(void) MoveGameInput(&widget, MOVE_RIGHT);
			(void) ReleaseDC(hWnd, widget.core.hDC);
			break;
		case WM_MBUTTONDOWN:
			widget.core.hDC = GetDC(hWnd);
			(void) SelectObject(widget.core.hDC, GetStockObject(NULL_PEN));
			(void) MoveGameInput(&widget, ROTATE);
			(void) ReleaseDC(hWnd, widget.core.hDC);
			break;
		case WM_LBUTTONDOWN:
			widget.core.hDC = GetDC(hWnd);
			(void) SelectObject(widget.core.hDC, GetStockObject(NULL_PEN));
			(void) MoveGameInput(&widget, MOVE_LEFT);
			(void) ReleaseDC(hWnd, widget.core.hDC);
			break;
#if (_WIN32_WINNT >= 0x0400) || (_WIN32_WINDOWS > 0x0400)
		case WM_MOUSEWHEEL:
			widget.core.hDC = GetDC(hWnd);
			(void) SelectObject(widget.core.hDC, GetStockObject(NULL_PEN));
			{
				int zDelta = ((short) HIWORD(wParam));

				if (zDelta > (WHEEL_DELTA >> 1)) {
					(void) MoveGameInput(&widget, REFLECT);
				} else if (zDelta < -(WHEEL_DELTA >> 1)) {
					(void) MoveGameInput(&widget, DROP);
				}
			}
			(void) ReleaseDC(hWnd, widget.core.hDC);
			break;
#endif
		case WM_COMMAND:
			switch (LOWORD(wParam)) {
				case IDM_EXIT:
					DestroyGame(brush);
					break;
				case IDM_HIDE:
					HideGame(&widget);
					break;
				case IDM_CLEAR:
					widget.core.hDC = GetDC(hWnd);
					(void) SelectObject(widget.core.hDC, GetStockObject(NULL_PEN));
					ClearGame(&widget);
					(void) ReleaseDC(hWnd, widget.core.hDC);
					break;
				case IDM_RIGHT:
				case IDM_LEFT:
				case IDM_ROTATE:
				case IDM_REFLECT:
				case IDM_DROP:
					{
						widget.core.hDC = GetDC(hWnd);
						(void) SelectObject(widget.core.hDC, GetStockObject(NULL_PEN));
						(void) MoveGameInput(&widget,
							(int) LOWORD(wParam) - IDM_DROP + 1);
						(void) ReleaseDC(hWnd, widget.core.hDC);
					}

					break;
				case IDM_INCLEVEL:
					(void) IncrementGameLevel(&widget);
					break;
				case IDM_DECLEVEL:
					(void) DecrementGameLevel(&widget);
					break;
				case IDM_INCUNITS:
					(void) IncrementGameUnits(&widget);
					break;
				case IDM_DECUNITS:
					(void) DecrementGameUnits(&widget);
					break;
				case IDM_INCCOLUMNS:
					widget.core.hDC = GetDC(hWnd);
					(void) SelectObject(widget.core.hDC, GetStockObject(NULL_PEN));
					(void) IncrementGameColumns(&widget);
					(void) ReleaseDC(hWnd, widget.core.hDC);
					break;
				case IDM_DECCOLUMNS:
					widget.core.hDC = GetDC(hWnd);
					(void) SelectObject(widget.core.hDC, GetStockObject(NULL_PEN));
					(void) DecrementGameColumns(&widget);
					(void) ReleaseDC(hWnd, widget.core.hDC);
					break;
				case IDM_FIXEDFILL:
					(void) FixedFillGame(&widget);
					break;
				case IDM_SUDDENAPPEAR:
					(void) SuddenAppearGame(&widget);
					break;
				case IDM_DIAGONAL:
					(void) DiagonalGame(&widget);
					break;
				case IDM_MIXED:
					(void) MixedGame(&widget);
					break;
				case IDM_BONUS:
					(void) BonusGame(&widget);
					break;
				case IDM_CW:
					(void) CwGame(&widget);
					break;
				case IDM_PLAINPIECE:
					widget.core.hDC = GetDC(hWnd);
					(void) SelectObject(widget.core.hDC, GetStockObject(NULL_PEN));
					(void) PlainPieceGame(&widget);
					(void) ReleaseDC(hWnd, widget.core.hDC);
					break;
				case IDM_GRID:
					widget.core.hDC = GetDC(hWnd);
					(void) SelectObject(widget.core.hDC, GetStockObject(NULL_PEN));
					(void) GridGame(&widget);
					(void) ReleaseDC(hWnd, widget.core.hDC);
					break;
				case IDM_SOUND:
					(void) SoundGame(&widget);
					break;
				case IDM_FILL0:
					widget.core.hDC = GetDC(hWnd);
					(void) SelectObject(widget.core.hDC, GetStockObject(NULL_PEN));
					(void) Fill0Game(&widget);
					(void) ReleaseDC(hWnd, widget.core.hDC);
					break;
				case IDM_FILL1:
					widget.core.hDC = GetDC(hWnd);
					(void) SelectObject(widget.core.hDC, GetStockObject(NULL_PEN));
					(void) Fill1Game(&widget);
					(void) ReleaseDC(hWnd, widget.core.hDC);
					break;
				case IDM_FILL2:
					widget.core.hDC = GetDC(hWnd);
					(void) SelectObject(widget.core.hDC, GetStockObject(NULL_PEN));
					(void) Fill2Game(&widget);
					(void) ReleaseDC(hWnd, widget.core.hDC);
					break;
				case IDM_FILL3:
					widget.core.hDC = GetDC(hWnd);
					(void) SelectObject(widget.core.hDC, GetStockObject(NULL_PEN));
					(void) Fill3Game(&widget);
					(void) ReleaseDC(hWnd, widget.core.hDC);
					break;
				case IDM_FILL4:
					widget.core.hDC = GetDC(hWnd);
					(void) SelectObject(widget.core.hDC, GetStockObject(NULL_PEN));
					(void) Fill4Game(&widget);
					(void) ReleaseDC(hWnd, widget.core.hDC);
					break;
				case IDM_FILL5:
					widget.core.hDC = GetDC(hWnd);
					(void) SelectObject(widget.core.hDC, GetStockObject(NULL_PEN));
					(void) Fill5Game(&widget);
					(void) ReleaseDC(hWnd, widget.core.hDC);
					break;
				case IDM_FILL6:
					widget.core.hDC = GetDC(hWnd);
					(void) SelectObject(widget.core.hDC, GetStockObject(NULL_PEN));
					(void) Fill6Game(&widget);
					(void) ReleaseDC(hWnd, widget.core.hDC);
					break;
				case IDM_FILL7:
					widget.core.hDC = GetDC(hWnd);
					(void) SelectObject(widget.core.hDC, GetStockObject(NULL_PEN));
					(void) Fill7Game(&widget);
					(void) ReleaseDC(hWnd, widget.core.hDC);
					break;
				case IDM_FILL8:
					widget.core.hDC = GetDC(hWnd);
					(void) SelectObject(widget.core.hDC, GetStockObject(NULL_PEN));
					(void) Fill8Game(&widget);
					(void) ReleaseDC(hWnd, widget.core.hDC);
					break;
				case IDM_FILL9:
					widget.core.hDC = GetDC(hWnd);
					(void) SelectObject(widget.core.hDC, GetStockObject(NULL_PEN));
					(void) Fill9Game(&widget);
					(void) ReleaseDC(hWnd, widget.core.hDC);
					break;
				case IDM_FILL10:
					widget.core.hDC = GetDC(hWnd);
					(void) SelectObject(widget.core.hDC, GetStockObject(NULL_PEN));
					(void) Fill10Game(&widget);
					(void) ReleaseDC(hWnd, widget.core.hDC);
					break;
				case IDM_SHOWNEXT:
					ShowNextNextGame(&widget2);
					if (widget2.tertris.showNext) {
						ShowWindow(widget2.core.hWnd, SW_SHOWNOACTIVATE);
					} else {
						ShowWindow(widget2.core.hWnd, SW_HIDE);
					}
#if 0
					if (widget2.core.hWnd == NULL) {
						widget2.core.hWnd = CreateDialog(widget.core.hInstance,
							"ShowNext", hWnd, (DLGPROC) ShowNext);
					} else {
						DestroyNextGame(brush2);
						(void) DestroyWindow(widget2.core.hWnd);
						widget2.core.hWnd = NULL;
					}
#endif
					break;
				case IDM_ABOUT:
					(void) DialogBox(widget.core.hInstance,
						"About", hWnd, (DLGPROC) About);
					break;
				case IDM_DESCRIPTION:
					(void) MessageBox(hWnd, descriptionHelp,
						"Description", MB_OK);
					break;
				case IDM_FEATURES:
					(void) MessageBox(hWnd, featuresHelp,
						"Features", MB_OK);
					break;
				case IDM_REFERENCES:
					(void) MessageBox(hWnd, referencesHelp,
						"References", MB_OK);
					break;
			}
			break;
		default:
			return (DefWindowProc(hWnd, message, wParam, lParam));
	}
	return FALSE;
}

int WINAPI
WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPSTR lpCmdLine,
		int numCmdShow)
{
	HWND hWnd;
	MSG msg;
	WNDCLASS wc;
	HACCEL hAccel;

	if (!hPrevInstance) {
		wc.style = CS_HREDRAW | CS_VREDRAW;
		wc.lpfnWndProc = WindowProc;
		wc.cbClsExtra = 0;
		wc.cbWndExtra = 0;
		wc.hInstance = hInstance;
		wc.hIcon = LoadIcon(hInstance, TITLE);
		wc.hCursor = LoadCursor((HINSTANCE) NULL, IDC_ARROW);
		wc.hbrBackground = (HBRUSH) GetStockObject(GRAY_BRUSH);
		wc.lpszMenuName = TITLE;
		wc.lpszClassName = TITLE;
		if (!RegisterClass(&wc))
			return FALSE;
	}
	widget.core.hInstance = hInstance;
	hWnd = CreateWindow(TITLE,
		TITLE,
		WS_OVERLAPPEDWINDOW,
		(signed) CW_USEDEFAULT,
		(signed) CW_USEDEFAULT,
		288,
		650,
		HWND_DESKTOP,
		(HMENU) NULL,
		hInstance,
		(void *) NULL);
	if (!hWnd)
		return FALSE;
	hAccel = (HACCEL) LoadAccelerators(hInstance, TITLE);
	(void) ShowWindow(hWnd, numCmdShow);
	SetTimer(hWnd, 1, 1000, NULL);
	(void) UpdateWindow(hWnd);
	while (GetMessage(&msg, (HWND) NULL, 0, 0))
		if (!IsDialogMessage(widget2.core.hWnd, &msg)) {
			if (!TranslateAccelerator(hWnd, hAccel, &msg)) {
				(void) TranslateMessage(&msg);
				(void) DispatchMessage(&msg);
			}
		}
	KillTimer(hWnd, 1);
	return (msg.wParam);
}

#else

static void
Usage(char * programName)
{
	unsigned int i;

	(void) fprintf(stderr, "usage: %s\n", programName);
	for (i = 0; i < strlen(optionsHelp); i++) {
		if (i == 0 || optionsHelp[i - 1] == '\n')
			(void) fprintf(stderr, "\t");
		(void) fprintf(stderr, "%c", (optionsHelp[i]));
	}
	exit(1);
}

static XrmOptionDescRec options[] =
{
	{(char *) "-mono", (char *) "*tertris.mono", XrmoptionNoArg, (char *) "TRUE"},
	{(char *) "-nomono", (char *) "*tertris.mono", XrmoptionNoArg, (char *) "FALSE"},
	{(char *) "-rv", (char *) "*tertris.reverse", XrmoptionNoArg, (char *) "TRUE"},
	{(char *) "-reverse", (char *) "*tertris.reverse", XrmoptionNoArg, (char *) "TRUE"},
	{(char *) "-norv", (char *) "*tertris.reverse", XrmoptionNoArg, (char *) "FALSE"},
	{(char *) "-noreverse", (char *) "*tertris.reverse", XrmoptionNoArg, (char *) "FALSE"},
	{(char *) "-fg", (char *) "*tertris.Foreground", XrmoptionSepArg, NULL},
	{(char *) "-foreground", (char *) "*tertris.Foreground", XrmoptionSepArg, NULL},
	{(char *) "-bg", (char *) "*Background", XrmoptionSepArg, NULL},
	{(char *) "-background", (char *) "*Background", XrmoptionSepArg, NULL},
	{(char *) "-bd", (char *) "*tertris.borderColor", XrmoptionSepArg, NULL},
	{(char *) "-border", (char *) "*tertris.borderColor", XrmoptionSepArg, NULL},
	{(char *) "-frame", (char *) "*tertris.frameColor", XrmoptionSepArg, NULL},
	{(char *) "-level", (char *) "*tertris.level", XrmoptionSepArg, NULL},
	{(char *) "-units", (char *) "*tertris.units", XrmoptionSepArg, NULL},
	{(char *) "-columns", (char *) "*tertris.columns", XrmoptionSepArg, NULL},
	{(char *) "-fill", (char *) "*tertris.fill", XrmoptionSepArg, NULL},
	{(char *) "-fixedFill", (char *) "*tertris.fixedFill", XrmoptionNoArg, (char *) "TRUE"},
	{(char *) "-nofixedFill", (char *) "*tertris.fixedFill", XrmoptionNoArg, (char *) "FALSE"},
	{(char *) "-suddenAppear", (char *) "*tertris.suddenAppear", XrmoptionNoArg, (char *) "TRUE"},
	{(char *) "-nosuddenAppear", (char *) "*tertris.suddenAppear", XrmoptionNoArg, (char *) "FALSE"},
	{(char *) "-diagonal", (char *) "*tertris.diagonal", XrmoptionNoArg, (char *) "TRUE"},
	{(char *) "-nodiagonal", (char *) "*tertris.diagonal", XrmoptionNoArg, (char *) "FALSE"},
	{(char *) "-mixed", (char *) "*tertris.mixed", XrmoptionNoArg, (char *) "TRUE"},
	{(char *) "-nomixed", (char *) "*tertris.mixed", XrmoptionNoArg, (char *) "FALSE"},
	{(char *) "-bonus", (char *) "*tertris.bonus", XrmoptionNoArg, (char *) "TRUE"},
	{(char *) "-nobonus", (char *) "*tertris.bonus", XrmoptionNoArg, (char *) "FALSE"},
	{(char *) "-cw", (char *) "*tertris.cw", XrmoptionNoArg, (char *) "TRUE"},
	{(char *) "-nocw", (char *) "*tertris.cw", XrmoptionNoArg, (char *) "FALSE"},
	{(char *) "-plain", (char *) "*tertris.plain", XrmoptionNoArg, (char *) "TRUE"},
	{(char *) "-noplain", (char *) "*tertris.plain", XrmoptionNoArg, (char *) "FALSE"},
	{(char *) "-showNext", (char *) "*tertris.showNext", XrmoptionNoArg, (char *) "TRUE"},
	{(char *) "-noshowNext", (char *) "*tertris.showNext", XrmoptionNoArg, (char *) "FALSE"},
	{(char *) "-grid", (char *) "*tertris.grid", XrmoptionNoArg, (char *) "TRUE"},
	{(char *) "-nogrid", (char *) "*tertris.grid", XrmoptionNoArg, (char *) "FALSE"},
	{(char *) "-sound", (char *) "*tertris.sound", XrmoptionNoArg, (char *) "TRUE"},
	{(char *) "-nosound", (char *) "*tertris.sound", XrmoptionNoArg, (char *) "FALSE"},
	{(char *) "-bumpSound", (char *) "*tertris.bumpSound", XrmoptionSepArg, NULL},
	{(char *) "-userName", (char *) "*tertris.userName", XrmoptionSepArg, NULL},
	{(char *) "-scoreFile", (char *) "*tertris.scoreFile", XrmoptionSepArg,
 NULL},
	{(char *) "-scores", (char *) "*tertris.scoreOnly", XrmoptionNoArg, (char *) "TRUE"},
	{(char *) "-version", (char *) "*tertris.versionOnly", XrmoptionNoArg, (char *) "TRUE"}
};

static void
playingTimer(XtPointer clientData, XtIntervalId *id)
{
	int delay;

	XtVaGetValues(tertris,
		XtNdelay, &delay, NULL);
	XtVaSetValues(tertris,
		XtNact, True, NULL);
	(void) XtAddTimeOut(7 * delay, playingTimer,
		clientData);
}

#ifdef HAVE_MOTIF
static void
CallbackGameRestart(Widget w, XtPointer clientData, XmAnyCallbackStruct * cbs)
{
	if (cbs->reason == XmCR_OK) {
		XtVaSetValues(tertris, XtNmenu, 4, NULL); /* menu choice */
	}
}

static void
LevelSlider(Widget w, XtPointer clientData, XmScaleCallbackStruct * cbs)
{
	int level = cbs->value, old;

	XtVaGetValues(tertris,
		XtNlevel, &old, NULL);
	if (old != level) {
		XtVaSetValues(tertris,
			XtNlevel, level, NULL);
	}
}

static void
UnitsSlider(Widget w, XtPointer clientData, XmScaleCallbackStruct * cbs)
{
	int units = cbs->value, old;
	Boolean diagonal, mixed, bonus;

	XtVaGetValues(tertris,
		XtNunits, &old,
		XtNdiagonal, &diagonal,
		XtNmixed, &mixed,
		XtNbonus, &bonus, NULL);
	if (old == units) {
		return;
	}
	if (LEGAL(units, diagonal, mixed, bonus)) {
		XtVaSetValues(tertris,
			XtNunits, units, NULL);
	} else if (LEGAL(units, diagonal, mixed, !bonus)) {
		/* Try toggling bonus */
		XmToggleButtonSetState(bonusSwitch, !bonus, True);
		XtVaSetValues(tertris,
			XtNunits, units,
			XtNbonus, !bonus, NULL);
	} else if (LEGAL(units, diagonal, !mixed, bonus)) {
		/* Try toggling mixed */
		XmToggleButtonSetState(mixedSwitch, !mixed, True);
		XtVaSetValues(tertris,
			XtNunits, units,
			XtNmixed, !mixed, NULL);
	} else if (LEGAL(units, !diagonal, mixed, bonus)) {
		/* Try toggling diagonal */
		XmToggleButtonSetState(diagonalSwitch, !diagonal, True);
		XtVaSetValues(tertris,
			XtNunits, units,
			XtNdiagonal, !diagonal, NULL);
	} else {
		/* Give up */
		XmScaleSetValue(unitsSlider, old);
	}
}

static void
FieldSlider(Widget w, XtPointer clientData, XmScaleCallbackStruct * cbs)
{
	int columns = cbs->value, old, odd;

	odd = 2 * ((columns + 1) / 2) - 1;
	XtVaGetValues(tertris,
		XtNcolumns, &old, NULL);
	if (old != odd) {
		XtVaSetValues(tertris,
			XtNcolumns, odd, NULL);
	}
	if (columns != odd) {
		XmScaleSetValue(fieldSlider, odd);
	}
}
static void
FillSlider(Widget w, XtPointer clientData, XmScaleCallbackStruct * cbs)
{
	int fill = cbs->value, old;

	XtVaGetValues(tertris,
		XtNfill, &old, NULL);
	if (old != fill) {
		XtVaSetValues(tertris,
			XtNfill, fill, NULL);
	}
}

static void
FixedFillToggle(Widget w, XtPointer clientData,
		XmToggleButtonCallbackStruct * cbs)
{
	Boolean fixedFill = cbs->set;

	XtVaSetValues(tertris,
		XtNfixedFill, fixedFill, NULL);
}

static void
SuddenAppearToggle(Widget w, XtPointer clientData,
		XmToggleButtonCallbackStruct * cbs)
{
	Boolean suddenAppear = cbs->set;

	XtVaSetValues(tertris,
		XtNsuddenAppear, suddenAppear, NULL);
}

static void
DiagonalToggle(Widget w, XtPointer clientData,
		XmToggleButtonCallbackStruct * cbs)
{
	int units;
	Boolean diagonal = cbs->set, mixed, bonus;

	XtVaGetValues(tertris,
		XtNunits, &units,
		XtNmixed, &mixed,
		XtNbonus, &bonus, NULL);
	if (LEGAL(units, diagonal, mixed, bonus)) {
		XtVaSetValues(tertris,
			XtNdiagonal, diagonal, NULL);
	} else if (LEGAL(units, diagonal, mixed, !bonus)) {
		/* Try toggling bonus */
		XmToggleButtonSetState(bonusSwitch, !bonus, True);
		XtVaSetValues(tertris,
			XtNdiagonal, diagonal,
			XtNbonus, !bonus, NULL);
	} else if (LEGAL(units, diagonal, !mixed, bonus)) {
		/* Try toggling mixed */
		XmToggleButtonSetState(mixedSwitch, !mixed, True);
		XtVaSetValues(tertris,
			XtNdiagonal, diagonal,
			XtNmixed, !mixed, NULL);
	} else {
		/* Give up */
		XmToggleButtonSetState(diagonalSwitch, !diagonal, True);
	}
}

static void
MixedToggle(Widget w, XtPointer clientData, XmToggleButtonCallbackStruct * cbs)
{
	int units;
	Boolean mixed = cbs->set, diagonal, bonus;

	XtVaGetValues(tertris,
		XtNunits, &units,
		XtNdiagonal, &diagonal,
		XtNbonus, &bonus, NULL);
	if (LEGAL(units, diagonal, mixed, bonus)) {
		XtVaSetValues(tertris,
			XtNmixed, mixed, NULL);
	} else if (LEGAL(units, diagonal, mixed, !bonus)) {
		/* Try toggling bonus */
		XmToggleButtonSetState(bonusSwitch, !bonus, True);
		XtVaSetValues(tertris,
			XtNmixed, mixed,
			XtNbonus, !bonus, NULL);
	} else {
		/* Give up */
		XmToggleButtonSetState(mixedSwitch, !mixed, True);
	}
}

static void
BonusToggle(Widget w, XtPointer clientData, XmToggleButtonCallbackStruct * cbs)
{
	int units;
	Boolean bonus = cbs->set, diagonal, mixed;

	XtVaGetValues(tertris,
		XtNunits, &units,
		XtNdiagonal, &diagonal,
		XtNmixed, &mixed,
		NULL);
	if (LEGAL(units, diagonal, mixed, bonus)) {
		XtVaSetValues(tertris,
			XtNbonus, bonus, NULL);
	} else {
		/* Give up */
		XmToggleButtonSetState(bonusSwitch, !bonus, True);
	}
}

static void
CwToggle(Widget w, XtPointer clientData, XmToggleButtonCallbackStruct * cbs)
{
	Boolean cw = cbs->set;

	XtVaSetValues(tertris,
		XtNcw, cw, NULL);
}

static void
PlainToggle(Widget w, XtPointer clientData, XmToggleButtonCallbackStruct * cbs)
{
	Boolean plainPiece = cbs->set;

	XtVaSetValues(tertris,
		XtNplainPiece, plainPiece, NULL);
	XtVaSetValues(nextTertris,
		XtNplainPiece, plainPiece, NULL);
}

static void
ShowNextToggle(Widget w, XtPointer clientData, XmToggleButtonCallbackStruct * cbs)
{
	Boolean showNext = cbs->set;

	XtVaSetValues(nextTertris,
		XtNshowNext, showNext, NULL);
}

static void
GridToggle(Widget w, XtPointer clientData, XmToggleButtonCallbackStruct * cbs)
{
	Boolean grid = cbs->set;

	XtVaSetValues(tertris,
		XtNgrid, grid, NULL);
	XtVaSetValues(nextTertris,
		XtNgrid, grid, NULL);
}

static void
SoundToggle(Widget w, XtPointer clientData, XmToggleButtonCallbackStruct * cbs)
{
	Boolean sound = cbs->set;

	XtVaSetValues(tertris,
		XtNsound, sound, NULL);
}

static void
fileCB(Widget w, void *value, void *clientData)
{
	int val = (int) value;

	if (val == 0)
		exit(0);
	XtVaSetValues(tertris,
		XtNmenu, val, NULL);
}

static void
closeDialog(Widget w, Widget shellDialog)
{
	XtDestroyWidget(shellDialog);
	setupDialog = NULL;

}

static Widget
createSetup(Widget w, char *title)
{
	Widget button, dialog, pane;
	Widget sliderRowCol, switchRowCol;
	int units, columns, fill;
	Boolean fixedFill, suddenAppear, diagonal, mixed;
	Boolean bonus, cw, plainPiece, showNext, grid, sound;
	char titleDsp[FILENAMELEN + 8];
	XmString titleString = NULL;

	(void) sprintf(titleDsp, "%s: %s\n", progDsp, title);
	titleString = XmStringCreateSimple((char *) titleDsp);
	XtSetArg(arg[0], XmNdeleteResponse, XmDESTROY);
	XtSetArg(arg[1], XmNdialogTitle, titleString);
	dialog = XmCreateDialogShell(w, (char *) "Setup", arg, 2);
	XmStringFree(titleString);
	pane = XtVaCreateWidget("pane", xmPanedWindowWidgetClass, dialog,
		XmNsashWidth, 1,
		XmNsashHeight, 1, NULL);
	sliderRowCol = XtVaCreateManagedWidget("sliderRowCol",
		xmRowColumnWidgetClass, pane,
		XmNnumColumns, 3,
		XmNorientation, XmHORIZONTAL,
		XmNpacking, XmPACK_COLUMN, NULL);
	switchRowCol = XtVaCreateManagedWidget("switchRowCol",
		xmRowColumnWidgetClass, pane,
		XmNnumColumns, 10,
		XmNorientation, XmHORIZONTAL,
		XmNpacking, XmPACK_COLUMN, NULL);
	XtVaGetValues(tertris,
		XtNunits, &units,
		XtNcolumns, &columns,
		XtNfill, &fill,
		XtNfixedFill, &fixedFill,
		XtNsuddenAppear, &suddenAppear,
		XtNdiagonal, &diagonal,
		XtNmixed, &mixed,
		XtNbonus, &bonus,
		XtNcw, &cw,
		XtNplainPiece, &plainPiece,
		XtNshowNext, &showNext,
		XtNgrid, &grid,
		XtNsound, &sound, NULL);
	unitsSlider = XtVaCreateManagedWidget("unitsSlider",
		xmScaleWidgetClass, sliderRowCol,
		XtVaTypedArg, XmNtitleString, XmRString, "Polyiamond Size", 15,
		XmNminimum, MIN_UNITS,
		XmNmaximum, MAX_UNITS,
		XmNvalue, units,
		XmNshowValue, True,
		XmNorientation, XmHORIZONTAL, NULL);
	XtAddCallback(unitsSlider, XmNvalueChangedCallback,
		(XtCallbackProc) UnitsSlider, (XtPointer) NULL);
	fieldSlider = XtVaCreateManagedWidget("fieldSlider",
		xmScaleWidgetClass, sliderRowCol,
		XtVaTypedArg, XmNtitleString, XmRString, "Field Size", 11,
		XmNminimum, MIN_COLUMNS + 1,
		XmNmaximum, MAX_COLUMNS - 1,
		XmNvalue, columns,
		XmNshowValue, True,
		XmNorientation, XmHORIZONTAL, NULL);
	XtAddCallback(fieldSlider, XmNvalueChangedCallback,
		(XtCallbackProc) FieldSlider, (XtPointer) NULL);
	fillSlider = XtVaCreateManagedWidget("fillSlider",
		xmScaleWidgetClass, sliderRowCol,
		XtVaTypedArg, XmNtitleString, XmRString, "Fill", 5,
		XmNminimum, 0,
		XmNmaximum, columns,
		XmNvalue, fill,
		XmNshowValue, True,
		XmNorientation, XmHORIZONTAL, NULL);
	XtAddCallback(fillSlider, XmNvalueChangedCallback,
		(XtCallbackProc) FillSlider, (XtPointer) NULL);
	fixedFillSwitch = XtVaCreateManagedWidget("Fixed Fill",
		xmToggleButtonWidgetClass, switchRowCol, NULL);
	XtAddCallback(fixedFillSwitch, XmNvalueChangedCallback,
		(XtCallbackProc) FixedFillToggle, (XtPointer) NULL);
	suddenAppearSwitch = XtVaCreateManagedWidget("Sudden Appear",
		xmToggleButtonWidgetClass, switchRowCol, NULL);
	XtAddCallback(suddenAppearSwitch, XmNvalueChangedCallback,
		(XtCallbackProc) SuddenAppearToggle, (XtPointer) NULL);
	diagonalSwitch = XtVaCreateManagedWidget("Diagonal",
		xmToggleButtonWidgetClass, switchRowCol, NULL);
	XtAddCallback(diagonalSwitch, XmNvalueChangedCallback,
		(XtCallbackProc) DiagonalToggle, (XtPointer) NULL);
	mixedSwitch = XtVaCreateManagedWidget("Mixed",
		xmToggleButtonWidgetClass, switchRowCol, NULL);
	XtAddCallback(mixedSwitch, XmNvalueChangedCallback,
		(XtCallbackProc) MixedToggle, (XtPointer) NULL);
	bonusSwitch = XtVaCreateManagedWidget("Bonus",
		xmToggleButtonWidgetClass, switchRowCol, NULL);
	XtAddCallback(bonusSwitch, XmNvalueChangedCallback,
		(XtCallbackProc) BonusToggle, (XtPointer) NULL);
	cwSwitch = XtVaCreateManagedWidget("Cw",
		xmToggleButtonWidgetClass, switchRowCol, NULL);
	XtAddCallback(cwSwitch, XmNvalueChangedCallback,
		(XtCallbackProc) CwToggle, (XtPointer) NULL);
	plainSwitch = XtVaCreateManagedWidget("Plain",
		xmToggleButtonWidgetClass, switchRowCol, NULL);
	XtAddCallback(plainSwitch, XmNvalueChangedCallback,
		(XtCallbackProc) PlainToggle, (XtPointer) NULL);
	showNextSwitch = XtVaCreateManagedWidget("ShowNext",
		xmToggleButtonWidgetClass, switchRowCol, NULL);
	XtAddCallback(showNextSwitch, XmNvalueChangedCallback,
		(XtCallbackProc) ShowNextToggle, (XtPointer) NULL);
	gridSwitch = XtVaCreateManagedWidget("Grid",
		xmToggleButtonWidgetClass, switchRowCol, NULL);
	XtAddCallback(gridSwitch, XmNvalueChangedCallback,
		(XtCallbackProc) GridToggle, (XtPointer) NULL);
	soundSwitch = XtVaCreateManagedWidget("Sound",
		xmToggleButtonWidgetClass, switchRowCol, NULL);
	XtAddCallback(soundSwitch, XmNvalueChangedCallback,
		(XtCallbackProc) SoundToggle, (XtPointer) NULL);
	XmToggleButtonSetState(fixedFillSwitch, fixedFill, True);
	XmToggleButtonSetState(suddenAppearSwitch, suddenAppear, True);
	XmToggleButtonSetState(diagonalSwitch, diagonal, True);
	XmToggleButtonSetState(mixedSwitch, mixed, True);
	XmToggleButtonSetState(bonusSwitch, bonus, True);
	XmToggleButtonSetState(cwSwitch, cw, True);
	XmToggleButtonSetState(plainSwitch, plainPiece, True);
	XmToggleButtonSetState(showNextSwitch, showNext, True);
	XmToggleButtonSetState(gridSwitch, grid, True);
	XmToggleButtonSetState(soundSwitch, sound, True);
	button = XtVaCreateManagedWidget("OK",
		xmPushButtonGadgetClass, pane,
		XmNtopAttachment, XmATTACH_FORM,
		XmNbottomAttachment, XmATTACH_FORM,
		XmNleftAttachment, XmALIGNMENT_CENTER,
		XmNshowAsDefault, True,
		XmNdefaultButtonShadowThickness, 1,
		NULL);
	XtAddCallback(button, XmNactivateCallback,
		(XtCallbackProc) closeDialog, dialog);
	XtManageChild(pane);
	return dialog;
}

static void
playCB(Widget w, void *value, void *clientData)
{
	int val = (int) value;

	if (val == 2) {
		/*XtManageChild(setupDialog); */
		if (setupDialog == NULL)
			setupDialog = createSetup(menuBar, (char *) "Setup");
		return;
	}
	XtVaSetValues(tertris,
		XtNmenu, val + 1, NULL); /* Q */
}

static Widget
createQuery(Widget w, char *text, char *title, XtCallbackProc callback)
{
	Widget button, messageBox;
	char titleDsp[FILENAMELEN + 8];
	XmString titleString = NULL, messageString = NULL;
	static XmStringCharSet charSet =
		(XmStringCharSet) XmSTRING_DEFAULT_CHARSET;

	messageString = XmStringCreateLtoR(text, charSet);
	(void) sprintf(titleDsp, "%s: %s\n", progDsp, title);
	titleString = XmStringCreateSimple((char *) titleDsp);
	XtSetArg(arg[0], XmNdialogTitle, titleString);
	XtSetArg(arg[1], XmNmessageString, messageString);
	messageBox = XmCreateWarningDialog(w, (char *) "queryBox",
		arg, 2);
	button = XmMessageBoxGetChild(messageBox, XmDIALOG_HELP_BUTTON);
	XtUnmanageChild(button);
	XmStringFree(titleString);
	XmStringFree(messageString);
	XtAddCallback(messageBox, XmNokCallback, callback, (XtPointer) NULL);
	XtAddCallback(messageBox, XmNcancelCallback, callback,
		(XtPointer) NULL);
	return messageBox;
}

static Widget
createHelp(Widget w, char *text, char *title)
{
	Widget button, messageBox;
	char titleDsp[FILENAMELEN + 8];
	XmString titleString = NULL, messageString = NULL, buttonString = NULL;
	static XmStringCharSet charSet =
		(XmStringCharSet) XmSTRING_DEFAULT_CHARSET;

	messageString = XmStringCreateLtoR(text, charSet);
	(void) sprintf(titleDsp, "%s: %s\n", progDsp, title);
	titleString = XmStringCreateSimple((char *) titleDsp);
	buttonString = XmStringCreateSimple((char *) "OK");
	XtSetArg(arg[0], XmNdialogTitle, titleString);
	XtSetArg(arg[1], XmNokLabelString, buttonString);
	XtSetArg(arg[2], XmNmessageString, messageString);
	messageBox = XmCreateInformationDialog(w, (char *) "helpBox",
		arg, 3);
	button = XmMessageBoxGetChild(messageBox, XmDIALOG_CANCEL_BUTTON);
	XtUnmanageChild(button);
	button = XmMessageBoxGetChild(messageBox, XmDIALOG_HELP_BUTTON);
	XtUnmanageChild(button);
	XmStringFree(titleString);
	XmStringFree(buttonString);
	XmStringFree(messageString);
	return messageBox;
}

static void
helpCB(Widget w, XtPointer value, XtPointer clientData)
{
	int val = (int) value;

	switch (val) {
	case 0:
		XtManageChild(descriptionDialog);
		break;
	case 1:
		XtManageChild(featuresDialog);
		break;
	case 2:
		XtManageChild(optionsDialog);
		break;
	case 3:
		XtManageChild(referencesDialog);
		break;
	case 4:
		XtManageChild(aboutDialog);
		break;
	default:
		{
			char *buf;

			intCat(&buf, "helpCB: %d\n", val);
			XtWarning(buf);
			free(buf);
		}
	}
}
#endif

int
main(int argc, char **argv)
{
#ifdef HAVE_MOTIF
	Widget pullDownMenu, widget;
	Widget menuBarPanel, mainPanel, controlPanel;
	Widget scoreRowCol, levelRowCol;
	Widget messageRowCol;
	XmString fileString, playString;
	XmString quitString;
	XmString pauseString, clearString, setupString;
#if 0
	XmString incrementString, decrementString;
	XmString fixedFillString, suddenAppearString;
	XmString diagonalString, mixedString, bonusString;
	XmString cwString, plainString, showNextString;
#endif
#else
	Boolean showNext;
	int x, y, width;
#endif

	progDsp = argv[0];
	topLevel = XtInitialize(argv[0], "Tertris",
		options, XtNumber(options), &argc, argv);
	if (argc != 1)
		Usage(argv[0]);

#if HAVE_XPM
	{
		XpmAttributes xpmAttributes;
		XpmColorSymbol transparentColor[1] = {{NULL,
			(char *) "none", 0 }};
		Pixel bg;

		xpmAttributes.valuemask = XpmColorSymbols | XpmCloseness;
		xpmAttributes.colorsymbols = transparentColor;
		xpmAttributes.numsymbols = 1;
		xpmAttributes.closeness = 40000;
		XtVaGetValues(topLevel, XtNbackground, &bg, NULL);
		transparentColor[0].pixel = bg;
		(void) XpmCreatePixmapFromData(XtDisplay(topLevel),
			RootWindowOfScreen(XtScreen(topLevel)),
			(char **) tertris_xpm, &tertrisIcon, NULL,
			&xpmAttributes);
	}
	if (tertrisIcon == (Pixmap) NULL)
#endif
		tertrisIcon = XCreateBitmapFromData(XtDisplay(topLevel),
			RootWindowOfScreen(XtScreen(topLevel)),
			(char *) tertris_bits,
			tertris_width, tertris_height);
#ifdef HAVE_MOTIF
	XtVaSetValues(topLevel,
		XtNiconPixmap, tertrisIcon,
		XmNkeyboardFocusPolicy, XmPOINTER, NULL); /* not XmEXPLICIT */
	menuBarPanel = XtVaCreateManagedWidget("menuBarPanel",
		xmPanedWindowWidgetClass, topLevel,
		XmNseparatorOn, False,
		XmNsashWidth, 1,
		XmNsashHeight, 1, NULL);
	fileString = XmStringCreateSimple((char *) "File");
	playString = XmStringCreateSimple((char *) "Play");
	menuBar = XmVaCreateSimpleMenuBar(menuBarPanel, (char *) "menuBar",
		XmVaCASCADEBUTTON, fileString, 'F',
		XmVaCASCADEBUTTON, playString, 'P',
		NULL);
	XmStringFree(fileString);
	XmStringFree(playString);
	quitString = XmStringCreateSimple((char *) "Quit");
	XmVaCreateSimplePulldownMenu(menuBar, (char *) "file_menu", 0, fileCB,
		XmVaSEPARATOR,
		XmVaPUSHBUTTON, quitString, 'Q', NULL, NULL,
		NULL);
	XmStringFree(quitString);
	pauseString = XmStringCreateSimple((char *) "Pause");
	clearString = XmStringCreateSimple((char *) "Clear");
	setupString = XmStringCreateSimple((char *) "Setup");
#if 0
	incrementString = XmStringCreateSimple((char *) "Increment");
	decrementString = XmStringCreateSimple((char *) "Decrement");
	fixedFillString = XmStringCreateSimple((char *) "FiXed Fill");
	suddenAppearString = XmStringCreateSimple((char *) "Sudden Appear");
	diagonalString = XmStringCreateSimple((char *) "DiaGonal");
	mixedString = XmStringCreateSimple((char *) "Mixed");
	bonusString = XmStringCreateSimple((char *) "Bonus");
	cwString = XmStringCreateSimple((char *) "Cw");
	plainString = XmStringCreateSimple((char *) "PLain");
	showNextString = XmStringCreateSimple((char *) "Show Next");
#endif
	XmVaCreateSimplePulldownMenu(menuBar, (char *) "play_menu", 1, playCB,
		XmVaPUSHBUTTON, pauseString, 'P', NULL, NULL,
		XmVaPUSHBUTTON, clearString, 'C', NULL, NULL,
		XmVaPUSHBUTTON, setupString, 'S', NULL, NULL,
#if 0
		XmVaPUSHBUTTON, incrementString, 'I', NULL, NULL,
		XmVaPUSHBUTTON, decrementString, 'D', NULL, NULL,
		XmVaPUSHBUTTON, fixedFillString, 'X', NULL, NULL,
		XmVaPUSHBUTTON, suddenAppearString, 'A', NULL, NULL,
		XmVaPUSHBUTTON, diagonalString, 'G', NULL, NULL,
		XmVaPUSHBUTTON, mixedString, 'M', NULL, NULL,
		XmVaPUSHBUTTON, bonusString, 'B', NULL, NULL,
		XmVaPUSHBUTTON, cwString, 'C', NULL, NULL,
		XmVaPUSHBUTTON, plainString, 'L', NULL, NULL,
		XmVaPUSHBUTTON, showNextString, 'N', NULL, NULL,
#endif
		NULL);
	XmStringFree(pauseString);
	XmStringFree(clearString);
	XmStringFree(setupString);
#if 0
	XmStringFree(incrementString);
	XmStringFree(decrementString);
	XmStringFree(fixedFillString);
	XmStringFree(suddenAppearString);
	XmStringFree(diagonalString);
	XmStringFree(mixedString);
	XmStringFree(bonusString);
	XmStringFree(cwString);
	XmStringFree(plainString);
	XmStringFree(showNextString);
#endif
	pullDownMenu = XmCreatePulldownMenu(menuBar,
		(char *) "helpPullDown", NULL, 0);
	widget = XtVaCreateManagedWidget("Help",
		xmCascadeButtonWidgetClass, menuBar,
		XmNsubMenuId, pullDownMenu,
		XmNmnemonic, 'H', NULL);
	XtVaSetValues(menuBar, XmNmenuHelpWidget, widget, NULL);
	widget = XtVaCreateManagedWidget("Description",
		xmPushButtonGadgetClass, pullDownMenu,
		XmNmnemonic, 'D', NULL);
		XtAddCallback(widget, XmNactivateCallback, helpCB, (char *) 0);
	widget = XtVaCreateManagedWidget("Features",
		xmPushButtonGadgetClass, pullDownMenu,
		XmNmnemonic, 'F', NULL);
	XtAddCallback(widget, XmNactivateCallback, helpCB, (char *) 1);
	widget = XtVaCreateManagedWidget("Options",
		xmPushButtonGadgetClass, pullDownMenu,
		XmNmnemonic, 'O', NULL);
	XtAddCallback(widget, XmNactivateCallback, helpCB, (char *) 2);
	widget = XtVaCreateManagedWidget("References",
		xmPushButtonGadgetClass, pullDownMenu,
		XmNmnemonic, 'R', NULL);
	XtAddCallback(widget, XmNactivateCallback, helpCB, (char *) 3);
	widget = XtVaCreateManagedWidget("About",
		xmPushButtonGadgetClass, pullDownMenu,
		XmNmnemonic, 'A', NULL);
	XtAddCallback(widget, XmNactivateCallback, helpCB, (char *) 4);
	XtManageChild(menuBar);
	descriptionDialog = createHelp(menuBar, (char *) descriptionHelp,
		(char *) "Description");
	featuresDialog = createHelp(menuBar, (char *) featuresHelp,
		(char *) "Features");
	optionsDialog = createHelp(menuBar, (char *) optionsHelp,
		(char *) "Options");
	referencesDialog = createHelp(menuBar, (char *) referencesHelp,
		(char *) "References");
	aboutDialog = createHelp(menuBar, (char *) aboutHelp,
		(char *) "About");
	restartDialog = createQuery(topLevel,
		(char *) "Are you sure you want to restart?",
		(char *) "Reset Query",
		(XtCallbackProc) CallbackGameRestart);
	mainPanel = XtVaCreateManagedWidget("mainPanel",
		xmPanedWindowWidgetClass, menuBarPanel,
		XmNorientation, XmHORIZONTAL,
		NULL);
	controlPanel = XtVaCreateManagedWidget("controlPanel",
		xmPanedWindowWidgetClass, mainPanel,
		XmNseparatorOn, False,
		XmNsashWidth, 1,
		XmNsashHeight, 1, NULL);
	levelRowCol = XtVaCreateManagedWidget("levelRowCol",
		xmRowColumnWidgetClass, controlPanel,
		XmNnumColumns, 1,
		XmNorientation, XmHORIZONTAL,
		NULL);
	scoreRowCol = XtVaCreateManagedWidget("Rowcol",
		xmRowColumnWidgetClass, controlPanel,
		XmNnumColumns, 3,
		XmNorientation, XmHORIZONTAL,
		XmNpacking, XmPACK_COLUMN, NULL);
	levelSlider = XtVaCreateManagedWidget("levelSlider",
		xmScaleWidgetClass, levelRowCol,
		XtVaTypedArg, XmNtitleString, XmRString, "Level", 6,
		XmNminimum, MIN_LEVELS,
		XmNmaximum, MAX_LEVELS,
		XmNvalue, MIN_LEVELS,
		XmNshowValue, True,
		XmNorientation, XmHORIZONTAL, NULL);
	XtAddCallback(levelSlider, XmNvalueChangedCallback,
		(XtCallbackProc) LevelSlider, (XtPointer) NULL);
	(void) XtVaCreateManagedWidget("rowsText",
		xmLabelGadgetClass, scoreRowCol,
		XtVaTypedArg, XmNlabelString, XmRString, "Rows", 5, NULL);
	rowsText = XtVaCreateManagedWidget("0",
		xmLabelWidgetClass, scoreRowCol, NULL);
	(void) XtVaCreateManagedWidget("scoreText",
		xmLabelGadgetClass, scoreRowCol,
		XtVaTypedArg, XmNlabelString, XmRString, "Score", 6, NULL);
	scoreText = XtVaCreateManagedWidget("0",
		xmLabelWidgetClass, scoreRowCol, NULL);
	(void) XtVaCreateManagedWidget("maxScoreText",
		xmLabelGadgetClass, scoreRowCol,
		XtVaTypedArg, XmNlabelString, XmRString, "Max Score", 10, NULL);
	maxScoreText = XtVaCreateManagedWidget("0",
		xmLabelWidgetClass, scoreRowCol, NULL);
	nextTertris = XtCreateManagedWidget("tertrisNext",
		nextTertrisWidgetClass, controlPanel, NULL, 0);
	messageRowCol = XtVaCreateManagedWidget("messageRowCol",
		xmRowColumnWidgetClass, controlPanel, NULL);
	message = XtVaCreateManagedWidget("Play Tertris!",
		xmLabelWidgetClass, messageRowCol, NULL);
	tertris = XtCreateManagedWidget("tertris",
		tertrisWidgetClass, mainPanel, NULL, 0);
#else
	tertris = XtCreateManagedWidget("tertris",
		tertrisWidgetClass, topLevel, NULL, 0);
	XtVaSetValues(topLevel,
		XtNiconPixmap, tertrisIcon,
		XtNinput, True, NULL);
	shell = XtCreateApplicationShell(argv[0],
		topLevelShellWidgetClass, NULL, 0);
	XtVaSetValues(shell,
		XtNiconPixmap, tertrisIcon,
		XtNinput, True, NULL);
	nextTertris = XtCreateManagedWidget("nextTertris",
		nextTertrisWidgetClass, shell, NULL, 0);
#endif
	XtAddCallback(tertris, XtNselectCallback,
		(XtCallbackProc) CallbackGame, (XtPointer) NULL);
	XtAddCallback(nextTertris, XtNselectCallback,
		(XtCallbackProc) CallbackNextGame, (XtPointer) NULL);
	Initialize(tertris);
	XtRealizeWidget(topLevel);
	VOID XGrabButton(XtDisplay(tertris), (unsigned int) AnyButton,
		AnyModifier, XtWindow(tertris), TRUE,
		(unsigned int) (ButtonPressMask),
		GrabModeAsync, GrabModeAsync, XtWindow(tertris),
		XCreateFontCursor(XtDisplay(tertris), XC_crosshair));
#ifndef HAVE_MOTIF
	XtVaGetValues(topLevel,
		XtNx, &x,
		XtNy, &y,
		XtNwidth, &width, NULL);
	XtVaSetValues(shell,
		XtNx, x + width,
		XtNy, y, NULL);
	XtVaGetValues(tertris,
		XtNshowNext, &showNext, NULL);
	if (showNext)
		XtRealizeWidget(shell);
#endif
	(void) XtAddTimeOut((unsigned long) 1000,
		playingTimer, (XtPointer) tertris);
	XtMainLoop();

#ifdef VMS
	return 1;
#else
	return 0;
#endif
}
#endif
