/*
# X-BASED TETRIS
#
#  Tetris.c
#
###
#
#  Copyright (c) 1993 - 2006	David Albert Bagley, bagleyd@tux.org
#
#  Taken from GENERIC X-BASED TETRIS
#
#  Copyright (c) 1992 - 1995	Q. Alex Zhao, azhao@cc.gatech.edu
#
#			All Rights Reserved
#
#  Permission to use, copy, modify, and distribute this software and
#  its documentation for any purpose and without fee is hereby granted,
#  provided that the above copyright notice appear in all copies and
#  that both that copyright notice and this permission notice appear in
#  supporting documentation, and that the name of the author not be
#  used in advertising or publicity pertaining to distribution of the
#  software without specific, written prior permission.
#
#  This program is distributed in the hope that it will be "playable",
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
*/

/* Methods file for Tetris */

#include "file.h"
#include "rngs.h"
#include "sound.h"
#include "TetrisP.h"

#ifdef WINVER
#ifndef LOGPATH
#define LOGPATH "/usr/tmp"
#endif
#ifndef INIFILE
#define INIFILE "wtetris.ini"
#endif

#define SECTION "setup"
#else

#ifndef LOGPATH
#ifdef VMS
#define LOGPATH "SYS$SCRATCH:"
#else
#define LOGPATH "/usr/tmp"
#endif
#endif

static void ResizeGame(TetrisWidget w);
static void SizeGame(TetrisWidget w);
static Boolean SetValuesGame(Widget current, Widget request, Widget renew);
static void DestroyGame(Widget old);
static void QuitGame(TetrisWidget w, XEvent *event, char **args, int nArgs);
static void InitializeGame(Widget request, Widget renew);
static void ExposeGame(Widget renew, XEvent *event, Region region);
static void HideGame(TetrisWidget w, XEvent *event, char **args, int nArgs);
static void ClearGame(TetrisWidget w, XEvent *event, char **args, int nArgs);
static void EnterGame(TetrisWidget w, XEvent *event, char **args, int nArgs);
static void LeaveGame(TetrisWidget w, XEvent *event, char **args, int nArgs);
static void ReflectGame(TetrisWidget w, XEvent *event, char **args, int nArgs);
static void RotateGame(TetrisWidget w, XEvent *event, char **args, int nArgs);
static void MoveGameLeft(TetrisWidget w,
	XEvent *event, char **args, int nArgs);
static void MoveGameRight(TetrisWidget w,
	XEvent *event, char **args, int nArgs);
static void DropGame(TetrisWidget w, XEvent *event, char **args, int nArgs);
static void MoveGameInput(TetrisWidget w, int direction);
static void IncrementGameLevel(TetrisWidget w,
	XEvent *event, char **args, int nArgs);
static void DecrementGameLevel(TetrisWidget w,
	XEvent *event, char **args, int nArgs);
static void IncrementGameUnits(TetrisWidget w,
	XEvent *event, char **args, int nArgs);
static void DecrementGameUnits(TetrisWidget w,
	XEvent *event, char **args, int nArgs);
static void IncrementGameColumns(TetrisWidget w,
	XEvent *event, char **args, int nArgs);
static void DecrementGameColumns(TetrisWidget w,
	XEvent *event, char **args, int nArgs);
static void FixedFillGame(TetrisWidget w,
	XEvent *event, char **args, int nArgs);
static void SuddenAppearGame(TetrisWidget w,
	XEvent *event, char **args, int nArgs);
static void DiagonalGame(TetrisWidget w,
	XEvent *event, char **args, int nArgs);
static void MixedGame(TetrisWidget w, XEvent *event, char **args, int nArgs);
static void BonusGame(TetrisWidget w, XEvent *event, char **args, int nArgs);
static void CwGame(TetrisWidget w, XEvent *event, char **args, int nArgs);
static void PlainPieceGame(TetrisWidget w,
	XEvent *event, char **args, int nArgs);
static void ShowNextGame(TetrisWidget w,
	XEvent *event, char **args, int nArgs);
static void SoundGame(TetrisWidget w, XEvent *event, char **args, int nArgs);
static void PauseGame(TetrisWidget w, XEvent *event, char **args, int nArgs);
static void Fill0Game(TetrisWidget w, XEvent *event, char **args, int nArgs);
static void Fill1Game(TetrisWidget w, XEvent *event, char **args, int nArgs);
static void Fill2Game(TetrisWidget w, XEvent *event, char **args, int nArgs);
static void Fill3Game(TetrisWidget w, XEvent *event, char **args, int nArgs);
static void Fill4Game(TetrisWidget w, XEvent *event, char **args, int nArgs);
static void Fill5Game(TetrisWidget w, XEvent *event, char **args, int nArgs);
static void Fill6Game(TetrisWidget w, XEvent *event, char **args, int nArgs);
static void Fill7Game(TetrisWidget w, XEvent *event, char **args, int nArgs);
static void Fill8Game(TetrisWidget w, XEvent *event, char **args, int nArgs);
static void Fill9Game(TetrisWidget w, XEvent *event, char **args, int nArgs);
static void Fill10Game(TetrisWidget w, XEvent *event, char **args, int nArgs);

static char defaultTranslationsGame[] =
"<KeyPress>q: Quit()\n\
 Ctrl<KeyPress>C: Quit()\n\
 <KeyPress>osfCancel: Hide()\n\
 <KeyPress>Escape: Hide()\n\
 <KeyPress>osfEscape: Hide()\n\
 Ctrl<KeyPress>[: Hide()\n\
 <KeyPress>0x1B: Hide()\n\
 <Btn4Down>: Reflect()\n\
 <KeyPress>i: Reflect()\n\
 <KeyPress>e: Reflect()\n\
 <KeyPress>KP_Divide: Reflect()\n\
 <KeyPress>R5: Reflect()\n\
 <Btn2Down>: Rotate()\n\
 <KeyPress>k: Rotate()\n\
 <KeyPress>d: Rotate()\n\
 <KeyPress>Up: Rotate()\n\
 <KeyPress>osfUp: Rotate()\n\
 <KeyPress>KP_Up: Rotate()\n\
 <KeyPress>R8: Rotate()\n\
 <KeyPress>KP_8: Rotate()\n\
 <KeyPress>Begin: Rotate()\n\
 <KeyPress>KP_5: Rotate()\n\
 <KeyPress>R11: Rotate()\n\
 <Btn1Down>: MoveLeft()\n\
 <KeyPress>j: MoveLeft()\n\
 <KeyPress>s: MoveLeft()\n\
 <KeyPress>Left: MoveLeft()\n\
 <KeyPress>osfLeft: MoveLeft()\n\
 <KeyPress>KP_Left: MoveLeft()\n\
 <KeyPress>KP_4: MoveLeft()\n\
 <KeyPress>R10: MoveLeft()\n\
 <Btn3Down>: MoveRight()\n\
 <KeyPress>l: MoveRight()\n\
 <KeyPress>f: MoveRight()\n\
 <KeyPress>Right: MoveRight()\n\
 <KeyPress>osfRight: MoveRight()\n\
 <KeyPress>KP_Right: MoveRight()\n\
 <KeyPress>KP_6: MoveRight()\n\
 <KeyPress>R12: MoveRight()\n\
 <Btn5Down>: Drop()\n\
 <KeyPress>Return: Drop()\n\
 <KeyPress>KP_Enter: Drop()\n\
 <KeyPress>Down: Drop()\n\
 <KeyPress>osfDown: Drop()\n\
 <KeyPress>KP_Down: Drop()\n\
 <KeyPress>KP_2: Drop()\n\
 <KeyPress>R14: Drop()\n\
 <KeyPress>KP_Space: Drop()\n\
 <KeyPress>space: Drop()\n\
 <Btn4Down>: Rotate()\n\
 <Btn5Down>: Drop()\n\
 <KeyPress>c: Clear()\n\
 <KeyPress>>: IncrementLevel()\n\
 <KeyPress>.: IncrementLevel()\n\
 <KeyPress><: DecrementLevel()\n\
 <KeyPress>+: IncrementUnits()\n\
 <KeyPress>=: IncrementUnits()\n\
 <KeyPress>-: DecrementUnits()\n\
 <KeyPress>_: DecrementUnits()\n\
 <KeyPress>]: IncrementColumns()\n\
 <KeyPress>}: IncrementColumns()\n\
 <KeyPress>[: DecrementColumns()\n\
 <KeyPress>{: DecrementColumns()\n\
 <KeyPress>x: FixedFill()\n\
 <KeyPress>r: SuddenAppear()\n\
 <KeyPress>g: Diagonal()\n\
 <KeyPress>m: Mixed()\n\
 <KeyPress>b: Bonus()\n\
 <KeyPress>w: Cw()\n\
 <KeyPress>p: PlainPiece()\n\
 <KeyPress>n: ShowNext()\n\
 Shift<KeyPress>2: Sound()\n\
 <KeyPress>|: Pause()\n\
 <KeyPress>`: Fill0()\n\
 <KeyPress>1: Fill1()\n\
 <KeyPress>2: Fill2()\n\
 <KeyPress>3: Fill3()\n\
 <KeyPress>4: Fill4()\n\
 <KeyPress>5: Fill5()\n\
 <KeyPress>6: Fill6()\n\
 <KeyPress>7: Fill7()\n\
 <KeyPress>8: Fill8()\n\
 <KeyPress>9: Fill9()\n\
 <KeyPress>0: Fill10()\n\
 <EnterWindow>: Enter()\n\
 <LeaveWindow>: Leave()";

/*
 <KeyPress>,: DecrementLevel()\n\
#if defined(UP_REFLECT)
 <KeyPress>KP_Up: Reflect()\n\
 <KeyPress>R8: Reflect()\n\
 <KeyPress>KP_8: Reflect()\n\
#endif
*/

static XtActionsRec actionsListGame[] =
{
	{(char *) "Quit", (XtActionProc) QuitGame},
	{(char *) "Hide", (XtActionProc) HideGame},
	{(char *) "Reflect", (XtActionProc) ReflectGame},
	{(char *) "Rotate", (XtActionProc) RotateGame},
	{(char *) "MoveLeft", (XtActionProc) MoveGameLeft},
	{(char *) "MoveRight", (XtActionProc) MoveGameRight},
	{(char *) "Drop", (XtActionProc) DropGame},
	{(char *) "Clear", (XtActionProc) ClearGame},
	{(char *) "IncrementLevel", (XtActionProc) IncrementGameLevel},
	{(char *) "DecrementLevel", (XtActionProc) DecrementGameLevel},
	{(char *) "IncrementUnits", (XtActionProc) IncrementGameUnits},
	{(char *) "DecrementUnits", (XtActionProc) DecrementGameUnits},
	{(char *) "IncrementColumns", (XtActionProc) IncrementGameColumns},
	{(char *) "DecrementColumns", (XtActionProc) DecrementGameColumns},
	{(char *) "FixedFill", (XtActionProc) FixedFillGame},
	{(char *) "SuddenAppear", (XtActionProc) SuddenAppearGame},
	{(char *) "Diagonal", (XtActionProc) DiagonalGame},
	{(char *) "Mixed", (XtActionProc) MixedGame},
	{(char *) "Bonus", (XtActionProc) BonusGame},
	{(char *) "Cw", (XtActionProc) CwGame},
	{(char *) "PlainPiece", (XtActionProc) PlainPieceGame},
	{(char *) "ShowNext", (XtActionProc) ShowNextGame},
	{(char *) "Sound", (XtActionProc) SoundGame},
	{(char *) "Pause", (XtActionProc) PauseGame},
	{(char *) "Fill0", (XtActionProc) Fill0Game},
	{(char *) "Fill1", (XtActionProc) Fill1Game},
	{(char *) "Fill2", (XtActionProc) Fill2Game},
	{(char *) "Fill3", (XtActionProc) Fill3Game},
	{(char *) "Fill4", (XtActionProc) Fill4Game},
	{(char *) "Fill5", (XtActionProc) Fill5Game},
	{(char *) "Fill6", (XtActionProc) Fill6Game},
	{(char *) "Fill7", (XtActionProc) Fill7Game},
	{(char *) "Fill8", (XtActionProc) Fill8Game},
	{(char *) "Fill9", (XtActionProc) Fill9Game},
	{(char *) "Fill10", (XtActionProc) Fill10Game},
	{(char *) "Enter", (XtActionProc) EnterGame},
	{(char *) "Leave", (XtActionProc) LeaveGame}
};

static XtResource resourcesGame[] =
{
	/* 30*10+2,24*10+2 */
	{XtNwidth, XtCWidth, XtRDimension, sizeof (Dimension),
	 XtOffset(TetrisWidget, core.width),
	 XtRString, (caddr_t) "302"},
	{XtNheight, XtCHeight, XtRDimension, sizeof (Dimension),
	 XtOffset(TetrisWidget, core.height),
	 XtRString, (caddr_t) "602"},
	{XtNmono, XtCMono, XtRBoolean, sizeof (Boolean),
	 XtOffset(TetrisWidget, tetris.mono),
	 XtRString, (caddr_t) "FALSE"},
	{XtNreverse, XtCReverse, XtRBoolean, sizeof (Boolean),
	 XtOffset(TetrisWidget, tetris.reverse),
	 XtRString, (caddr_t) "FALSE"},
	{XtNforeground, XtCForeground, XtRPixel, sizeof (Pixel),
	 XtOffset(TetrisWidget, tetris.foreground),
	 XtRString, (caddr_t) XtDefaultForeground},
	{XtNbackground, XtCBackground, XtRPixel, sizeof (Pixel),
	 XtOffset(TetrisWidget, tetris.background),
	 XtRString, (caddr_t) XtDefaultBackground},
	{XtNframeColor, XtCColor, XtRPixel, sizeof (Pixel),
	 XtOffset(TetrisWidget, tetris.frameColor),
	 XtRString, (caddr_t) "cyan" /*XtDefaultForeground*/},
	{XtNborderColor, XtCColor, XtRPixel, sizeof (Pixel),
	 XtOffset(TetrisWidget, tetris.borderColor),
	 XtRString, (caddr_t) "gray25" /*XtDefaultForeground*/},
	{XtNdelay, XtCDelay, XtRInt, sizeof (int),
	 XtOffset(TetrisWidget, tetris.delay),
	 XtRString, (caddr_t) "1000"},
	{XtNsound, XtCSound, XtRBoolean, sizeof (Boolean),
	 XtOffset(TetrisWidget, tetris.sound),
	 XtRString, (caddr_t) "FALSE"},
	{XtNbumpSound, XtCBumpSound, XtRString, sizeof (String),
	 XtOffset(TetrisWidget, tetris.bumpSound),
	 XtRString, (caddr_t) BUMPSOUND},
	{XtNcolumns, XtCColumns, XtRInt, sizeof (int),
	 XtOffset(TetrisWidget, tetris.columns),
	 XtRString, (caddr_t) "10"},
	{XtNlevel, XtCLevel, XtRInt, sizeof (int),
	 XtOffset(TetrisWidget, tetris.level),
	 XtRString, (caddr_t) "3"},
	{XtNunits, XtCUnits, XtRInt, sizeof (int),
	 XtOffset(TetrisWidget, tetris.units),
	 XtRString, (caddr_t) "4"},
	{XtNfill, XtCFill, XtRInt, sizeof (int),
	 XtOffset(TetrisWidget, tetris.fill),
	 XtRString, (caddr_t) "0"},
	{XtNfixedFill, XtCFixedFill, XtRBoolean, sizeof (Boolean),
	 XtOffset(TetrisWidget, tetris.fixedFill),
	 XtRString, (caddr_t) "TRUE"},
	{XtNsuddenAppear, XtCSuddenAppear, XtRBoolean, sizeof (Boolean),
	 XtOffset(TetrisWidget, tetris.suddenAppear),
	 XtRString, (caddr_t) "FALSE"},
	{XtNdiagonal, XtCDiagonal, XtRBoolean, sizeof (Boolean),
	 XtOffset(TetrisWidget, tetris.diagonal),
	 XtRString, (caddr_t) "FALSE"},
	{XtNmixed, XtCMixed, XtRBoolean, sizeof (Boolean),
	 XtOffset(TetrisWidget, tetris.mixed),
	 XtRString, (caddr_t) "FALSE"},
	{XtNbonus, XtCBonus, XtRBoolean, sizeof (Boolean),
	 XtOffset(TetrisWidget, tetris.bonus),
	 XtRString, (caddr_t) "TRUE"},
	{XtNcw, XtCCw, XtRBoolean, sizeof (Boolean),
	 XtOffset(TetrisWidget, tetris.cw),
	 XtRString, (caddr_t) "TRUE"},
	{XtNplainPiece, XtCPlainPiece, XtRBoolean, sizeof (Boolean),
	 XtOffset(TetrisWidget, tetris.plainPiece),
	 XtRString, (caddr_t) "FALSE"},
	{XtNshowNext, XtCShowNext, XtRBoolean, sizeof (Boolean),
	 XtOffset(TetrisWidget, tetris.showNext),
	 XtRString, (caddr_t) "TRUE"},
	{XtNscore, XtCBoolean, XtRInt, sizeof (int),
	 XtOffset(TetrisWidget, tetris.score),
	 XtRString, (caddr_t) "0"},
	{XtNcompletedRows, XtCBoolean, XtRInt, sizeof (int),
	 XtOffset(TetrisWidget, tetris.completedRows),
	 XtRString, (caddr_t) "0"},
	{XtNuserName, XtCUserName, XtRString, sizeof (String),
	 XtOffset(TetrisWidget, tetris.userName),
	 XtRString, (caddr_t) ""},
	{XtNscoreFile, XtCScoreFile, XtRString, sizeof (String),
	 XtOffset(TetrisWidget, tetris.scoreFile),
	 XtRString, (caddr_t) ""},
	{XtNscoreOnly, XtCBoolean, XtRBoolean, sizeof (Boolean),
	 XtOffset(TetrisWidget, tetris.scoreOnly),
	 XtRString, (caddr_t) "FALSE"},
	{XtNversionOnly, XtCBoolean, XtRBoolean, sizeof (Boolean),
	 XtOffset(TetrisWidget, tetris.versionOnly),
	 XtRString, (caddr_t) "FALSE"},
	{XtNstart, XtCBoolean, XtRBoolean, sizeof (Boolean),
	 XtOffset(TetrisWidget, tetris.started),
	 XtRString, (caddr_t) "FALSE"},
	{XtNact, XtCBoolean, XtRBoolean, sizeof (Boolean),
	 XtOffset(TetrisWidget, tetris.act),
	 XtRString, (caddr_t) "FALSE"},
	{XtNmenu, XtCMenu, XtRInt, sizeof (int),
	 XtOffset(TetrisWidget, tetris.menu),
	 XtRString, (caddr_t) "-1"},
	{XtNselectCallback, XtCCallback, XtRCallback, sizeof (caddr_t),
	 XtOffset(TetrisWidget, tetris.select),
	 XtRCallback, (caddr_t) NULL}
};

TetrisClassRec tetrisClassRec =
{
	{
		(WidgetClass) & widgetClassRec,		/* superclass */
		(char *) "Tetris",	/* class name */
		sizeof (TetrisRec),	/* widget size */
		NULL,		/* class initialize */
		NULL,		/* class part initialize */
		FALSE,		/* class inited */
		(XtInitProc) InitializeGame,	/* initialize */
		NULL,		/* initialize hook */
		XtInheritRealize,	/* realize */
		actionsListGame,	/* actions */
		XtNumber(actionsListGame),	/* num actions */
		resourcesGame,	/* resources */
		XtNumber(resourcesGame),	/* num resources */
		NULLQUARK,	/* xrm class */
		TRUE,		/* compress motion */
		TRUE,		/* compress exposure */
		TRUE,		/* compress enterleave */
		TRUE,		/* visible interest */
		(XtWidgetProc) DestroyGame,	/* destroy */
		(XtWidgetProc) ResizeGame,	/* resize */
		(XtExposeProc) ExposeGame,	/* expose */
		(XtSetValuesFunc) SetValuesGame,	/* set values */
		NULL,		/* set values hook */
		XtInheritSetValuesAlmost,	/* set values almost */
		NULL,		/* get values hook */
		NULL,		/* accept focus */
		XtVersion,	/* version */
		NULL,		/* callback private */
		defaultTranslationsGame,	/* tm table */
		NULL,		/* query geometry */
		NULL,		/* display accelerator */
		NULL		/* extension */
	},
	{
		0		/* ignore */
	}
};

WidgetClass tetrisWidgetClass = (WidgetClass) & tetrisClassRec;

static void
setGame(TetrisWidget w, int reason)
{
	tetrisCallbackStruct cb;

	cb.reason = reason;
	XtCallCallbacks((Widget) w, (char *) XtNselectCallback, &cb);
}

static void
setGameNext(TetrisWidget w, int reason,
		int randomNumber, int randomRotation, int randomReflection,
		int units, Boolean diagonal, Boolean mixed, Boolean newBonus)
{
	tetrisCallbackStruct cb;

	cb.reason = reason;
	cb.randomNumber = randomNumber;
	cb.randomRotation = randomRotation;
	cb.randomReflection = randomReflection;
	cb.units = units;
	cb.diagonal = diagonal;
	cb.mixed = mixed;
	cb.newBonus = newBonus;
	XtCallCallbacks((Widget) w, (char *) XtNselectCallback, &cb);
}

#endif

#include "image.h"

static Pixmap dr = 0; /* dummy for future double buffering */

static const int delay[MAX_LEVELS] = {
	100, 92, 84, 75, 68, 60, 52, 45, 39, 32,
	26, 21, 16, 12, 8, 5, 3, 2, 1, 0
};

static void
CheckThings(TetrisWidget w)
{
	char *buf1 = NULL, *buf2 = NULL;

	if (w->tetris.units < MIN_UNITS || w->tetris.units > MAX_UNITS) {
		intCat(&buf1, "Number of units out of bounds, use ",
			MIN_UNITS);
		stringCat(&buf2, buf1, "...");
		free(buf1);
		intCat(&buf1, buf2, MAX_UNITS);
		free(buf2);
		stringCat(&buf2, buf1, ", defaulting to ");
		free(buf1);
		intCat(&buf1, buf2, DEFAULT_UNITS);
		free(buf2);
		DISPLAY_WARNING(buf1);
		free(buf1);
		w->tetris.units = DEFAULT_UNITS;
		w->tetris.next.units = w->tetris.units;
	}
	if (w->tetris.columns < MIN_COLUMNS) {
		intCat(&buf1, "Number of columns out of bounds, use at least ",
			MIN_COLUMNS);
		stringCat(&buf2, buf1, ", defaulting to ");
		free(buf1);
		intCat(&buf1, buf2, DEFAULTCOLUMNS);
		free(buf2);
		DISPLAY_WARNING(buf1);
		free(buf1);
		w->tetris.columns = DEFAULTCOLUMNS;
	}
	if (!LEGAL(w->tetris.units, w->tetris.diagonal,
			w->tetris.mixed, FALSE)) {
		intCat(&buf1, "Illegal configuration of units ",
			w->tetris.units);
		stringCat(&buf2, buf1, ", diagonal ");
		free(buf1);
		intCat(&buf1, buf2, (int) w->tetris.diagonal);
		free(buf2);
		stringCat(&buf2, buf1, ", mixed ");
		free(buf1);
		intCat(&buf1, buf2, (int) w->tetris.mixed);
		free(buf2);
		stringCat(&buf2, buf1, ", defaulting to units ");
		free(buf1);
		intCat(&buf1, buf2, DEFAULT_UNITS);
		free(buf2);
		stringCat(&buf2, buf1, ", diagonal ");
		free(buf1);
		intCat(&buf1, buf2, (int) DEFAULTDIAGONAL);
		free(buf2);
		stringCat(&buf2, buf1, ", mixed ");
		free(buf1);
		intCat(&buf1, buf2, (int) DEFAULTMIXED);
		free(buf2);
		DISPLAY_WARNING(buf1);
		free(buf1);
		w->tetris.units = DEFAULT_UNITS;
		w->tetris.diagonal = FALSE;
		w->tetris.mixed = FALSE;
		w->tetris.next.units = w->tetris.units;
		w->tetris.next.diagonal = w->tetris.diagonal;
		w->tetris.next.mixed = w->tetris.mixed;
	}
}

static void
drawUnit(TetrisWidget w, int pmid, int cid, int i, int j)
{
	if (i >= 0 && i < w->tetris.columns && j >= 0 && j < w->tetris.rows) {
#ifdef IMAGE
		Boolean image = True;
#ifdef WINVER
		HBITMAP hOldBitmap, monoOldBitmap, hBitmap;
		HDC monoDC;
#endif

		if (images[FALSE][pmid % BITMAPS] == None) {
			if (images[FALSE][BITMAPS - 1] != None)
				(void) printf("Bitmap unknown %d.\n", pmid);
			image = False;
		}
#ifdef WINVER
		if (image && (monoDC = CreateCompatibleDC(0)) == NULL) {
			(void) printf("Can not create mono DC.\n");
			image = False;
		}
		if (image && (w->core.memDC = CreateCompatibleDC(w->core.hDC)) == NULL) {
			(void) printf("Can not create compatible DC.\n");
			image = False;
		}
		if (image && (hBitmap = CreateCompatibleBitmap(w->core.hDC,
				w->tetris.boxSize, w->tetris.boxSize)) == NULL) {
			(void) printf("Can not create compatible bitmap.\n");
			image = False;
		}
#endif
		if (image) {
#ifdef WINVER
			FILLRECTANGLE(w, hBitmap, w->tetris.thingGC[cid],
				0, 0,
				w->tetris.boxSize, w->tetris.boxSize);
			monoOldBitmap = (HBITMAP) SelectObject(monoDC,
				images[FALSE][pmid % BITMAPS]);
			hOldBitmap = SelectObject(w->core.memDC, hBitmap);
			BitBlt(w->core.memDC,
				0, 0, w->tetris.boxSize, w->tetris.boxSize,
				monoDC, 0, 0,
				MERGEPAINT);
			(void) SelectObject(monoDC, monoOldBitmap);
			DeleteDC(monoDC);
			BitBlt(w->core.hDC,
				w->tetris.boxSize * i + w->tetris.offset.x,
				w->tetris.boxSize * j + w->tetris.offset.y,
				w->tetris.boxSize, w->tetris.boxSize,
				w->core.memDC, 0, 0,
				SRCCOPY);
			(void) SelectObject(w->core.memDC, hOldBitmap);
			DeleteObject(hBitmap);
			DeleteDC(w->core.memDC);
#else
			(void) XPutImage(XtDisplay(w), XtWindow(w),
				w->tetris.thingGC[cid],
				images[FALSE][pmid % BITMAPS],
				0, 0,
				w->tetris.boxSize * i + w->tetris.offset.x,
				w->tetris.boxSize * j + w->tetris.offset.y,
				w->tetris.boxSize, w->tetris.boxSize);
#endif
		} else
#endif
		{
			FILLRECTANGLE(w, dr, w->tetris.thingGC[cid],
				w->tetris.boxSize * i + w->tetris.offset.x,
				w->tetris.boxSize * j + w->tetris.offset.y,
				w->tetris.boxSize, w->tetris.boxSize);
		}
	}
}

static void
clearUnit(TetrisWidget w, int i, int j)
{
	if (i >= 0 && i < w->tetris.columns && j >= 0 && j < w->tetris.rows) {
		FILLRECTANGLE(w, dr, w->tetris.inverseGC,
			w->tetris.boxSize * i + w->tetris.offset.x,
			w->tetris.boxSize * j + w->tetris.offset.y,
			w->tetris.boxSize, w->tetris.boxSize);
	}
}

static void
xorUnit(TetrisWidget w, int i, int j)
{
#ifdef WINVER
	HDC memDC;
	HBITMAP hBitmap, hOldBitmap;

	if ((memDC = CreateCompatibleDC(w->core.hDC)) == NULL) {
		DISPLAY_ERROR("Not enough memory (xor memDC), exiting.");
	}
	if ((hBitmap = CreateCompatibleBitmap(w->core.hDC,
			w->tetris.boxSize,
			w->tetris.boxSize)) == NULL) {
		DISPLAY_ERROR("Not enough memory (xor hBitmap), exiting.");
	}
	hOldBitmap = SelectObject(memDC, hBitmap);
	BitBlt(memDC, 0, 0,
		w->tetris.boxSize, w->tetris.boxSize,
		w->core.hDC,
		w->tetris.boxSize * i + w->tetris.offset.x,
		w->tetris.boxSize * j + w->tetris.offset.y,
		SRCCOPY);
	BitBlt(w->core.hDC,
		w->tetris.boxSize * i + w->tetris.offset.x,
		w->tetris.boxSize * j + w->tetris.offset.y,
		w->tetris.boxSize, w->tetris.boxSize,
		memDC,
		0, 0,
		DSTINVERT);
	(void)  SelectObject(memDC, hOldBitmap);
	DeleteObject(hBitmap);
	DeleteDC(memDC);
#else
	FILLRECTANGLE(w, dr, w->tetris.xorGC,
		w->tetris.boxSize * i + w->tetris.offset.x,
		w->tetris.boxSize * j + w->tetris.offset.y,
		w->tetris.boxSize, w->tetris.boxSize);
#endif
}

static void
moveUnit(TetrisWidget w, int ai, int aj, int bi, int bj)
{
#ifdef WINVER
	BitBlt(w->core.hDC,
		w->tetris.offset.x + bi * w->tetris.boxSize,
		w->tetris.offset.y + bj * w->tetris.boxSize,
		w->tetris.boxSize, w->tetris.boxSize,
		w->core.hDC,
		w->tetris.offset.x + ai * w->tetris.boxSize,
		w->tetris.offset.y + aj * w->tetris.boxSize,
		SRCCOPY);
#else
	VOID XSetGraphicsExposures(XtDisplay(w), w->tetris.inverseGC,
		False);
	VOID XCopyArea(XtDisplay(w), XtWindow(w), XtWindow(w),
		w->tetris.inverseGC,
		w->tetris.offset.x + ai * w->tetris.boxSize,
		w->tetris.offset.y + aj * w->tetris.boxSize,
		w->tetris.boxSize, w->tetris.boxSize,
		w->tetris.offset.x + bi * w->tetris.boxSize,
		w->tetris.offset.y + bj * w->tetris.boxSize);
#endif
}

static void
drawPartialField(TetrisWidget w, int start, int finish)
{
	int i, j, offset;

	for (j = start; j < finish; j++) {
		offset = j * w->tetris.columns;
		for (i = 0; i < w->tetris.columns; i++) {
			if (w->tetris.field[offset + i].pmid >= 0)
				drawUnit(w, w->tetris.field[offset + i].pmid,
					w->tetris.field[offset + i].cid,
					i, j);
		}
	}
}

static void
drawField(TetrisWidget w)
{
	drawPartialField(w, 0, w->tetris.rows);
}

static void
clearField(TetrisWidget w)
{
	int i, j;

	for (j = 0; j < w->tetris.rows; j++) {
		for (i = 0; i < w->tetris.columns; i++) {
			clearUnit(w, i, j);
		}
	}
}

static void
drawThing(TetrisWidget w)
{
	int i, j;

	for (i = 0; i < w->tetris.curThing.size; i++)
		for (j = 0; j < w->tetris.curThing.size; j++)
			if (tris[w->tetris.curThing.units].polyomino
					[w->tetris.curThing.polyominoNumber].shape[j][i] != 0)
				drawUnit(w, polyPixmap(i, j,
					(tris[w->tetris.curThing.units].polyomino
					[w->tetris.curThing.polyominoNumber].shape),
					w->tetris.curThing.size),
					w->tetris.curThing.colorNumber,
					w->tetris.curThing.xpos + i,
					w->tetris.curThing.ypos + j);
}

static void
drawThingDiff(TetrisWidget w, Thing oldThing)
{
	int i, j, ox, oy;


	for (i = 0; i < w->tetris.curThing.size; i++)
		for (j = 0; j < w->tetris.curThing.size; j++) {
			if ((w->tetris.curThing.ypos + j >= 0) &&
			tris[w->tetris.curThing.units].polyomino
			[w->tetris.curThing.polyominoNumber].shape[j][i] != 0)
				drawUnit(w, polyPixmap(i, j,
					(tris[w->tetris.curThing.units].polyomino
					[w->tetris.curThing.polyominoNumber].shape),
w->tetris.curThing.size),
					w->tetris.curThing.colorNumber,
					w->tetris.curThing.xpos + i,
					w->tetris.curThing.ypos + j);
		}
	for (i = 0; i < w->tetris.curThing.size; i++)
		for (j = 0; j < w->tetris.curThing.size; j++) {
			ox = oldThing.xpos + i -
				w->tetris.curThing.xpos;
			oy = oldThing.ypos + j -
				w->tetris.curThing.ypos;
			if (tris[oldThing.units].polyomino
			[oldThing.polyominoNumber].shape[j][i] != 0 &&
			((ox < 0) || (ox >= w->tetris.curThing.size) ||
			(oy < 0) || (oy >= w->tetris.curThing.size) ||
		 	tris[w->tetris.curThing.units].polyomino
			[w->tetris.curThing.polyominoNumber].shape[oy][ox] == 0))
		clearUnit(w, (oldThing.xpos + i), (oldThing.ypos + j));
	}
}

static void
EraseFrame(TetrisWidget w)
{
	FILLRECTANGLE(w, dr, w->tetris.inverseGC,
		0, 0, w->core.width, w->core.height);
}

static void
DrawFrame(TetrisWidget w, Boolean focus)
{
	DRAWRECTANGLE(w, dr, (focus) ? w->tetris.frameGC : w->tetris.borderGC,
		w->tetris.offset.x - 1, w->tetris.offset.y - 1,
		w->tetris.viewLength + 1, 2 * w->tetris.viewLength + 1);
}

static void
drawNext(TetrisWidget w)
{
	setGameNext(w, GAME_DRAWNEXT,
		w->tetris.nextThing.randomNumber,
		w->tetris.nextThing.randomRotation,
		w->tetris.nextThing.randomReflection,
		w->tetris.units, w->tetris.diagonal, w->tetris.mixed,
		w->tetris.next.bonus);
}

static void
drawStatus(TetrisWidget w)
{
	setGame(w, GAME_STATUS);
}

static void
redoNext(TetrisWidget w)
{
	int diagonal = (w->tetris.next.diagonal) ? 1 : 0;
	int turnStyle = tris[w->tetris.next.units -
		MIN_UNITS].mode[diagonal].turnStyle;
	int next_start, i;

	w->tetris.nextThing.units = w->tetris.next.units - MIN_UNITS +
		w->tetris.nextThing.bonus;
	next_start = tris[w->tetris.nextThing.units].mode
		[diagonal].number[(w->tetris.nextThing.bonus) ?
	       	NOMIX : w->tetris.next.mixed][turnStyle];
	if (!next_start) {
		DISPLAY_ERROR("Error, no polyominoes for this combination");
	}
	next_start = w->tetris.nextThing.randomNumber % next_start;
	w->tetris.nextThing.colorNumber = next_start;

	if (w->tetris.next.mixed && !w->tetris.nextThing.bonus) {
		w->tetris.nextThing.units = 0;
		while (next_start >= tris[w->tetris.nextThing.units].mode
				[diagonal].number[NOMIX][turnStyle]) {
			next_start -= tris[w->tetris.nextThing.units].mode
				[diagonal].number[NOMIX][turnStyle];
			w->tetris.nextThing.units++;
		}
	}
	w->tetris.nextThing.polyominoNumber =
		tris[w->tetris.nextThing.units].mode
		[diagonal].start[next_start][turnStyle];
	if (tris[w->tetris.nextThing.units].mode[diagonal].turnStyle > NONE)
		for (i = 0; i < w->tetris.nextThing.randomRotation; i++)
			w->tetris.nextThing.polyominoNumber =
				tris[w->tetris.nextThing.units].polyomino
				[w->tetris.nextThing.polyominoNumber].rotation;
	if (tris[w->tetris.nextThing.units].mode[diagonal].turnStyle == ALL)
		for (i = 0; i < w->tetris.nextThing.randomReflection; i++)
			w->tetris.nextThing.polyominoNumber =
				tris[w->tetris.nextThing.units].polyomino
				[w->tetris.nextThing.polyominoNumber].reflection;
	w->tetris.nextThing.size = tris[w->tetris.nextThing.units].polyomino
		[w->tetris.nextThing.polyominoNumber].size;
	w->tetris.nextThing.xpos =
		NRAND(w->tetris.columns - w->tetris.nextThing.size + 1);
	/* nextThing.xpos = (w->tetris.columns - nextThing.size) / 2; */
	w->tetris.nextThing.ypos =
		-tris[w->tetris.nextThing.units].polyomino
		[w->tetris.nextThing.polyominoNumber].startHeight[(w->tetris.suddenAppear) ? JUMPIN : GRADUAL];
}

static void
initFill(TetrisWidget w)
{
	int i, j, col, offset, fixedUnits;

	if (w->tetris.fixedFill) {
		/* This algorithm randomly picks fixedUnits per row
		 * (4 for DEFAULTCOLUMNS, ie. 10/2-1)
		 * This is more in keeping with the original */
		fixedUnits = w->tetris.columns / 2 - 1;
		for (j = 0; j < w->tetris.rows; j++) {
			offset = j * w->tetris.columns;
			for (i = 0; i < w->tetris.columns; i++) {
				w->tetris.field[offset + i].pmid = -1;
				w->tetris.field[offset + i].cid = 0;
			}
			if (j >= w->tetris.rows - w->tetris.fill)
				for (i = 0; i < fixedUnits; i++) {
					do
						col = NRAND(w->tetris.columns);
					while (w->tetris.field[offset + col].pmid != -1);
					w->tetris.field[offset + col].pmid = BITMAPS - 1;
					w->tetris.field[offset + col].cid = NRAND(w->tetris.numberPolyominoes);
				}
		}
	} else {
		/* This algorithm randomly fills boxes to a certain level
		 * The problem with this is that it could potentially fill
		 * up a whole row */
		for (j = 0; j < w->tetris.rows; j++) {
			offset = j * w->tetris.columns;
			for (i = 0; i < w->tetris.columns; i++)
				if ((j >= w->tetris.rows - w->tetris.fill) &&
						(NRAND(2) == 0)) {
					w->tetris.field[offset + i].pmid = BITMAPS - 1;
					w->tetris.field[offset + i].cid = NRAND(w->tetris.numberPolyominoes);
				} else {
					w->tetris.field[offset + i].pmid = -1;
					w->tetris.field[offset + i].cid = 0;
				}
		}
	}
}

static void
fillLines(TetrisWidget w)
{
	int i, j;
	int diagonal = (w->tetris.current.diagonal) ? 1 : 0;
	int mixed = (w->tetris.current.mixed) ? 1 : 0;
	int turnStyle = tris[w->tetris.current.units -
		MIN_UNITS].mode[diagonal].turnStyle;

	FLUSH(w);
	for (j = 0; j <= w->tetris.rows / 2; j++) {
		for (i = 0; i < w->tetris.columns; i++) {
			drawUnit(w, BITMAPS - 1, NRAND(tris[w->tetris.current.units -
				MIN_UNITS].mode[diagonal].number[mixed][turnStyle]), i, j);
			drawUnit(w, BITMAPS - 1, NRAND(tris[w->tetris.current.units -
				MIN_UNITS].mode[diagonal].number[mixed][turnStyle]), i, w->tetris.rows - j - 1);
		}
		FLUSH(w);
		Sleep(100);
	}
	Sleep(1000);
}

static Boolean
overlapping(TetrisWidget w)
{
	int i, j;
	int x = w->tetris.curThing.xpos, y = w->tetris.curThing.ypos;

	for (i = 0; i < w->tetris.curThing.size; i++)
		for (j = 0; j < w->tetris.curThing.size; j++)
			if (tris[w->tetris.curThing.units].polyomino
			[w->tetris.curThing.polyominoNumber].shape[j][i] != 0) {
				int xi = x + i;
				int yj = y + j;

				if ((yj >= w->tetris.rows) || (xi < 0) ||
						(xi >= w->tetris.columns))
					return True;
				if (w->tetris.suddenAppear) { /* creationalism */
/* This method does not allow turning polyomino to an area above screen */
					if ((yj < 0) || (w->tetris.field[yj *
							w->tetris.columns + xi].pmid >= 0)) {
						return True;
					}
				} else { /* evolution */
/* This method one can turn polyomino to an area above of screen,
 * also part of the polyomino may not be visible initially */
					if ((yj >= 0) && (w->tetris.field[yj *
							w->tetris.columns + xi].pmid >= 0)) {
						return True;
					}
				}
			}
	return False;
}

static Boolean
atBottom(TetrisWidget w)
{
	int i, j;
	int x = w->tetris.curThing.xpos, y = w->tetris.curThing.ypos;

	for (i = 0; i < w->tetris.curThing.size; i++)
		for (j = 0; j < w->tetris.curThing.size; j++) {
			int xi = x + i;
			int yj = y + j;

			if ((yj >= -1) && tris[w->tetris.curThing.units].polyomino
				[w->tetris.curThing.polyominoNumber].shape[j][i] != 0)
				if ((yj >= w->tetris.rows - 1) ||
					(xi < 0) ||
					(xi >= w->tetris.columns) ||
					(w->tetris.field[(yj + 1) *
					w->tetris.columns + xi].pmid >= 0))
					return True;
		}
	return False;
}

static void
putBox(TetrisWidget w)
{
	int i, j;
	int x = w->tetris.curThing.xpos, y = w->tetris.curThing.ypos;

	for (i = 0; i < w->tetris.curThing.size; i++)
		for (j = 0; j < w->tetris.curThing.size; j++) {
			int xi = x + i;
			int yj = y + j;

			if ((yj >= 0) && tris[w->tetris.curThing.units].polyomino
				[w->tetris.curThing.polyominoNumber].shape[j][i]!= 0) {
				w->tetris.field[(yj) * w->tetris.columns + xi].pmid =
					(polyPixmap(i, j,
					(tris[w->tetris.curThing.units].polyomino
					[w->tetris.curThing.polyominoNumber].shape),
					w->tetris.curThing.size)) % BITMAPS;
				w->tetris.field[(yj) * w->tetris.columns + xi].cid =
					w->tetris.curThing.colorNumber;
			}
		}
}

static int
checkLines(TetrisWidget w)
{
	int *lSet, nSet = 0;
	int i, j, k, y, offset;

	if  (!(lSet = (int *) malloc(sizeof (int) * w->tetris.rows))) {
		DISPLAY_ERROR("Not enough memory (lSet), exiting.");
	}
	/* Find filled rows */
	for (j = 0; j < w->tetris.rows; j++) {
		lSet[j] = 0;
		offset = j * w->tetris.columns;
		for (i = 0; i < w->tetris.columns; i++)
			if (w->tetris.field[offset + i].pmid >= 0)
				lSet[j]++;
		if (lSet[j] == w->tetris.columns)
			nSet++;
	}

	if (nSet == 0) {
		free(lSet);
		return 0;
	}
	for (k = 0; k < 2 * NUM_FLASHES; k++) {
		for (j = 0; j < w->tetris.rows; j++) {
			if (lSet[j] == w->tetris.columns) {
				for (i = 0; i < w->tetris.columns; i++)
					xorUnit(w, i, j);
			}
		}
		FLUSH(w);
		Sleep(100);
	}
	/* Shrink */
	for (j = w->tetris.rows - 1; j >= 0; j--)
		if (lSet[j] == w->tetris.columns) {
			for (y = j; y > 0; y--)
				for (i = 0; i < w->tetris.columns; i++) {
					w->tetris.field[y * w->tetris.columns + i] =
						w->tetris.field[(y - 1) * w->tetris.columns + i];
					moveUnit(w, i, y - 1, i, y);
				}
			for (i = 0; i < w->tetris.columns; i++) {
				w->tetris.field[i].pmid = -1;
				clearUnit(w, i, 0);
			}
			for (i = j; i > 0; i--)
				lSet[i] = lSet[i - 1];
			lSet[0] = 0;
			if (j > 0)
				for (i = 0; i < w->tetris.columns; i++) {
					int tmp = w->tetris.field[j * w->tetris.columns + i].pmid;

					if ((tmp >= 0) && (tmp != CHECKDOWN(tmp))) {
						w->tetris.field[j * w->tetris.columns + i].pmid = CHECKDOWN(tmp);
						drawUnit(w, w->tetris.field[j * w->tetris.columns + i].pmid,
						w->tetris.field[j * w->tetris.columns + i].cid, i, j);
					}
				}
			j++;
			if (j < w->tetris.rows)
				for (i = 0; i < w->tetris.columns; i++) {
					int tmp = w->tetris.field[j * w->tetris.columns + i].pmid;

					if ((tmp >= 0) && (tmp != CHECKUP(tmp))) {
						w->tetris.field[j * w->tetris.columns + i].pmid = CHECKUP(tmp);
						drawUnit(w, w->tetris.field[j * w->tetris.columns + i].pmid,
						w->tetris.field[j * w->tetris.columns + i].cid, i, j);
					}
				}
			FLUSH(w);
#ifdef USE_SOUND
			if (w->tetris.sound) {
				playSound((char *) BUMPSOUND);
			}
#endif
		}
	free(lSet);
	return nSet;
}

static void
newThing(TetrisWidget w)
{
	w->tetris.curThing = w->tetris.nextThing;
	/*w->tetris.current = w->tetris.next;*/
	w->tetris.current.units = w->tetris.next.units;
	w->tetris.current.diagonal = w->tetris.next.diagonal;
	w->tetris.current.mixed = w->tetris.next.mixed;
	w->tetris.current.bonus = w->tetris.next.bonus;
	w->tetris.nextThing.randomNumber = LRAND();
	w->tetris.nextThing.randomRotation = NRAND(MAX_SIDES);
	w->tetris.nextThing.randomReflection = NRAND(2);
	w->tetris.nextThing.bonus = (w->tetris.next.bonus) ? 1 : 0;
	redoNext(w);
}

static void
ClearThings(TetrisWidget w, Boolean draw)
{
	if (draw)
		clearField(w);
	initFill(w);
	if (draw)
		drawField(w);
	w->tetris.delay = delay[w->tetris.level - 1];
	w->tetris.score = w->tetris.fill * w->tetris.level * 10;
	w->tetris.started = True;
	w->tetris.paused = False;
	w->tetris.pauseStart = 0;
	w->tetris.completedRows = 0;
	w->tetris.newRows = 0;
	drawStatus(w);
}

static void
gameOver(TetrisWidget w, Boolean fast)
{
	if (!w->tetris.started)
		return;
	if (!fast)
		fillLines(w);
	setGame(w, GAME_GAMEOVER);
	ClearThings(w, True);
}

static void
tryMove(TetrisWidget w, move_t move)
{
	int diagonal = (w->tetris.current.diagonal) ? 1 : 0;
	Thing old;
	int i;

	old = w->tetris.curThing;

	switch (move) {
	case FALL:
		w->tetris.curThing.ypos++;
		break;

	case DROP:
		do {
			w->tetris.curThing.ypos++;
			w->tetris.score += w->tetris.level + w->tetris.fill;
		} while (!overlapping(w));
		w->tetris.curThing.ypos--;
		break;

	case MOVE_LEFT:
		w->tetris.curThing.xpos--;
		break;

	case MOVE_RIGHT:
		w->tetris.curThing.xpos++;
		break;

	case ROTATE:
		if (tris[w->tetris.current.units -
				MIN_UNITS].mode[diagonal].turnStyle > NONE) {
			if (w->tetris.cw)
				for (i = 0; i < MAX_SIDES - 1; i++)
					w->tetris.curThing.polyominoNumber =
						tris[w->tetris.curThing.units].polyomino
						[w->tetris.curThing.polyominoNumber].rotation;
			else
				w->tetris.curThing.polyominoNumber =
					tris[w->tetris.curThing.units].polyomino
					[w->tetris.curThing.polyominoNumber].rotation;
			w->tetris.curThing.xpos = old.xpos;
			w->tetris.curThing.ypos = old.ypos;
	}
	break;

	case REFLECT: /* reflect on y axis */
		if (tris[w->tetris.current.units -
				MIN_UNITS].mode[diagonal].turnStyle == ALL) {
		w->tetris.curThing.polyominoNumber =
			tris[w->tetris.curThing.units].polyomino
			[w->tetris.curThing.polyominoNumber].reflection;
		w->tetris.curThing.xpos = old.xpos;
		w->tetris.curThing.ypos = old.ypos;
		}
		break;
	}

	if (!overlapping(w)) {
		drawThingDiff(w, old);
	} else {
		w->tetris.curThing = old;
	}
}

static Boolean
moveOne(TetrisWidget w, move_t move)
{
	int lines;

	if ((move == DROP) || ((move == FALL) && atBottom(w))) {
		tryMove(w, move);
		putBox(w);
		lines = checkLines(w);
		w->tetris.score += (w->tetris.fill + w->tetris.level +
			lines + ((w->tetris.suddenAppear) ? 1 : 0)) *
			((w->tetris.current.bonus) ? 1 : 3) *
			((w->tetris.showNext) ? 1 : 3) /
			(1 + w->tetris.current.mixed);
		w->tetris.score += ((w->tetris.showNext) ? 3 : 5) *
			w->tetris.level * w->tetris.level * lines * lines *
			((w->tetris.current.bonus) ? 3 : 5) *
			w->tetris.startLevel *
			(w->tetris.current.units + ((w->tetris.current.diagonal) ? 1 : 0)) *
			(w->tetris.current.units + ((w->tetris.current.diagonal) ? 1 : 0)) /
			(16 * (1 + ((w->tetris.current.mixed) ? 1 : 0)));
		w->tetris.completedRows += lines;
		w->tetris.newRows += lines;
		w->tetris.next.bonus = False;
		if ((w->tetris.newRows > THRESHOLD(w->tetris.level)) &&
				(w->tetris.level < MAX_LEVELS - 1)) {
			w->tetris.level++;
			w->tetris.newRows = 0;
			if (LEGAL(w->tetris.next.units,
					w->tetris.next.diagonal,
					w->tetris.next.mixed,
					w->tetris.bonus))
				w->tetris.next.bonus = True;
		}
		drawStatus(w);
		newThing(w);
		drawThing(w);
		drawNext(w);
		FLUSH(w);
		if (overlapping(w))
			gameOver(w, False);
		return True;
	} else {
		tryMove(w, move);
		if ((w->tetris.newRows > THRESHOLD(w->tetris.level)) &&
				(w->tetris.level < MAX_LEVELS - 1)) {
			w->tetris.next.bonus = False;
			w->tetris.level++;
			w->tetris.newRows = 0;
			if (LEGAL(w->tetris.next.units,
					w->tetris.next.diagonal,
					w->tetris.next.mixed,
					w->tetris.bonus))
				w->tetris.next.bonus = True;
			drawStatus(w);
			drawNext(w);
		}
		return False;
	}
}

#ifndef WINVER
static
#endif
void
FalldownGame(TetrisWidget w)
{
	if (!w->tetris.paused) {
		w->tetris.gotNew = moveOne(w, FALL) || w->tetris.gotNew;
	} else {
		if (w->tetris.pauseStart >= 32) {
			ClearThings(w, True);
		}
		if (!w->tetris.started)
			w->tetris.pauseStart++;
	}
}

#ifndef WINVER
static
#endif
void
MoveGameInput(TetrisWidget w, int direction)
{
	if (!w->tetris.paused)
		w->tetris.gotNew = moveOne(w, (move_t) direction) ||
			w->tetris.gotNew;
}

static void
ResetThings(TetrisWidget w)
{
	if (w->tetris.field)
		free(w->tetris.field);
	w->tetris.rows = 2 * w->tetris.columns;
	if (!(w->tetris.field = (Field *) malloc(
			w->tetris.rows * w->tetris.columns * sizeof(Field)))) {
		DISPLAY_ERROR("Not enough memory (field), exiting.");
	}
	w->tetris.started = False;
}

static void
ResizeThings(TetrisWidget w)
{
#ifdef IMAGE
#ifndef WINVER
	Display *display = XtDisplay(w);
	Window window = XtWindow(w);
	XWindowAttributes xgwa;

	(void) XGetWindowAttributes(display, window, &xgwa);
#endif
	free_images(FALSE);
	if (w->tetris.boxSize > 4)
		(void) create_images(
#ifndef WINVER
			display, xgwa.visual,
#endif
			(w->tetris.plainPiece) ? 0 : 1, FALSE,
			w->tetris.boxSize);
#endif
}

#ifdef WINVER
static void
SetValuesGame(TetrisWidget w)
{
	struct tagColor {
		int red, green, blue;
	} color;
	char szBuf[80], buf[20];
	int thing;

	w->tetris.level = GetPrivateProfileInt(SECTION, "level",
		DEFAULTLEVELS, INIFILE);
	w->tetris.units = GetPrivateProfileInt(SECTION, "units",
		DEFAULTUNITS, INIFILE);
	w->tetris.columns = GetPrivateProfileInt(SECTION, "columns",
		DEFAULTCOLUMNS, INIFILE);
	w->tetris.fill = GetPrivateProfileInt(SECTION, "fill",
		0, INIFILE);
	w->tetris.fixedFill = (BOOL) GetPrivateProfileInt(SECTION,
		"fixedFill", TRUE, INIFILE);
	w->tetris.suddenAppear = (BOOL) GetPrivateProfileInt(SECTION,
		"suddenAppear", FALSE, INIFILE);
	w->tetris.diagonal = (BOOL) GetPrivateProfileInt(SECTION, "diagonal",
		DEFAULTDIAGONAL, INIFILE);
	w->tetris.mixed = (BOOL) GetPrivateProfileInt(SECTION, "mixed",
		DEFAULTMIXED, INIFILE);
	w->tetris.bonus = (BOOL) GetPrivateProfileInt(SECTION, "bonus",
		TRUE, INIFILE);
	w->tetris.cw = (BOOL) GetPrivateProfileInt(SECTION, "cw",
		TRUE, INIFILE);
	w->tetris.plainPiece = (BOOL) GetPrivateProfileInt(SECTION,
		"plainPiece", FALSE, INIFILE);
	w->tetris.showNext = (BOOL) GetPrivateProfileInt(SECTION, "showNext",
		FALSE, INIFILE);
	w->tetris.sound = (BOOL) GetPrivateProfileInt(SECTION, "sound",
		FALSE, INIFILE);
	(void) GetPrivateProfileString(SECTION, "bumpSound", BUMPSOUND,
		szBuf, sizeof (szBuf), INIFILE);
	(void) strcpy(w->tetris.bumpSound, szBuf);
	w->tetris.bumpSound[80] = 0;
	w->tetris.mono = (BOOL) GetPrivateProfileInt(SECTION, "mono",
		DEFAULTMONO, INIFILE);
	w->tetris.reverse = (BOOL) GetPrivateProfileInt(SECTION, "reverse",
		DEFAULTREVERSE, INIFILE);
	/* cyan */
	(void) GetPrivateProfileString(SECTION, "frameColor", "0 255 255",
		szBuf, sizeof (szBuf), INIFILE);
	(void) sscanf(szBuf, "%d %d %d",
		&(color.red), &(color.green), &(color.blue));
	w->tetris.frameGC = RGB(color.red, color.green, color.blue);
	/* gray25 */
	(void) GetPrivateProfileString(SECTION, "borderColor", "64 64 64",
		szBuf, sizeof (szBuf), INIFILE);
	(void) sscanf(szBuf, "%d %d %d",
		&(color.red), &(color.green), &(color.blue));
	w->tetris.borderGC = RGB(color.red, color.green, color.blue);
	/* #AEB2C3 */
	(void) GetPrivateProfileString(SECTION, "background", "255 255 255",
		szBuf, sizeof (szBuf), INIFILE);
	(void) sscanf(szBuf, "%d %d %d",
		&(color.red), &(color.green), &(color.blue));
	w->tetris.inverseGC = RGB(color.red, color.green, color.blue);
	for (thing = 0; thing < MAX_START_POLYOMINOES; thing++) {
		(void) sprintf(buf, "thingColor%d", thing);
		(void) GetPrivateProfileString(SECTION, buf,
			polyominoColorString[thing],
			szBuf, sizeof (szBuf), INIFILE);
		(void) sscanf(szBuf, "%d %d %d",
			&(color.red), &(color.green), &(color.blue));
		w->tetris.thingGC[thing] =
			RGB(color.red, color.green, color.blue);
	}
	(void) GetPrivateProfileString(SECTION, "userName", "Guest",
		szBuf, sizeof (szBuf), INIFILE);
	(void) strcpy(w->tetris.userName, szBuf);
		w->tetris.userName[80] = 0;
	(void) GetPrivateProfileString(SECTION, "scoreFile", "",
		szBuf, sizeof (szBuf), INIFILE);
	(void) strcpy(w->tetris.scoreFile, szBuf);
		w->tetris.scoreFile[80] = 0;
}

void
DestroyGame(HBRUSH brush)
{
	(void) DeleteObject(brush);
	PostQuitMessage(0);
}

#else
static void
GetColor(TetrisWidget w, int thing)
{
	XGCValues values;
	XtGCMask valueMask;
	XColor colorCell, rgb;
	Boolean gotColor = False;

	valueMask = GCForeground | GCBackground;
	if (w->tetris.reverse) {
		values.background = w->tetris.foreground;
	} else {
		values.background = w->tetris.background;
	}
	if (!w->tetris.mono) {
		if (XAllocNamedColor(XtDisplay(w),
				DefaultColormapOfScreen(XtScreen(w)),
				polyominoColorName[thing],
				&colorCell, &rgb)) {
			values.foreground = w->tetris.thingColor[thing] =
				colorCell.pixel;
			gotColor = True;
		} else {
			char *buf1, *buf2;

			stringCat(&buf1, "Color name \"",
				polyominoColorName[thing]);
			stringCat(&buf2, buf1, "\" is not defined for polyomino ");
			free(buf1);
			intCat(&buf1, buf2, thing);
			free(buf2);
			DISPLAY_WARNING(buf1);
			free(buf1);
		}
	}
	if (!gotColor) {
		if (w->tetris.reverse) {
			values.background = w->tetris.foreground;
			values.foreground = w->tetris.background;
		} else {
			values.background = w->tetris.background;
			values.foreground = w->tetris.foreground;
		}
	}
	if (w->tetris.thingGC[thing])
		XtReleaseGC((Widget) w, w->tetris.thingGC[thing]);
	w->tetris.thingGC[thing] = XtGetGC((Widget) w, valueMask, &values);
}

static void
SetAllColors(TetrisWidget w)
{
	XGCValues values;
	XtGCMask valueMask;
	int polyomino;

	valueMask = GCForeground | GCBackground;

	if (w->tetris.reverse) {
		values.background = w->tetris.background;
		values.foreground = w->tetris.foreground;
	} else {
		values.foreground = w->tetris.background;
		values.background = w->tetris.foreground;
	}
	if (w->tetris.inverseGC)
		XtReleaseGC((Widget) w, w->tetris.inverseGC);
	w->tetris.inverseGC = XtGetGC((Widget) w, valueMask, &values);
	if (w->tetris.mono) {
		if (w->tetris.reverse) {
			values.background = w->tetris.foreground;
			values.foreground = w->tetris.background;
		} else {
			values.foreground = w->tetris.foreground;
			values.background = w->tetris.background;
		}
	} else {
		values.foreground = w->tetris.frameColor;
		values.background = w->tetris.background;
	}
	if (w->tetris.frameGC)
		XtReleaseGC((Widget) w, w->tetris.frameGC);
	w->tetris.frameGC = XtGetGC((Widget) w, valueMask, &values);
	if (w->tetris.mono) {
		if (w->tetris.reverse) {
			values.background = w->tetris.foreground;
			values.foreground = w->tetris.background;
		} else {
			values.foreground = w->tetris.foreground;
			values.background = w->tetris.background;
		}
	} else {
		values.foreground = w->tetris.borderColor;
		values.background = w->tetris.background;
	}
	if (w->tetris.borderGC)
		XtReleaseGC((Widget) w, w->tetris.borderGC);
	w->tetris.borderGC = XtGetGC((Widget) w, valueMask, &values);
	for (polyomino = 0; polyomino < MAX_START_POLYOMINOES; polyomino++)
		GetColor(w, polyomino);
	valueMask = GCForeground | GCBackground | GCFunction;
	values.foreground = 0xFFFFFF;
	values.background = 0;
	values.function = GXxor;
	if (w->tetris.xorGC)
		XtReleaseGC((Widget) w, w->tetris.xorGC);
	w->tetris.xorGC = XtGetGC((Widget) w, valueMask, &values);
}

static Boolean
SetValuesGame(Widget current, Widget request, Widget renew)
{
	TetrisWidget c = (TetrisWidget) current, w = (TetrisWidget) renew;
	Boolean redraw = False, setColors = False;

	if (w->tetris.act) {
		FalldownGame(w);
		w->tetris.act = False;
	}
	if (w->tetris.level != c->tetris.level)
		w->tetris.delay = delay[w->tetris.level - 1];
	CheckThings(w);
	if (w->tetris.borderColor != c->tetris.borderColor ||
			w->tetris.frameColor != c->tetris.frameColor ||
			w->tetris.reverse != c->tetris.reverse ||
			w->tetris.mono != c->tetris.mono) {
		SetAllColors(w);
		redraw = True;
	} else if (w->tetris.background != c->tetris.background ||
			w->tetris.foreground != c->tetris.foreground ||
			setColors) {
		SetAllColors(w);
		redraw = True;
	} else if (w->tetris.plainPiece != c->tetris.plainPiece) {
		redraw = True;
	}
	if (w->tetris.columns != c->tetris.columns) {
		w->tetris.paused = True;
		w->tetris.started = False;
		EraseFrame(w);
		SizeGame(w);
		ResizeThings(w);
		newThing(w);
		newThing(w);
		ClearThings(w, False);
		DrawFrame(w, w->tetris.focus);
		redraw = True;
	} else if (w->tetris.fill != c->tetris.fill) {
		ClearThings(w, True);
		redraw = True;
	} else if (w->tetris.units != c->tetris.units) {
		w->tetris.next.units = w->tetris.units;
		w->tetris.next.bonus = False;
		w->tetris.newRows = 0;
		redoNext(w);
		drawNext(w);
	} else if (w->tetris.diagonal != c->tetris.diagonal) {
		w->tetris.next.diagonal = w->tetris.diagonal;
		w->tetris.next.bonus = False;
		w->tetris.newRows = 0;
		redoNext(w);
		drawNext(w);
	} else if (w->tetris.mixed != c->tetris.mixed) {
		w->tetris.next.mixed = w->tetris.mixed;
		w->tetris.next.bonus = False;
		w->tetris.newRows = 0;
		redoNext(w);
		drawNext(w);
	} else if (w->tetris.bonus != c->tetris.bonus) {
		w->tetris.next.bonus = False;
		w->tetris.newRows = 0;
		redoNext(w);
		drawNext(w);
	}
	if (w->tetris.menu != -1) {
		switch (w->tetris.menu) {
		case 2:
			w->tetris.menu = -1;
			ClearThings(w, True);
			redraw = True;
			break;
		case 4:
			w->tetris.menu = -1;
			/*ConfirmThings(w);*/
			redraw = True;
			break;
		default:
			w->tetris.menu = -1;
			break;
		}
	}
	return (redraw);
}

static void
QuitGame(TetrisWidget w, XEvent *event, char **args, int nArgs)
{
	Display *display = XtDisplay(w);

	gameOver(w, True);
	XtCloseDisplay(display);
	exit(0);
}

static void
DestroyGame(Widget old)
{
	TetrisWidget w = (TetrisWidget) old;
	int thing;

	for (thing = 0; thing < MAX_START_POLYOMINOES; thing++)
		XtReleaseGC(old, w->tetris.thingGC[thing]);
	XtReleaseGC(old, w->tetris.borderGC);
	XtReleaseGC(old, w->tetris.frameGC);
	XtReleaseGC(old, w->tetris.inverseGC);
	XtReleaseGC(old, w->tetris.xorGC);
	XtRemoveCallbacks(old, XtNselectCallback, w->tetris.select);
}
#endif

#ifndef WINVER
static
#endif
void
ResizeGame(TetrisWidget w)
{
#ifdef WINVER
	RECT rect;

	/* Determine size of client area */
	(void) GetClientRect(w->core.hWnd, &rect);
	w->core.width = rect.right;
	w->core.height = rect.bottom;
#endif
	w->tetris.boxSize = MIN((w->core.height - 2) / 2, w->core.width - 2) /
		w->tetris.columns;
	w->tetris.viewLength = w->tetris.boxSize * w->tetris.columns;
	w->tetris.offset.x = ((int) w->core.width -
		w->tetris.viewLength) / 2;
	w->tetris.offset.y = ((int) w->core.height -
		2 * w->tetris.viewLength) / 2;
}

#ifndef WINVER
static
#endif
void
SizeGame(TetrisWidget w)
{
	ResetThings(w);
	ResizeGame(w);
}

#ifndef WINVER
static
#endif
void
InitializeGame(
#ifdef WINVER
TetrisWidget w, HBRUSH brush
#else
Widget request, Widget renew
#endif
)
{
	int diagonal, mixed, turnStyle;
#ifdef IMAGE
	int bitmap;
#endif
#ifndef WINVER
	int poly;
	TetrisWidget w = (TetrisWidget) renew;
#endif

#ifdef WINVER
	SetValuesGame(w);
#else
	w->tetris.mono = (DefaultDepthOfScreen(XtScreen(w)) < 2 ||
		w->tetris.mono);
	for (poly = 0; poly < MAX_START_POLYOMINOES; poly++)
		w->tetris.thingGC[poly] = NULL;
	w->tetris.borderGC = NULL;
	w->tetris.frameGC = NULL;
	w->tetris.inverseGC = NULL;
	w->tetris.xorGC = NULL;
#endif
#ifdef IMAGE
	for (bitmap = 0; bitmap < BITMAPS; bitmap++)
		images[FALSE][bitmap] = None;
#endif
	w->tetris.field = NULL;
	w->tetris.started = False;
	w->tetris.quit = False;
	w->tetris.completedRows = 0;
	w->tetris.newRows = 0;
	w->tetris.fall = False;
	w->tetris.focus = False;
	w->tetris.next.units = w->tetris.units;
	w->tetris.next.diagonal = w->tetris.diagonal;
	w->tetris.next.mixed = w->tetris.mixed;
	w->tetris.next.bonus = False;
	w->tetris.startLevel = w->tetris.level;
	w->tetris.nextThing.units = 0;
	w->tetris.nextThing.polyominoNumber = 0;
	w->tetris.nextThing.xpos = 0;
	w->tetris.nextThing.ypos = 0;
	w->tetris.nextThing.size = 0;
	w->tetris.nextThing.colorNumber = 0;
	w->tetris.nextThing.bonus = 0;
	w->tetris.nextThing.randomRotation = 0;
	w->tetris.nextThing.randomReflection = 0;
	w->tetris.nextThing.randomNumber = 0;
	w->tetris.gotNew = False;
	CheckThings(w);
	SizeGame(w);
#ifdef WINVER
	brush = CreateSolidBrush(w->tetris.inverseGC);
	SETBACK(w->core.hWnd, brush);
	(void) SRAND(time(NULL));
#else
	(void) SRAND(getpid());
	SetAllColors(w);
#endif
	diagonal = (w->tetris.next.diagonal) ? 1 : 0;
	mixed = (w->tetris.next.mixed) ? 1 : 0;
	readThings();
	turnStyle = tris[w->tetris.next.units -
		MIN_UNITS].mode[diagonal].turnStyle;
	/* use intCat */
	if (turnStyle < 0 || turnStyle >= MAX_TURNSTYLE_TYPES) {
		(void) fprintf(stderr, "Tetris: corrupted polyomino file\n");
		(void) fprintf(stderr, "\tturnStyle = %d.\n", turnStyle);
		exit(1);
	}
	w->tetris.numberPolyominoes = tris[w->tetris.next.units -
		MIN_UNITS].mode[diagonal].number[mixed][turnStyle];
	if (w->tetris.numberPolyominoes <= 0 ||
			w->tetris.numberPolyominoes > MAX_START_POLYOMINOES) {
		(void) fprintf(stderr, "Tetris: corrupted polyomino file\n");
		(void) fprintf(stderr, "\tnumberPolyominoes = %d.\n",
			w->tetris.numberPolyominoes);
		exit(1);
	}
	ClearThings(w, False);
	newThing(w);
	newThing(w);
}

#ifndef WINVER
static
#endif
void
ExposeGame(
#ifdef WINVER
TetrisWidget w
#else
Widget renew, XEvent *event, Region region
#endif
)
{
#ifndef WINVER
	TetrisWidget w = (TetrisWidget) renew;

	if (!w->core.visible)
		return;
#endif
	EraseFrame(w);
	ResizeThings(w);
	DrawFrame(w, w->tetris.focus);
	drawField(w);
	drawThing(w);
	drawNext(w);
}

#ifndef WINVER
static
#endif
void
HideGame(TetrisWidget w
#ifndef WINVER
, XEvent *event, char **args, int nArgs
#endif
)
{
	if (w->tetris.sound)
		w->tetris.sound = False;
	w->tetris.paused = True;
	setGame(w, GAME_HIDE);
}

#ifndef WINVER
static
#endif
void
ClearGame(TetrisWidget w
#ifndef WINVER
, XEvent *event, char **args, int nArgs
#endif
)
{
	ClearThings(w, True);
}

#ifndef WINVER
static
#endif
void
EnterGame(TetrisWidget w
#ifndef WINVER
, XEvent *event, char **args, int nArgs
#endif
)
{
	w->tetris.focus = True;
	w->tetris.paused = False;
	DrawFrame(w, w->tetris.focus);
}

#ifndef WINVER
static
#endif
void
LeaveGame(TetrisWidget w
#ifndef WINVER
, XEvent *event, char **args, int nArgs
#endif
)
{
	/* if (w->tetris.firstFall) */
	w->tetris.focus = False;
	w->tetris.paused = True;
	DrawFrame(w, w->tetris.focus);
}

#ifndef WINVER
static void
ReflectGame(TetrisWidget w, XEvent *event, char **args, int nArgs)
{
	MoveGameInput(w, REFLECT);
}

static void
RotateGame(TetrisWidget w, XEvent *event, char **args, int nArgs)
{
	MoveGameInput(w, ROTATE);
}

static void
MoveGameLeft(TetrisWidget w, XEvent *event, char **args, int nArgs)
{
	MoveGameInput(w, MOVE_LEFT);
}

static void
MoveGameRight(TetrisWidget w, XEvent *event, char **args, int nArgs)
{
	MoveGameInput(w, MOVE_RIGHT);
}

static void
DropGame(TetrisWidget w, XEvent *event, char **args, int nArgs)
{
	MoveGameInput(w, DROP);
}

#endif

#ifdef WINVER
Boolean
#else
static void
#endif
IncrementGameLevel(TetrisWidget w
#ifndef WINVER
, XEvent *event, char **args, int nArgs
#endif
)
{
	if (!w->tetris.paused && w->tetris.level < MAX_LEVELS) {
		w->tetris.level++;
		w->tetris.delay = delay[w->tetris.level - 1];
		w->tetris.newRows = 0;
		drawStatus(w);
#ifdef WINVER
		return True;
#endif
	}
#ifdef WINVER
		return False;
#endif
}

#ifdef WINVER
Boolean
#else
static void
#endif
DecrementGameLevel(TetrisWidget w
#ifndef WINVER
, XEvent *event, char **args, int nArgs
#endif
)
{
#ifdef CHEAT
	if (!w->tetris.paused && w->tetris.level > MIN_LEVELS) {
		w->tetris.level--;
		w->tetris.delay = delay[w->tetris.level - 1];
		w->tetris.newRows = 0;
		drawStatus(w);
#ifdef WINVER
		return True;
#endif
	}
#endif
#ifdef WINVER
		return False;
#endif
}

#ifndef WINVER
static
#endif
void
PauseGame(TetrisWidget w
#ifndef WINVER
, XEvent *event, char **args, int nArgs
#endif
)
{
	if (w->tetris.sound)
		w->tetris.sound = False;
	w->tetris.paused = !w->tetris.paused;
#if 0
	if (w->tetris.paused) {
		banner(MSG_PAUSED);
		(void) XIconifyWindow(display, mainWin, screen_num);
	} else {
		drawNext(w);
	}
#endif
}

#ifdef WINVER
Boolean
#else
static void
#endif
IncrementGameUnits(TetrisWidget w
#ifndef WINVER
, XEvent *event, char **args, int nArgs
#endif
)
{
	if (!w->tetris.paused && LEGAL(w->tetris.next.units + 1,
			w->tetris.next.diagonal, w->tetris.next.mixed,
			False)) {
		setGame(w, GAME_INCUNITS);
#ifdef WINVER
		w->tetris.next.units = w->tetris.units;
		w->tetris.next.bonus = False;
		w->tetris.newRows = 0;
		redoNext(w);
		drawNext(w);
		return True;
#endif
	}
#ifdef WINVER
	return False;
#endif
}

#ifdef WINVER
Boolean
#else
static void
#endif
DecrementGameUnits(TetrisWidget w
#ifndef WINVER
, XEvent *event, char **args, int nArgs
#endif
)
{
	if (!w->tetris.paused && LEGAL(w->tetris.next.units - 1,
			w->tetris.next.diagonal, w->tetris.next.mixed,
			False)) {
		setGame(w, GAME_DECUNITS);
#ifdef WINVER
		w->tetris.next.units = w->tetris.units;
		w->tetris.next.bonus = False;
		w->tetris.newRows = 0;
		redoNext(w);
		drawNext(w);
		return True;
#endif
	}
#ifdef WINVER
	return False;
#endif
}

#ifdef WINVER
Boolean
#else
static void
#endif
IncrementGameColumns(TetrisWidget w
#ifndef WINVER
, XEvent *event, char **args, int nArgs
#endif
)
{
	if (!w->tetris.paused && w->tetris.columns < MAX_COLUMNS) {
#ifdef WINVER
		w->tetris.paused = True;
		w->tetris.started = False;
		EraseFrame(w);
		w->tetris.columns++;
		SizeGame(w);
		ResizeThings(w);
		newThing(w);
		newThing(w);
		ClearThings(w, True);
		DrawFrame(w, w->tetris.focus);
		return True;
#else
		setGame(w, GAME_INCCOLUMNS);
#endif
	}
#ifdef WINVER
	return False;
#endif
}

#ifdef WINVER
Boolean
#else
static void
#endif
DecrementGameColumns(TetrisWidget w
#ifndef WINVER
, XEvent *event, char **args, int nArgs
#endif
)
{
	if (!w->tetris.paused && w->tetris.columns > MIN_COLUMNS) {
#ifdef WINVER
		w->tetris.paused = True;
		w->tetris.started = False;
		EraseFrame(w);
		w->tetris.columns--;
		SizeGame(w);
		ResizeThings(w);
		newThing(w);
		newThing(w);
		ClearThings(w, True);
		DrawFrame(w, w->tetris.focus);
		return True;
#else
		setGame(w, GAME_DECCOLUMNS);
#endif
	}
#ifdef WINVER
	return False;
#endif
}

#ifndef WINVER
static
#endif
void
FixedFillGame(TetrisWidget w
#ifndef WINVER
, XEvent *event, char **args, int nArgs
#endif
)
{
	setGame(w, GAME_FIXEDFILL);
}

#ifndef WINVER
static
#endif
void
SuddenAppearGame(TetrisWidget w
#ifndef WINVER
, XEvent *event, char **args, int nArgs
#endif
)
{
	setGame(w, GAME_SUDDENAPPEAR);
#ifdef WINVER
	redoNext(w);
#endif
}

#ifdef WINVER
Boolean
#else
static void
#endif
DiagonalGame(TetrisWidget w
#ifndef WINVER
, XEvent *event, char **args, int nArgs
#endif
)
{
	if (!w->tetris.paused && LEGAL(w->tetris.next.units,
			!w->tetris.next.diagonal, w->tetris.next.mixed,
			False)) {
		setGame(w, GAME_DIAGONAL);
#ifdef WINVER
		w->tetris.next.diagonal = w->tetris.diagonal;
		w->tetris.next.bonus = False;
		w->tetris.newRows = 0;
		redoNext(w);
		drawNext(w);
		return True;
#endif
	}
#ifdef WINVER
	return False;
#endif
}

#ifdef WINVER
Boolean
#else
static void
#endif
MixedGame(TetrisWidget w
#ifndef WINVER
, XEvent *event, char **args, int nArgs
#endif
)
{
	if (!w->tetris.paused && LEGAL(w->tetris.next.units,
			w->tetris.next.diagonal, !w->tetris.next.mixed,
			False)) {
		setGame(w, GAME_MIXED);
#ifdef WINVER
		w->tetris.next.mixed = w->tetris.mixed;
		w->tetris.next.bonus = False;
		w->tetris.newRows = 0;
		redoNext(w);
		drawNext(w);
		return True;
#endif
	}
#ifdef WINVER
	return False;
#endif
}

#ifndef WINVER
static
#endif
void
BonusGame(TetrisWidget w
#ifndef WINVER
, XEvent *event, char **args, int nArgs
#endif
)
{
		setGame(w, GAME_BONUS);
#ifdef WINVER
		w->tetris.next.bonus = False;
		w->tetris.newRows = 0;
		redoNext(w);
		drawNext(w);
#endif
}

#ifndef WINVER
static
#endif
void
CwGame(TetrisWidget w
#ifndef WINVER
, XEvent *event, char **args, int nArgs
#endif
)
{
	setGame(w, GAME_CW);
}

#ifndef WINVER
static
#endif
void
PlainPieceGame(TetrisWidget w
#ifndef WINVER
, XEvent *event, char **args, int nArgs
#endif
)
{
	setGame(w, GAME_PLAINPIECE);
#ifdef WINVER
	ResizeThings(w);
	drawField(w);
#endif
}

#ifndef WINVER
static
#endif
void
ShowNextGame(TetrisWidget w
#ifndef WINVER
, XEvent *event, char **args, int nArgs
#endif
)
{
	if (!w->tetris.paused) {
		setGame(w, GAME_SHOWNEXT);
	}
}

#ifndef WINVER
static
#endif
void
SoundGame(TetrisWidget w
#ifndef WINVER
, XEvent *event, char **args, int nArgs
#endif
)
{
	setGame(w, GAME_SOUND);
}

#ifndef WINVER
static
#endif
void
Fill0Game(TetrisWidget w
#ifndef WINVER
, XEvent *event, char **args, int nArgs
#endif
)
{
	setGame(w, GAME_FILL0);
#ifdef WINVER
	ClearThings(w, True);
#endif
}

#ifndef WINVER
static
#endif
void
Fill1Game(TetrisWidget w
#ifndef WINVER
, XEvent *event, char **args, int nArgs
#endif
)
{
	setGame(w, GAME_FILL1);
#ifdef WINVER
	ClearThings(w, True);
#endif
}

#ifndef WINVER
static
#endif
void
Fill2Game(TetrisWidget w
#ifndef WINVER
, XEvent *event, char **args, int nArgs
#endif
)
{
	setGame(w, GAME_FILL2);
#ifdef WINVER
	ClearThings(w, True);
#endif
}

#ifndef WINVER
static
#endif
void
Fill3Game(TetrisWidget w
#ifndef WINVER
, XEvent *event, char **args, int nArgs
#endif
)
{
	setGame(w, GAME_FILL3);
#ifdef WINVER
	ClearThings(w, True);
#endif
}

#ifndef WINVER
static
#endif
void
Fill4Game(TetrisWidget w
#ifndef WINVER
, XEvent *event, char **args, int nArgs
#endif
)
{
	setGame(w, GAME_FILL4);
#ifdef WINVER
	ClearThings(w, True);
#endif
}

#ifndef WINVER
static
#endif
void
Fill5Game(TetrisWidget w
#ifndef WINVER
, XEvent *event, char **args, int nArgs
#endif
)
{
	setGame(w, GAME_FILL5);
#ifdef WINVER
	ClearThings(w, True);
#endif
}

#ifndef WINVER
static
#endif
void
Fill6Game(TetrisWidget w
#ifndef WINVER
, XEvent *event, char **args, int nArgs
#endif
)
{
	setGame(w, GAME_FILL6);
#ifdef WINVER
	ClearThings(w, True);
#endif
}

#ifndef WINVER
static
#endif
void
Fill7Game(TetrisWidget w
#ifndef WINVER
, XEvent *event, char **args, int nArgs
#endif
)
{
	setGame(w, GAME_FILL7);
#ifdef WINVER
	ClearThings(w, True);
#endif
}

#ifndef WINVER
static
#endif
void
Fill8Game(TetrisWidget w
#ifndef WINVER
, XEvent *event, char **args, int nArgs
#endif
)
{
	setGame(w, GAME_FILL8);
#ifdef WINVER
	ClearThings(w, True);
#endif
}

#ifndef WINVER
static
#endif
void
Fill9Game(TetrisWidget w
#ifndef WINVER
, XEvent *event, char **args, int nArgs
#endif
)
{
	setGame(w, GAME_FILL9);
#ifdef WINVER
	ClearThings(w, True);
#endif
}

#ifndef WINVER
static
#endif
void
Fill10Game(TetrisWidget w
#ifndef WINVER
, XEvent *event, char **args, int nArgs
#endif
)
{
	setGame(w, GAME_FILL10);
#ifdef WINVER
	ClearThings(w, True);
#endif
}
