/*
# X-BASED TETRIS
#
#  TetrisP.h
#
###
#
#  Copyright (c) 1993 - 2006		David A. Bagley, bagleyd@tux.org
#
#  Taken from GENERIC X-BASED TETRIS
#
#  Copyright (c) 1992 - 1995		Q. Alex Zhao, azhao@cc.gatech.edu
#
#			All Rights Reserved
#
#  Permission to use, copy, modify, and distribute this software and
#  its documentation for any purpose and without fee is hereby granted,
#  provided that the above copyright notice appear in all copies and
#  that both that copyright notice and this permission notice appear in
#  supporting documentation, and that the name of the author not be
#  used in advertising or publicity pertaining to distribution of the
#  software without specific, written prior permission.
#
#  This program is distributed in the hope that it will be "playable",
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
*/

/* Private header file for Tetris */

#ifndef _TetrisP_h
#define _TetrisP_h

#include "xwin.h"
#include "timer.h"
#include "polyomino.h"
#include "Tetris.h"

#ifdef WINVER
extern void DestroyGame(HBRUSH brush);
extern void ResizeGame(TetrisWidget w);
extern void SizeGame(TetrisWidget w);
extern void InitializeGame(TetrisWidget w, HBRUSH brush);
extern void ExposeGame(TetrisWidget w);
extern void HideGame(TetrisWidget w);
extern void ClearGame(TetrisWidget w);
extern void EnterGame(TetrisWidget w);
extern void LeaveGame(TetrisWidget w);
extern void MoveGameInput(TetrisWidget w, int direction);
extern void SoundGame(TetrisWidget w);
extern Boolean IncrementGameLevel(TetrisWidget w);
extern Boolean DecrementGameLevel(TetrisWidget w);
extern Boolean IncrementGameUnits(TetrisWidget w);
extern Boolean DecrementGameUnits(TetrisWidget w);
extern Boolean IncrementGameColumns(TetrisWidget w);
extern Boolean DecrementGameColumns(TetrisWidget w);
extern void FixedFillGame(TetrisWidget w);
extern void SuddenAppearGame(TetrisWidget w);
extern Boolean DiagonalGame(TetrisWidget w);
extern Boolean MixedGame(TetrisWidget w);
extern void BonusGame(TetrisWidget w);
extern void CwGame(TetrisWidget w);
extern void PlainPieceGame(TetrisWidget w);
extern void FalldownGame(TetrisWidget w);
extern void PauseGame(TetrisWidget w);
extern void ShowNextGame(TetrisWidget w);
extern void Fill0Game(TetrisWidget w);
extern void Fill1Game(TetrisWidget w);
extern void Fill2Game(TetrisWidget w);
extern void Fill3Game(TetrisWidget w);
extern void Fill4Game(TetrisWidget w);
extern void Fill5Game(TetrisWidget w);
extern void Fill6Game(TetrisWidget w);
extern void Fill7Game(TetrisWidget w);
extern void Fill8Game(TetrisWidget w);
extern void Fill9Game(TetrisWidget w);
extern void Fill10Game(TetrisWidget w);
extern void setGame(TetrisWidget w, int reason);
extern void setGameNext(TetrisWidget w, int reason,
	int randomNumber, int randomRotation, int randomReflection,
	int units, Boolean diagonal, Boolean mixed, Boolean newBonus);

#else

/* This gets around C's inability to do inheritance */
typedef struct _TetrisClassPart {
	int         ignore;
} TetrisClassPart;

typedef struct _TetrisClassRec {
	CoreClassPart coreClass;
	TetrisClassPart tetrisClass;
} TetrisClassRec;

extern TetrisClassRec tetrisClassRec;
#endif

#ifndef BUMPSOUND
#define BUMPSOUND "bump"
#endif

#define SYMBOL ':'

#define IGNORE_DIR (-1)

#define NUM_FLASHES 4

#define MAX_SIDES 4
#define MIN_LEVELS 1
#define MAX_LEVELS 20

#define THRESHOLD(x)	((x)*10)

/*** types ***/

typedef enum {FALL, DROP, MOVE_LEFT, MOVE_RIGHT, ROTATE, REFLECT} move_t;

typedef struct {
	int         units;
	Boolean     diagonal, mixed, bonus;
} Spec;

typedef struct {
	int         units, polyominoNumber;
	int         xpos, ypos;
	int         size, colorNumber;
	int         bonus, randomRotation, randomReflection;
	long        randomNumber;
} Thing;

typedef struct {
	int         pmid, cid;
} Field;

typedef struct _TetrisPart {
	Thing       curThing, nextThing;
	Field      *field;
	GC          thingGC[MAX_START_POLYOMINOES];
	Boolean     fixedFill, suddenAppear, cw, plainPiece, showNext;
	Boolean     sound, resetTime, firstFall, paused, gotNew;
	int         level, fill, completedRows;
	int         startLevel, score, newRows;
	Spec        current, next;
	int         units, columns, rows;
	Boolean     diagonal, mixed, bonus;
	int         delay, numberPolyominoes, pauseStart;
	int         currentPosition;
	Boolean     fall, started, quit, act;
	Boolean     mono, reverse, focus;
	Boolean     scoreOnly, versionOnly;
	int         boxSize;
	Position    viewLength;
	Point       offset;
	GC          frameGC, borderGC, inverseGC, xorGC;
#ifdef WINVER
	char        userName[81], scoreFile[81], bumpSound[81];
#else
	int         menu;
	char       *userName, *scoreFile, *bumpSound;
	Pixel       foreground, background;
	Pixel       frameColor, borderColor;
	Pixel       thingColor[MAX_START_POLYOMINOES];
	XtCallbackList select;
#endif
} TetrisPart;

typedef struct _TetrisRec {
	CorePart    core;
	TetrisPart  tetris;
} TetrisRec;

#endif /* _TetrisP_h */
