/*
# X-BASED WELLTRIS
#
#  WelltrisN.c
#
###
#
#  Copyright (c) 1993 - 2006	David A. Bagley, bagleyd@tux.org
#
#			All Rights Reserved
#
#  Permission to use, copy, modify, and distribute this software and
#  its documentation for any purpose and without fee is hereby granted,
#  provided that the above copyright notice appear in all copies and
#  that both that copyright notice and this permission notice appear in
#  supporting documentation, and that the name of the author not be
#  used in advertising or publicity pertaining to distribution of the
#  software without specific, written prior permission.
#
#  This program is distributed in the hope that it will be "playable",
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
*/

/* Methods file for Welltris */
#include "file.h"
#include "rngs.h"
#include "WelltrisNP.h"

#ifdef WINVER
#ifndef LOGPATH
#define LOGPATH "/usr/tmp"
#endif
#ifndef INIFILE
#define INIFILE "wwelltris.ini"
#endif

#define SECTION "setup"

#else

#ifndef LOGPATH
#ifdef VMS
#define LOGPATH "SYS$SCRATCH:"
#else
#define LOGPATH "/usr/tmp"
#endif
#endif

static void ResizeNextGame(NextWelltrisWidget w);
static void DestroyNextGame(Widget old);
#if 0
static void SizeNextGame(NextWelltrisWidget w);
#endif
static Boolean SetValuesNextGame(Widget current, Widget request, Widget renew);
static void QuitNextGame(NextWelltrisWidget w,
	XEvent *event, char **args, int nArgs);
static void InitializeNextGame(Widget request, Widget renew);
static void ExposeNextGame(Widget renew, XEvent *event, Region region);
static void HideNextGame(NextWelltrisWidget w,
	XEvent *event, char **args, int nArgs);
static void EnterNextGame(NextWelltrisWidget w,
	XEvent *event, char **args, int nArgs);
static void LeaveNextGame(NextWelltrisWidget w,
	XEvent *event, char **args, int nArgs);
static void ShowNextNextGame(NextWelltrisWidget w,
	XEvent *event, char **args, int nArgs);

static char defaultTranslationsNextGame[] =
"<KeyPress>q: Quit()\n\
 Ctrl<KeyPress>C: Quit()\n\
 <KeyPress>osfCancel: Hide()\n\
 <KeyPress>Escape: Hide()\n\
 <KeyPress>osfEscape: Hide()\n\
 Ctrl<KeyPress>[: Hide()\n\
 <KeyPress>0x1B: Hide()\n\
 <KeyPress>n: ShowNext()\n\
 <EnterWindow>: Enter()\n\
 <LeaveWindow>: Leave()";

static XtActionsRec actionsListNextGame[] =
{
	{(char *) "Quit", (XtActionProc) QuitNextGame},
	{(char *) "Hide", (XtActionProc) HideNextGame},
	{(char *) "ShowNext", (XtActionProc) ShowNextNextGame},
	{(char *) "Enter", (XtActionProc) EnterNextGame},
	{(char *) "Leave", (XtActionProc) LeaveNextGame}
};

static XtResource resourcesNextGame[] =
{
	/* 18*6+2 */
	{XtNwidth, XtCWidth, XtRDimension, sizeof (Dimension),
	 XtOffset(NextWelltrisWidget, core.width),
	 XtRString, (caddr_t) "146"},
	{XtNheight, XtCHeight, XtRDimension, sizeof (Dimension),
	 XtOffset(NextWelltrisWidget, core.height),
	 XtRString, (caddr_t) "146"},
	{XtNmono, XtCMono, XtRBoolean, sizeof (Boolean),
	 XtOffset(NextWelltrisWidget, welltris.mono),
	 XtRString, (caddr_t) "FALSE"},
	{XtNreverse, XtCReverse, XtRBoolean, sizeof (Boolean),
	 XtOffset(NextWelltrisWidget, welltris.reverse),
	 XtRString, (caddr_t) "FALSE"},
	{XtNforeground, XtCForeground, XtRPixel, sizeof (Pixel),
	 XtOffset(NextWelltrisWidget, welltris.foreground),
	 XtRString, (caddr_t) XtDefaultForeground},
	{XtNbackground, XtCBackground, XtRPixel, sizeof (Pixel),
	 XtOffset(NextWelltrisWidget, welltris.background),
	 XtRString, (caddr_t) XtDefaultBackground},
	{XtNframeColor, XtCColor, XtRPixel, sizeof (Pixel),
	 XtOffset(NextWelltrisWidget, welltris.frameColor),
	 XtRString, (caddr_t) "cyan" /*XtDefaultForeground*/},
	{XtNborderColor, XtCColor, XtRPixel, sizeof (Pixel),
	 XtOffset(NextWelltrisWidget, welltris.borderColor),
	 XtRString, (caddr_t) "gray25" /*XtDefaultForeground*/},
	{XtNdiameter, XtCDiameter, XtRInt, sizeof (int),
	 XtOffset(NextWelltrisWidget, welltris.diameter),
	 XtRString, (caddr_t) "8"},
	{XtNunits, XtCUnits, XtRInt, sizeof (int),
	 XtOffset(NextWelltrisWidget, welltris.units),
	 XtRString, (caddr_t) "4"},
	{XtNdiagonal, XtCDiagonal, XtRBoolean, sizeof (Boolean),
	 XtOffset(NextWelltrisWidget, welltris.diagonal),
	 XtRString, (caddr_t) "FALSE"},
	{XtNmixed, XtCMixed, XtRBoolean, sizeof (Boolean),
	 XtOffset(NextWelltrisWidget, welltris.mixed),
	 XtRString, (caddr_t) "FALSE"},
	{XtNbonus, XtCBonus, XtRBoolean, sizeof (Boolean),
	 XtOffset(NextWelltrisWidget, nextWelltris.newBonus),
	 XtRString, (caddr_t) "FALSE"},
	{XtNplainPiece, XtCPlainPiece, XtRBoolean, sizeof (Boolean),
	 XtOffset(NextWelltrisWidget, welltris.plainPiece),
	 XtRString, (caddr_t) "FALSE"},
	{XtNshowNext, XtCShowNext, XtRBoolean, sizeof (Boolean),
	 XtOffset(NextWelltrisWidget, welltris.showNext),
	 XtRString, (caddr_t) "TRUE"},
	{XtNmenu, XtCMenu, XtRInt, sizeof (int),
	 XtOffset(NextWelltrisWidget, welltris.menu),
	 XtRString, (caddr_t) "-1"},
	{XtNstart, XtCBoolean, XtRBoolean, sizeof (Boolean),
	 XtOffset(NextWelltrisWidget, welltris.started),
	 XtRString, (caddr_t) "FALSE"},
	{XtNrandomNumber, XtCRandomNumber, XtRInt, sizeof (int),
	 XtOffset(NextWelltrisWidget, nextWelltris.randomNumber),
	 XtRString, (caddr_t) "-1"},
	{XtNrandomRotation, XtCRandomRotation, XtRInt, sizeof (int),
	 XtOffset(NextWelltrisWidget, nextWelltris.randomRotation),
	 XtRString, (caddr_t) "-1"},
	{XtNrandomReflection, XtCRandomReflection, XtRInt, sizeof (int),
	 XtOffset(NextWelltrisWidget, nextWelltris.randomReflection),
	 XtRString, (caddr_t) "-1"},
	{XtNselectCallback, XtCCallback, XtRCallback, sizeof (caddr_t),
	 XtOffset(NextWelltrisWidget, welltris.select),
	 XtRCallback, (caddr_t) NULL}
};

NextWelltrisClassRec nextWelltrisClassRec =
{
	{
		(WidgetClass) & widgetClassRec,		/* superclass */
		(char *) "NextWelltris",	/* class name */
		sizeof (NextWelltrisRec),	/* widget size */
		NULL,		/* class initialize */
		NULL,		/* class part initialize */
		FALSE,		/* class inited */
		(XtInitProc) InitializeNextGame,	/* initialize */
		NULL,		/* initialize hook */
		XtInheritRealize,	/* realize */
		actionsListNextGame,	/* actions */
		XtNumber(actionsListNextGame),	/* num actions */
		resourcesNextGame,	/* resources */
		XtNumber(resourcesNextGame),	/* num resources */
		NULLQUARK,	/* xrm class */
		TRUE,		/* compress motion */
		TRUE,		/* compress exposure */
		TRUE,		/* compress enterleave */
		TRUE,		/* visible interest */
		(XtWidgetProc) DestroyNextGame,	/* destroy */
		(XtWidgetProc) ResizeNextGame,	/* resize */
		(XtExposeProc) ExposeNextGame,	/* expose */
		(XtSetValuesFunc) SetValuesNextGame,	/* set values */
		NULL,		/* set values hook */
		XtInheritSetValuesAlmost,	/* set values almost */
		NULL,		/* get values hook */
		NULL,		/* accept focus */
		XtVersion,	/* version */
		NULL,		/* callback private */
		defaultTranslationsNextGame,	/* tm table */
		NULL,		/* query geometry */
		NULL,		/* display accelerator */
		NULL		/* extension */
	},
	{
		0		/* ignore */
	}
};

WidgetClass nextWelltrisWidgetClass = (WidgetClass) & nextWelltrisClassRec;

static void
setNextGame(NextWelltrisWidget w, int reason)
{
	welltrisCallbackStruct cb;

	cb.reason = reason;
	XtCallCallbacks((Widget) w, (char *) XtNselectCallback, &cb);
}
#endif

#include "image.h"

static Pixmap dr = 0; /* dummy for future double buffering */

static void
CheckThings(NextWelltrisWidget w)
{
	char *buf1 = NULL, *buf2 = NULL;

	if (w->welltris.units < MIN_UNITS || w->welltris.units > MAX_UNITS) {
		intCat(&buf1, "Number of units out of bounds, use ",
			MIN_UNITS);
		stringCat(&buf2, buf1, "...");
		free(buf1);
		intCat(&buf1, buf2, MAX_UNITS);
		free(buf2);
		stringCat(&buf2, buf1, ", defaulting to ");
		free(buf1);
		intCat(&buf1, buf2, DEFAULT_UNITS);
		free(buf2);
		DISPLAY_WARNING(buf1);
		free(buf1);
		w->welltris.units = DEFAULT_UNITS;
		w->welltris.next.units = w->welltris.units;
	}
}

static void
drawUnit(NextWelltrisWidget w, int pmid, int cid, int i, int j)
{
#ifdef IMAGE
	Boolean image = True;
#ifdef WINVER
	HBITMAP hOldBitmap, monoOldBitmap, hBitmap;
	HDC monoDC;
#endif

	if (images[TRUE][pmid % BITMAPS] == None) {
		if (images[TRUE][BITMAPS - 1] != None)
			(void) printf("Bitmap unknown %d (Next).\n", pmid);
		image = False;
	}
#ifdef WINVER
	if (image && (monoDC = CreateCompatibleDC(0)) == NULL) {
		(void) printf("Can not create mono DC (Next).\n");
		image = False;
	}
	if (image && (w->core.memDC = CreateCompatibleDC(w->core.hDC)) == NULL) {
		(void) printf("Can not create compatible DC (Next).\n");
		image = False;
	}
	if (image && (hBitmap = CreateCompatibleBitmap(w->core.hDC,
			w->welltris.boxSize, w->welltris.boxSize)) == NULL) {
		(void) printf("Can not create compatible bitmap (Next).\n");
		image = False;
	}
#endif
	if (image) {
#ifdef WINVER
		FILLRECTANGLE(w, hBitmap, w->welltris.thingGC[cid],
			0, 0,
			w->welltris.boxSize, w->welltris.boxSize);
		monoOldBitmap = (HBITMAP) SelectObject(monoDC, images[TRUE][pmid % BITMAPS]);
		hOldBitmap = SelectObject(w->core.memDC, hBitmap);
		BitBlt(w->core.memDC,
			0, 0, w->welltris.boxSize, w->welltris.boxSize,
			monoDC, 0, 0,
			MERGEPAINT);
		(void) SelectObject(monoDC, monoOldBitmap);
		DeleteDC(monoDC);
		BitBlt(w->core.hDC,
			w->welltris.boxSize * i + w->welltris.offset.x,
			w->welltris.boxSize * j + w->welltris.offset.y,
			w->welltris.boxSize, w->welltris.boxSize,
			w->core.memDC, 0, 0,
			SRCCOPY);
		(void) SelectObject(w->core.memDC, hOldBitmap);
		DeleteObject(hBitmap);
		DeleteDC(w->core.memDC);
#else
		(void) XPutImage(XtDisplay(w), XtWindow(w),
			w->welltris.thingGC[cid],
			images[TRUE][pmid % BITMAPS],
			0, 0,
			w->welltris.boxSize * i + w->welltris.offset.x,
			w->welltris.boxSize * j + w->welltris.offset.y,
			w->welltris.boxSize, w->welltris.boxSize);
#endif
	} else
#endif
	{
		FILLRECTANGLE(w, dr, w->welltris.thingGC[cid],
			w->welltris.boxSize * i + w->welltris.offset.x,
			w->welltris.boxSize * j + w->welltris.offset.y,
			w->welltris.boxSize, w->welltris.boxSize);
	}
}

static void
clearThing(NextWelltrisWidget w)
{
	FILLRECTANGLE(w, dr, w->welltris.inverseGC,
		w->welltris.offset.x, w->welltris.offset.y,
	        w->welltris.viewLength, w->welltris.viewLength);
}

static void
drawThing(NextWelltrisWidget w)
{
	int i, j;

	if (w->welltris.nextThing.polyominoNumber == GAME_IGNORE)
		return;
	clearThing(w);
	for (i = 0; i < w->welltris.nextThing.size; i++)
		for (j = 0; j < w->welltris.nextThing.size; j++)
			if (tris[w->welltris.nextThing.units].polyomino
					[w->welltris.nextThing.polyominoNumber].shape[j][i] != 0) {
				drawUnit(w, polyPixmap(i, j,
					(tris[w->welltris.nextThing.units].polyomino
					[w->welltris.nextThing.polyominoNumber].shape),
					w->welltris.nextThing.size),
					w->welltris.nextThing.colorNumber,
					w->welltris.nextThing.xpos + i,
					w->welltris.nextThing.ypos + j);
	    		}
}

static void
redoNext(NextWelltrisWidget w)
{
	int diagonal = (w->welltris.next.diagonal) ? 1 : 0;
	int turnStyle = tris[w->welltris.next.units -
		MIN_UNITS].mode[diagonal].turnStyle;
	int next_start, i;

	w->welltris.nextThing.units = w->welltris.next.units - MIN_UNITS +
		w->welltris.nextThing.bonus;
	next_start = w->welltris.nextThing.randomNumber %
		tris[w->welltris.nextThing.units].mode
		[diagonal].number[(w->welltris.nextThing.bonus) ?
	       	NOMIX : w->welltris.next.mixed][turnStyle];
	w->welltris.nextThing.colorNumber = next_start;

	if (w->welltris.next.mixed && !w->welltris.nextThing.bonus) {
		w->welltris.nextThing.units = 0;
		while (next_start >= tris[w->welltris.nextThing.units].mode
				[diagonal].number[NOMIX][turnStyle]) {
			next_start -= tris[w->welltris.nextThing.units].mode
				[diagonal].number[NOMIX][turnStyle];
			w->welltris.nextThing.units++;
		}
	}
	w->welltris.nextThing.polyominoNumber =
		tris[w->welltris.nextThing.units].mode
		[diagonal].start[next_start][turnStyle];
	if (tris[w->welltris.nextThing.units].mode[diagonal].turnStyle > NONE)
		for (i = 0; i < w->welltris.nextThing.randomRotation; i++)
			w->welltris.nextThing.polyominoNumber =
				tris[w->welltris.nextThing.units].polyomino
				[w->welltris.nextThing.polyominoNumber].rotation;
	if (tris[w->welltris.nextThing.units].mode[diagonal].turnStyle == ALL)
		for (i = 0; i < w->welltris.nextThing.randomReflection; i++)
			w->welltris.nextThing.polyominoNumber =
				tris[w->welltris.nextThing.units].polyomino
				[w->welltris.nextThing.polyominoNumber].reflection;
	w->welltris.nextThing.size =
		tris[w->welltris.nextThing.units].polyomino
		[w->welltris.nextThing.polyominoNumber].size;
	w->welltris.nextThing.xpos =
		(MAX_UNITS - w->welltris.nextThing.size) / 2;
	w->welltris.nextThing.ypos =
		(MAX_UNITS - w->welltris.nextThing.size) / 2;
}

static void
EraseFrame(NextWelltrisWidget w)
{
	FILLRECTANGLE(w, dr, w->welltris.inverseGC,
		w->welltris.offset.x, w->welltris.offset.y,
		w->welltris.viewLength, w->welltris.viewLength);
}

static void
DrawFrame(NextWelltrisWidget w, Boolean focus)
{
	DRAWRECTANGLE(w, dr, (focus) ? w->welltris.frameGC : w->welltris.borderGC,
		w->welltris.offset.x - 1, w->welltris.offset.y - 1,
		w->welltris.viewLength + 1, w->welltris.viewLength + 1);
}

#if 0
static void
ResetThings(NextWelltrisWidget w)
{
}
#endif

static void
ResizeThings(NextWelltrisWidget w)
{
#ifdef IMAGE
#ifndef WINVER
	Display *display = XtDisplay(w);
	Window window = XtWindow(w);
	XWindowAttributes xgwa;

	(void) XGetWindowAttributes(display, window, &xgwa);
#endif
	free_images(TRUE);
	if (w->welltris.boxSize > 4)
		(void) create_images(
#ifndef WINVER
			display, xgwa.visual,
#endif
			(w->welltris.plainPiece) ? 0 : 1, TRUE,
			w->welltris.boxSize);
#endif
}

#ifdef WINVER
static void
SetValuesNextGame(NextWelltrisWidget w)
{
	struct tagColor {
		int red, green, blue;
	} color;
	char szBuf[80], buf[20];
	int thing;

	w->welltris.units = GetPrivateProfileInt(SECTION, "units",
		DEFAULTUNITS, INIFILE);
	w->welltris.diagonal = (BOOL) GetPrivateProfileInt(SECTION, "diagonal",
		DEFAULTDIAGONAL, INIFILE);
	w->welltris.mixed = (BOOL) GetPrivateProfileInt(SECTION, "mixed",
		DEFAULTMIXED, INIFILE);
	w->welltris.plainPiece = (BOOL) GetPrivateProfileInt(SECTION, "plainPiece",
		FALSE, INIFILE);
	w->welltris.showNext = (BOOL) GetPrivateProfileInt(SECTION, "showNext",
		FALSE, INIFILE);
	w->welltris.mono = (BOOL) GetPrivateProfileInt(SECTION, "mono",
		DEFAULTMONO, INIFILE);
	w->welltris.reverse = (BOOL) GetPrivateProfileInt(SECTION, "reverse",
		DEFAULTREVERSE, INIFILE);
	/* cyan */
	(void) GetPrivateProfileString(SECTION, "frameColor", "0 255 255",
		szBuf, sizeof (szBuf), INIFILE);
	(void) sscanf(szBuf, "%d %d %d",
		&(color.red), &(color.green), &(color.blue));
	w->welltris.frameGC = RGB(color.red, color.green, color.blue);
	/* gray25 */
	(void) GetPrivateProfileString(SECTION, "borderColor", "64 64 64",
		szBuf, sizeof (szBuf), INIFILE);
	(void) sscanf(szBuf, "%d %d %d",
		&(color.red), &(color.green), &(color.blue));
	w->welltris.borderGC = RGB(color.red, color.green, color.blue);
	/* #AEB2C3 */
	(void) GetPrivateProfileString(SECTION, "background", "255 255 255",
		szBuf, sizeof (szBuf), INIFILE);
	(void) sscanf(szBuf, "%d %d %d",
		&(color.red), &(color.green), &(color.blue));
	w->welltris.inverseGC = RGB(color.red, color.green, color.blue);
	for (thing = 0; thing < MAX_START_POLYOMINOES; thing++) {
		(void) sprintf(buf, "thingColor%d", thing);
		(void) GetPrivateProfileString(SECTION, buf,
			polyominoColorString[thing],
			szBuf, sizeof (szBuf), INIFILE);
		(void) sscanf(szBuf, "%d %d %d",
			&(color.red), &(color.green), &(color.blue));
		w->welltris.thingGC[thing] =
			RGB(color.red, color.green, color.blue);
	}
}

void
DestroyNextGame(HBRUSH brush)
{
	(void) DeleteObject(brush);
	/*PostQuitMessage(0);*/
}

#else
static void
GetColor(NextWelltrisWidget w, int thing)
{
	XGCValues values;
	XtGCMask valueMask;
	XColor colorCell, rgb;
	Boolean gotColor = False;

	valueMask = GCForeground | GCBackground;
	if (w->welltris.reverse) {
		values.background = w->welltris.foreground;
	} else {
		values.background = w->welltris.background;
	}
	if (!w->welltris.mono) {
		if (XAllocNamedColor(XtDisplay(w),
				DefaultColormapOfScreen(XtScreen(w)),
				polyominoColorName[thing],
				&colorCell, &rgb)) {
			values.foreground = w->welltris.thingColor[thing] =
				colorCell.pixel;
			gotColor = True;
		} else {
			char *buf1, *buf2;

			stringCat(&buf1, "Color name \"",
				polyominoColorName[thing]);
			stringCat(&buf2, buf1, "\" is not defined for polyomino ");
			free(buf1);
			intCat(&buf1, buf2, thing);
			free(buf2);
			DISPLAY_WARNING(buf1);
			free(buf1);
		}
	}
	if (!gotColor) {
		if (w->welltris.reverse) {
			values.background = w->welltris.foreground;
			values.foreground = w->welltris.background;
		} else {
			values.background = w->welltris.background;
			values.foreground = w->welltris.foreground;
		}
	}
	if (w->welltris.thingGC[thing])
		XtReleaseGC((Widget) w, w->welltris.thingGC[thing]);
	w->welltris.thingGC[thing] = XtGetGC((Widget) w, valueMask, &values);
}

static void
SetAllColors(NextWelltrisWidget w)
{
	XGCValues values;
	XtGCMask valueMask;
	int polyomino;

	valueMask = GCForeground | GCBackground;

	if (w->welltris.reverse) {
		values.background = w->welltris.background;
		values.foreground = w->welltris.foreground;
	} else {
		values.foreground = w->welltris.background;
		values.background = w->welltris.foreground;
	}
	if (w->welltris.inverseGC)
		XtReleaseGC((Widget) w, w->welltris.inverseGC);
	w->welltris.inverseGC = XtGetGC((Widget) w, valueMask, &values);
	if (w->welltris.mono) {
		if (w->welltris.reverse) {
			values.background = w->welltris.foreground;
			values.foreground = w->welltris.background;
		} else {
			values.foreground = w->welltris.foreground;
			values.background = w->welltris.background;
		}
	} else {
		values.foreground = w->welltris.frameColor;
		values.background = w->welltris.background;
	}
	if (w->welltris.frameGC)
		XtReleaseGC((Widget) w, w->welltris.frameGC);
	w->welltris.frameGC = XtGetGC((Widget) w, valueMask, &values);
	if (w->welltris.mono) {
		if (w->welltris.reverse) {
			values.background = w->welltris.foreground;
			values.foreground = w->welltris.background;
		} else {
			values.foreground = w->welltris.foreground;
			values.background = w->welltris.background;
		}
	} else {
		values.foreground = w->welltris.borderColor;
		values.background = w->welltris.background;
	}
	if (w->welltris.borderGC)
		XtReleaseGC((Widget) w, w->welltris.borderGC);
	w->welltris.borderGC = XtGetGC((Widget) w, valueMask, &values);
	for (polyomino = 0; polyomino < MAX_START_POLYOMINOES; polyomino++)
		GetColor(w, polyomino);
}

static Boolean
SetValuesNextGame(Widget current, Widget request, Widget renew)
{
	NextWelltrisWidget c = (NextWelltrisWidget) current;
	NextWelltrisWidget w = (NextWelltrisWidget) renew;
	Boolean redraw = False, setColors = False;

	CheckThings(w);
	if (w->welltris.showNext != c->welltris.showNext) {
		redraw = True;
	}
	if (w->welltris.borderColor != c->welltris.borderColor ||
			w->welltris.frameColor != c->welltris.frameColor ||
			w->welltris.reverse != c->welltris.reverse ||
			w->welltris.mono != c->welltris.mono) {
		SetAllColors(w);
		redraw = True;
	} else if (w->welltris.background != c->welltris.background ||
			w->welltris.foreground != c->welltris.foreground ||
			setColors) {
		SetAllColors(w);
		redraw = True;
	} else if (w->welltris.plainPiece != c->welltris.plainPiece) {
		redraw = True;
	}
#if 0
	if (w->welltris.units != c->welltris.units) {
		w->welltris.next.units = w->welltris.units;
		ResetThings(w);
	} else if (w->welltris.diagonal != c->welltris.diagonal) {
		w->welltris.next.diagonal = w->welltris.diagonal;
		ResetThings(w);
	} else if (w->welltris.mixed != c->welltris.mixed) {
		w->welltris.next.mixed = w->welltris.mixed;
		ResetThings(w);
	}
#endif
	if (w->welltris.menu != -1) {
		switch (w->welltris.menu) {
		case 3:
			w->welltris.menu = -1;
			clearThing(w);
			redraw = True;
			break;
		default:
			w->welltris.menu = -1;
			break;
		}
	}
	if (w->nextWelltris.randomNumber == GAME_CLEARNEXT) {
		if (w->welltris.showNext) {
			drawThing(w);
		} else {
			clearThing(w);
		}
		w->nextWelltris.randomNumber = GAME_IGNORE;
		redraw = True;
	} else if (w->nextWelltris.randomNumber != GAME_IGNORE) {
		w->welltris.nextThing.randomNumber = w->nextWelltris.randomNumber;
		w->welltris.nextThing.randomRotation = w->nextWelltris.randomRotation;
		w->welltris.nextThing.randomReflection = w->nextWelltris.randomReflection;
		w->welltris.nextThing.bonus = w->nextWelltris.newBonus;
		w->welltris.next.units = w->welltris.units;
		w->welltris.next.diagonal = w->welltris.diagonal;
		w->welltris.next.mixed = w->welltris.mixed;
		redoNext(w);
		if (w->welltris.showNext) {
			drawThing(w);
		}
		w->nextWelltris.randomNumber = GAME_IGNORE;
		redraw = True;
	}
	return (redraw);
}

static void
QuitNextGame(NextWelltrisWidget w, XEvent *event, char **args, int nArgs)
{
	Display *display = XtDisplay(w);

	XtCloseDisplay(display);
	exit(0);
}

static void
DestroyNextGame(Widget old)
{
	NextWelltrisWidget w = (NextWelltrisWidget) old;
	int thing;

	for (thing = 0; thing < MAX_START_POLYOMINOES; thing++)
		XtReleaseGC(old, w->welltris.thingGC[thing]);
	XtReleaseGC(old, w->welltris.borderGC);
	XtReleaseGC(old, w->welltris.frameGC);
	XtReleaseGC(old, w->welltris.inverseGC);
	XtRemoveCallbacks(old, XtNselectCallback, w->welltris.select);
}
#endif

#ifndef WINVER
static
#endif
void
ResizeNextGame(NextWelltrisWidget w)
{
#ifdef WINVER
	RECT rect;

	/* Determine size of client area */
	(void) GetClientRect(w->core.hWnd, &rect);
	w->core.width = rect.right;
	w->core.height = rect.bottom;
#endif
	w->welltris.boxSize = (MIN(w->core.height, w->core.width) - 2) /
		MAX_UNITS;
	w->welltris.viewLength = w->welltris.boxSize * MAX_UNITS;
	w->welltris.offset.x = ((int) w->core.width - w->welltris.viewLength) /
		2;
	w->welltris.offset.y = ((int) w->core.height - w->welltris.viewLength) /
		2;
}

#if 0
#ifndef WINVER
static
#endif
void
SizeNextGame(NextWelltrisWidget w)
{
	ResetThings(w);
	ResizeNextGame(w);
}
#endif

#ifndef WINVER
static
#endif
void
InitializeNextGame(
#ifdef WINVER
NextWelltrisWidget w, HBRUSH brush
#else
Widget request, Widget renew
#endif
)
{
	int diagonal, mixed, turnStyle;
#ifdef IMAGE
	int bitmap;
#endif
#ifndef WINVER
	int poly;
	NextWelltrisWidget w = (NextWelltrisWidget) renew;
#endif

#ifdef WINVER
	SetValuesNextGame(w);
#else
	w->welltris.mono = (DefaultDepthOfScreen(XtScreen(w)) < 2 ||
		w->welltris.mono);
	for (poly = 0; poly < MAX_START_POLYOMINOES; poly++)
		w->welltris.thingGC[poly] = NULL;
	w->welltris.borderGC = NULL;
	w->welltris.frameGC = NULL;
	w->welltris.inverseGC = NULL;
#endif
#ifdef IMAGE
	for (bitmap = 0; bitmap < BITMAPS; bitmap++)
		images[TRUE][bitmap] = None;
#endif

	w->welltris.quit = False;
	w->welltris.focus = False;
	w->welltris.next.units = w->welltris.units;
	w->welltris.next.diagonal = w->welltris.diagonal;
	w->welltris.next.mixed = w->welltris.mixed;
	w->welltris.nextThing.polyominoNumber = GAME_IGNORE;
	CheckThings(w);
#if 0
	SizeNextGame(w);
#else
	ResizeNextGame(w);
#endif
#ifdef WINVER
	brush = CreateSolidBrush(w->welltris.inverseGC);
	SETBACK(w->core.hWnd, brush);
	(void) SRAND(time(NULL));
#else
	(void) SRAND(getpid());
	SetAllColors(w);
#endif
	diagonal = (w->welltris.next.diagonal) ? 1 : 0;
	mixed = (w->welltris.next.mixed) ? 1 : 0;
	readThings();
	turnStyle = tris[w->welltris.next.units -
		MIN_UNITS].mode[diagonal].turnStyle;
	/* use intCat */
	if (turnStyle < 0 || turnStyle >= MAX_TURNSTYLE_TYPES) {
		(void) fprintf(stderr, "NextWelltris: corrupted polyomino file\n");
		(void) fprintf(stderr, "\tturnStyle = %d.\n", turnStyle);
		exit(1);
	}
	w->welltris.numberPolyominoes = tris[w->welltris.next.units -
		MIN_UNITS].mode[diagonal].number[mixed][turnStyle];
	if (w->welltris.numberPolyominoes <= 0 ||
			w->welltris.numberPolyominoes > MAX_START_POLYOMINOES) {
		(void) fprintf(stderr, "NextWelltris: corrupted polyomino file\n");
		(void) fprintf(stderr, "\tnumberPolyominoes = %d.\n",
			w->welltris.numberPolyominoes);
		exit(1);
	}
}

#ifndef WINVER
static
#endif
void
ExposeNextGame(
#ifdef WINVER
NextWelltrisWidget w
#else
Widget renew, XEvent *event, Region region
#endif
)
{
#ifndef WINVER
	NextWelltrisWidget w = (NextWelltrisWidget) renew;

	if (!w->core.visible)
		return;
#endif
	EraseFrame(w);
	ResizeThings(w);
	DrawFrame(w, w->welltris.focus);
	if (w->welltris.showNext) {
		drawThing(w);
	}
}

#ifndef WINVER
static
#endif
void
HideNextGame(NextWelltrisWidget w
#ifndef WINVER
, XEvent *event, char **args, int nArgs
#endif
)
{
	setNextGame(w, GAME_HIDE);
}

#ifdef WINVER
void
ValueNextGame(NextWelltrisWidget w,
		int randomNumber, int randomRotation, int randomReflection,
		int units, Boolean diagonal, Boolean mixed, Boolean newBonus)
{
	w->welltris.nextThing.randomNumber = randomNumber;
	w->welltris.nextThing.randomRotation = randomRotation;
	w->welltris.nextThing.randomReflection = randomReflection;
	w->welltris.next.units = w->welltris.units = units;
	w->welltris.next.diagonal = w->welltris.diagonal = diagonal;
	w->welltris.next.mixed = w->welltris.mixed = mixed;
	w->welltris.nextThing.bonus = newBonus;
	redoNext(w);
	if (w->welltris.showNext) {
		drawThing(w);

	}
}
#endif

#ifndef WINVER
static
#endif
void
EnterNextGame(NextWelltrisWidget w
#ifndef WINVER
, XEvent *event, char **args, int nArgs
#endif
)
{
	w->welltris.focus = True;
	DrawFrame(w, w->welltris.focus);
}

#ifndef WINVER
static
#endif
void
LeaveNextGame(NextWelltrisWidget w
#ifndef WINVER
, XEvent *event, char **args, int nArgs
#endif
)
{
	w->welltris.focus = False;
	DrawFrame(w, w->welltris.focus);
}

#ifndef WINVER
static
#endif
void
ShowNextNextGame(NextWelltrisWidget w
#ifndef WINVER
, XEvent *event, char **args, int nArgs
#endif
)
{
	if (!w->welltris.paused) {
		w->welltris.showNext = !w->welltris.showNext;
		if (w->welltris.showNext) {
			drawThing(w);
		} else {
			clearThing(w);
		}
		FLUSH(w);
	}
}
