/*
# X-BASED WELLTRIS
#
#  WelltrisP.h
#
###
#
#  Copyright (c) 1993 - 2006		David A. Bagley, bagleyd@tux.org
#
#  Taken from GENERIC X-BASED TETRIS
#
#  Copyright (c) 1992 - 1995		Q. Alex Zhao, azhao@cc.gatech.edu
#
#			All Rights Reserved
#
#  Permission to use, copy, modify, and distribute this software and
#  its documentation for any purpose and without fee is hereby granted,
#  provided that the above copyright notice appear in all copies and
#  that both that copyright notice and this permission notice appear in
#  supporting documentation, and that the name of the author not be
#  used in advertising or publicity pertaining to distribution of the
#  software without specific, written prior permission.
#
#  This program is distributed in the hope that it will be "playable",
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
*/

/* Private header file for Welltris */

#ifndef _WelltrisP_h
#define _WelltrisP_h

#include "xwin.h"
#include "timer.h"
#include "polyomino.h"
#include "Welltris.h"

#ifdef WINVER
extern void DestroyGame(HBRUSH brush);
extern void ResizeGame(WelltrisWidget w);
extern void SizeGame(WelltrisWidget w);
extern void InitializeGame(WelltrisWidget w, HBRUSH brush);
extern void ExposeGame(WelltrisWidget w);
extern void HideGame(WelltrisWidget w);
extern void ClearGame(WelltrisWidget w);
extern void EnterGame(WelltrisWidget w);
extern void LeaveGame(WelltrisWidget w);
extern void MoveGameInput(WelltrisWidget w, int direction);
extern void SoundGame(WelltrisWidget w);
extern Boolean IncrementGameLevel(WelltrisWidget w);
extern Boolean DecrementGameLevel(WelltrisWidget w);
extern Boolean IncrementGameUnits(WelltrisWidget w);
extern Boolean DecrementGameUnits(WelltrisWidget w);
extern Boolean IncrementGameDiameter(WelltrisWidget w);
extern Boolean DecrementGameDiameter(WelltrisWidget w);
extern void FixedFillGame(WelltrisWidget w);
extern void SuddenAppearGame(WelltrisWidget w);
extern Boolean DiagonalGame(WelltrisWidget w);
extern Boolean MixedGame(WelltrisWidget w);
extern void BonusGame(WelltrisWidget w);
extern void CwGame(WelltrisWidget w);
extern void PlainPieceGame(WelltrisWidget w);
extern void FalldownGame(WelltrisWidget w);
extern void PauseGame(WelltrisWidget w);
extern void ShowNextGame(WelltrisWidget w);
extern void GridGame(WelltrisWidget w);
extern void Fill0Game(WelltrisWidget w);
extern void Fill1Game(WelltrisWidget w);
extern void Fill2Game(WelltrisWidget w);
extern void Fill3Game(WelltrisWidget w);
extern void Fill4Game(WelltrisWidget w);
extern void Fill5Game(WelltrisWidget w);
extern void Fill6Game(WelltrisWidget w);
extern void Fill7Game(WelltrisWidget w);
extern void Fill8Game(WelltrisWidget w);
extern void Fill9Game(WelltrisWidget w);
extern void Fill10Game(WelltrisWidget w);
extern void setGame(WelltrisWidget w, int reason);
extern void setGameNext(WelltrisWidget w, int reason,
	int randomNumber, int randomRotation, int randomReflection,
	int units, Boolean diagonal, Boolean mixed, Boolean newBonus);

#else

/* This gets around C's inability to do inheritance */
typedef struct _WelltrisClassPart {
	int         ignore;
} WelltrisClassPart;

typedef struct _WelltrisClassRec {
	CoreClassPart coreClass;
	WelltrisClassPart welltrisClass;
} WelltrisClassRec;

extern WelltrisClassRec welltrisClassRec;
#endif

#ifndef BUMPSOUND
#define BUMPSOUND "bump"
#endif
#ifndef DRIPSOUND
#define DRIPSOUND "drip"
#endif

#define SYMBOL ':'

#define IGNORE_DIR (-1)

#define NUM_FLASHES 4

#define MAX_SIDES 4
#define MIN_LEVELS 1
#define MAX_LEVELS 20

#define THRESHOLD(x)	((x)*10)
#define EMPTY		(-1)
#define DRAW_COLOR	0
#define XOR_COLOR	(-1)
#define XOR_CLEAR_COLOR	(-2)
#define CLEAR_COLOR	(-3)
#define BORDER_COLOR	(-4)
/* OLD
#define MAX_WIDTH	8
#define MAX_DEPTH	12
#define MAX_SIDES	4
#define MAX_PERIMETER	(MAX_SIDES*MAX_WIDTH)

#define	BOXSIZE		18
#define	NEXTBOXSIZE	16
#define DELTA		1
#define BASE_WIDTH	(MAX_WIDTH*BOXSIZE+DELTA)
#define ENDPT		(3*BASE_WIDTH)
#define GAME_WIDTH	(ENDPT+2*DELTA)
#define MIDPT		(ENDPT/2+DELTA)
#define	OFFSET		20
*/

/*** types ***/

typedef enum {FALL, DROP, MOVE_LEFT, MOVE_RIGHT, ROTATE, REFLECT} move_t;

typedef struct {
	int         units;
	Boolean     diagonal, mixed, bonus;
} Spec;

typedef struct {
	int         units, polyominoNumber;
	int         xpos, ypos;
	int         size, colorNumber;
	int         bonus, randomRotation, randomReflection;
	long        randomNumber;
} Thing;

typedef struct {
	int         pmid, cid;
} Field;

typedef struct _WelltrisPart {
	Thing       curThing, nextThing;
	Field      *wall, *base;
	GC          thingGC[MAX_START_POLYOMINOES];
	GC          thingBrightGC[MAX_START_POLYOMINOES];
	GC          thingDarkGC[MAX_START_POLYOMINOES];
	GC          thingBrighterGC[MAX_START_POLYOMINOES];
	GC          thingDarkerGC[MAX_START_POLYOMINOES];
	Boolean     fixedFill, suddenAppear, cw, plainPiece, showNext, grid;
	Boolean     sound, resetTime, firstFall, paused, gotNew;
	int         level, fill, completedRows;
	int         startLevel, score, newRows;
	Spec        current, next;
	int         units;
	int         frozenWall[MAX_SIDES];
	Boolean     diagonal, mixed, bonus;
	int         delay, numberPolyominoes, pauseStart;
	int         currentPosition;
	Boolean     fall, started, quit, act;
	Boolean     mono, reverse, focus;
	Boolean     scoreOnly, versionOnly;
	int         boxSize;
	Position    viewLength;
	Point       offset;
	GC          frameGC, borderGC, inverseGC, xorGC;
	int         diameter, depth, perimeter, endPt;
#ifdef WINVER
	char        userName[81], scoreFile[81], bumpSound[81], dripSound[81];
#else
	int         menu;
	char       *userName, *scoreFile, *bumpSound, *dripSound;
	Pixel       foreground, background;
	Pixel       frameColor, borderColor;
	Pixel       thingColor[MAX_START_POLYOMINOES];
	XtCallbackList select;
#endif
} WelltrisPart;

typedef struct _WelltrisRec {
	CorePart    core;
	WelltrisPart  welltris;
} WelltrisRec;
#endif /* _WelltrisP_h */
